# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Type::SFDC;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension::ExternalBugs::Type);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Error;

use Bugzilla::WebService::Server::XMLRPC;
use Bugzilla::WebService::Bug;

use Bugzilla::Extension::ExternalBugs::Util qw(mail_error sfdc_get_directory);

use Data::Dumper;
use File::Spec::Functions qw(catfile);
use SOAP::Lite;
use WWW::Salesforce::Simple;

use constant TYPE => 'SFDC';

our $HOSTNAME = Bugzilla->params->{'sfdc_live_server'};
our $SERVER   = "https://login.salesforce.com/services/Soap/u/46.0";
our $CACHE    = sfdc_get_directory('LIVE', $HOSTNAME);
our $WSDL
  = 'file:' . catfile(bz_locations()->{datadir}, 'prod-BugWebServices.xml');

$WWW::Salesforce::SF_URI = 'urn:partner.soap.sforce.com';

my $logger = Bugzilla->logger;

sub get_data {
  my ($self, $ext_bug) = @_;

  unless ($ext_bug->ext_bz_bug_id =~ /^\d+$/) {
    $logger->debug("SFDC Bug id " . $ext_bug->ext_bz_bug_id . " is not a number.");
    return;
  }

  unless ($self->ext_login && $self->ext_password) {
    $logger->debug(
      "SFDC SOAP: Warning: No username/password information for " . $self->url);
    return;
  }

  my $result = undef;
  eval {
    my $sforce = WWW::Salesforce::Simple->new(
      serverurl => $SERVER,
      username  => $self->ext_login,
      password  => $self->ext_password,
    );
    if (not $sforce) {
      die "SFDC SOAP: WWW::Salesforce login failure";
    }
    $logger->debug("SFDC SOAP: Login successful");

    my $soql
      = "SELECT Severity__c, Status, Subject, Account.Name FROM Case WHERE CaseNumber = '"
      . $ext_bug->ext_bz_bug_id . "'";

    $result = $sforce->do_query($soql);
  };

  if ($@) {
    ThrowCodeError('eval_error', {command => "SFDC SOAP", msg => $@});
  }

  if (!$result) {

    # Probably not a valid case number
    return;
  }

  my %data
    = (priority => $result->[0]{Severity__c}, status => $result->[0]{Status},);

  if ( defined($result->[0]{'Account'})
    && defined($result->[0]{'Account'}->{'Name'}))
  {
    $data{description}
      = '[' . $result->[0]{'Account'}->{'Name'} . '] ' . $result->[0]{Subject};
  }
  else {
    $data{description} = '[unknown] ' . $result->[0]{Subject};
  }

  return \%data;
}

sub can_sync {
  my ($self, $ext_bug) = @_;
  return 1;
}

# This does the actual work of sending data to SFDC.
sub send_to_sfdc {
  my $self             = shift;
  my $bug_id           = shift;
  my $bug_xml_data     = shift;
  my $bug_xml_comments = shift;

  unless ($self->ext_login && $self->ext_password) {
    $logger->debug(
      "SFDC SOAP: Warning: No username/password information for " . $self->url);
    die "No user name and password supplied";
  }

  eval {
    my $sforce = WWW::Salesforce::Simple->new(
      serverurl => $SERVER,
      username  => $self->ext_login,
      password  => $self->ext_password,
    );
    if (not $sforce) {
      die "SFDC SOAP: WWW::Salesforce login failure";
    }
    $logger->debug("SFDC SOAP: Login successful");
    my $client = $sforce->get_client(1);
    my @errors;
    my $result = $client->cache_dir($CACHE)->service($WSDL)
      ->updateBug($bug_xml_data, $sforce->get_session_header());
    push(@errors, "updateBug returned failure ($result)")
      if (!$result || $result eq 'false');
    if ($bug_xml_comments) {
      $result = $client->cache_dir($CACHE)->service($WSDL)
        ->upsertComments($bug_xml_comments, $sforce->get_session_header());
      push(@errors, "upsertComments returned failure ($result)")
        if (!$result || $result eq 'false');
    }

    if (@errors) {
      die("Call(s) failed: " . join(", ", @errors));
    }
  };


  if (my $call_error = $@) {
    die $call_error;
  }

  $logger->debug("SFDC SOAP: Success: changes sent for bug $bug_id");
  return 1;
}

sub sfdc_hostname {

  # As per bug 653915 comment 40, the SFDC
  #  link location is different to the RPC call.
  return Bugzilla->params->{sfdc_hostname};
}

1;

__END__

=head1 NAME

Bugzilla::Extension::ExternalBugs::Type::SFDC

=head1 DESCRIPTION

Type represents a SFDC issue. It is an implementation of
L<Bugzilla::Extension::ExternalBugs::Type>, and thus provides all methods that
L<Bugzilla::Extension::ExternalBugs::Type> provides.

This type can sync data from remote SFDC sites to the local site.

This type can push comments to remote SFDC sites.

=cut
