# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::PlotlyReports::Graphs::BugStatus;

use 5.10.1;
use strict;
use warnings;


use strict;
use warnings;

use Carp;
use Bugzilla;
use DateTime;
use Date::Parse;
use Data::Dumper;
use Bugzilla::Extension::PlotlyReports::Util;

my $interval_periods = '(days|weeks|months|years)';

sub new {
  my $class = shift;
  my $opt   = shift;
  my $self  = bless {
    x_label => 'months',
    y_label => 'bugs',

    groups       => [],
    date         => [],
    labels       => [],
    data         => [],
    new          => {},
    open         => {},
    stalled      => {},
    closed       => {},
    averages     => {},
    n_day_open   => {},
    n_day_new    => {},
    n_day_closed => {},

    opt => {%$opt},
  }, $class;
  $self->_init_vars();
  $self->_init_output();
  $self->_init_dates();
  return $self;
}

sub _dbh {
  my $self = shift;
  if (not $self->{dbh}) {
    $self->{dbh} = Bugzilla->dbh();
  }
  return $self->{dbh};
}


sub _init_output {
  my $self = shift;

  # XXX this code could be split into the subclasses
  my $opt = $self->{opt};

  if (!$opt->{chart_title}) {
    $opt->{chart_title} = '';
    if ($opt->{bz_email}) {
      my @emails  = split(',', $opt->{bz_email});
      my $address = $emails[0];
      $address =~ s/\@.*$//;
      $opt->{chart_title} .= $address . ' ';
    }
    elsif ($opt->{orgchart_group}) {
      $opt->{chart_title} .= $opt->{orgchart_group} . ' ';
    }
    if ($opt->{bz_keyword}) {
      $opt->{chart_title} .= $opt->{bz_keyword} . ' ';
    }

    if ($opt->{bz_by_severity}) {
      $opt->{chart_title} .= 'Severity';
    }
    elsif ($opt->{bz_by_priority}) {
      $opt->{chart_title} .= 'Priority';
    }
    elsif ($opt->{show_trend}) {
      $opt->{chart_title} .= 'Trend';
    }
    else {
      $opt->{chart_title} .= 'Status';
    }

    if ($opt->{show_stalled}) {
      $opt->{chart_title} .= ' Stalled';
    }

    if ($opt->{show_averages}) {
      $opt->{chart_title} .= ' Averages';
    }
  }
}

sub _init_vars {
  my $self = shift;

  my $opt = $self->{opt};

  if ($opt->{interval_period}
    && ($opt->{interval_period} !~ /^$interval_periods$/i))
  {
    croak("Unknown interval period: '"
        . $opt->{interval_period}
        . "'\nTry one of $interval_periods\n");
  }

  if ($opt->{interval_count} < 0) {
    $opt->{interval_count} *= -1;
  }

  $self->{x_label} = 'months';
  if ($opt->{interval_period}) {
    $self->{x_label} = $opt->{interval_period};
  }
  $self->{y_label} = 'bugs';

  $self->{groups} = [];
  $self->{data}   = [];
  $self->{labels} = [];
}

# generate dates for comparison and prefill arrays

sub _init_dates {
  my $self = shift;
  my $opt  = $self->{opt};
  my $date = $self->{date};

  if ($opt->{date_from} || $opt->{date_to} || $opt->{interval_period}) {
    my $period = ($opt->{interval_period} || 'months');
    if (!$opt->{date_from}) {
      $opt->{date_from}
        = DateTime->now()->subtract($period => $opt->{interval_count})->ymd('');
    }
    if (!$opt->{date_to}) {
      $opt->{date_to} = DateTime->now()->ymd('');
    }
    my $end_date   = undef;
    my $start_date = undef;
    if (not $opt->{interval_period}) {
      $opt->{date_to} =~ /(....)(..)(..)/;
      $end_date = DateTime->new(year => $1, month => $2, day => $3)->add(months => 1)
        ->set_day(1);

      $opt->{date_from} =~ /(....)(..)(..)/;
      $start_date
        = DateTime->new(year => $1, month => $2, day => $3)->subtract(months => 1)
        ->set_day(1);

      $opt->{interval_count} = -2;
      while (DateTime->compare($end_date, $start_date) > 0) {
        push @$date, $end_date->ymd('');
        $opt->{interval_count}++;
        $end_date->subtract(months => 1);
      }
    }
    else {
      if ($opt->{date_to} =~ /(....)(..)(..)/) {
        $end_date = DateTime->new(year => $1, month => $2, day => $3)->add(days => 1);
      }

      $opt->{date_from} =~ /(....)(..)(..)/;
      $start_date = DateTime->new(year => $1, month => $2, day => $3);

      $opt->{interval_count} = -1;
      while (DateTime->compare($end_date, $start_date) > 0) {
        push @$date, $end_date->ymd('');
        $opt->{interval_count}++;
        $end_date->subtract($opt->{interval_period} => 1);
      }
      push @$date, $end_date->ymd('');
    }
  }
  else {

    my $tmp_date = DateTime->today()->set_day(1);
    push @$date, $tmp_date->ymd('');
    for (my $i = 0; $i <= $opt->{interval_count} + 1; $i++) {
      $tmp_date->subtract(months => 1);
      push @$date, $tmp_date->ymd('');
    }
  }

  $self->{first_date} = $date->[-1];

  return;
}

sub chart_title { $_[0]->{opt}{chart_title} }

sub sub_title { $_[0]->{opt}{title} }

sub _query { }

sub _process_row {
  my ($self, $row) = @_;

  my $opt = $self->{opt};

  my $groups = $self->{groups};

  my $gcount = 0;
  if (defined $row->{by_group}) {
    while ($gcount < @$groups) {
      if ($groups->[$gcount] eq $row->{by_group}) {
        last;
      }
      $gcount++;
    }
    if ($gcount == @$groups) {
      push @$groups, $row->{by_group};
    }
  }

  my $date         = $self->{date};
  my $new          = $self->{new};
  my $open         = $self->{open};
  my $closed       = $self->{closed};
  my $averages     = $self->{averages};
  my $n_day_open   = $self->{n_day_open};
  my $n_day_new    = $self->{n_day_new};
  my $n_day_closed = $self->{n_day_closed};

  for (my $i = 1; $i <= $opt->{interval_count} + 1; $i++) {
    my $id = $i . '-' . $gcount;
    if ( $row->{closed}
      && (defined $row->{closure_ts})
      && ($row->{closure_ts} >= $date->[$i])
      && ($row->{closure_ts} < $date->[$i - 1]))
    {
      if (defined $closed->{$id}) {
        $closed->{$id} += $row->{count};
      }
      else {
        $closed->{$id} = $row->{count};
      }

      if ($opt->{show_averages}) {
        $row->{creation_ts} =~ m/(....)(..)(..)/;
        my $year1  = $1;
        my $month1 = $2;
        my $day1   = $3;
        $row->{closure_ts} =~ m/(....)(..)(..)/;
        my $year2  = $1;
        my $month2 = $2;
        my $day2   = $3;

        $averages->{$id}{'days'} = ($averages->{$id}{'days'} || 0)
          + Delta_Days($year1, $month1, $day1, $year2, $month2, $day2);
        $averages->{$id}{'count'} = ($averages->{$id}{'count'} || 0) + 1;
      }

      # add cumulative n-day average, if applicable

      if ($opt->{n_day_average}) {
        my $j = $i;
        my $c = 0;
        while ($c <= $opt->{n_day_average} && $j >= 1) {
          my $nid = $j . '-' . $gcount;
          if (defined $n_day_closed->{$nid}) {
            $n_day_closed->{$nid} += $row->{count};
          }
          else {
            $n_day_closed->{$nid} = $row->{count};
          }
          $c++;
          $j--;
        }
      }
    }
    if (
      (
        ($row->{creation_ts} >= $date->[$i]) && ($row->{creation_ts} < $date->[$i - 1])
      )
      && (!($row->{closed} && (!defined $row->{closure_ts})))
      )
    {
      if (defined $new->{$id}) {
        $new->{$id} += $row->{count};
      }
      else {
        $new->{$id} = $row->{count};
      }

      # add cumulative n-day average, if applicable

      if ($opt->{n_day_average}) {
        my $j = $i;
        my $c = 0;
        while ($c <= $opt->{n_day_average} && $j >= 1) {
          my $nid = $j . '-' . $gcount;
          if (defined $n_day_new->{$nid}) {
            $n_day_new->{$nid} += $row->{count};
          }
          else {
            $n_day_new->{$nid} = $row->{count};
          }
          $c++;
          $j--;
        }
      }
    }
    if (
      (
        ($row->{creation_ts} < $date->[$i - 1])
        && ( (!$row->{closed})
          || (!defined $row->{closure_ts})
          || ($row->{closure_ts} >= $date->[$i - 1]))
      )
      && (!($row->{closed} && (!defined $row->{closure_ts})))
      )
    {
      if (defined $open->{$id}) {
        $open->{$id} += $row->{count};
      }
      else {
        $open->{$id} = $row->{count};
      }

      # add cumulative n-day average, if applicable

      if ($opt->{n_day_average}) {
        my $j = $i;
        my $c = 0;
        while ($c <= $opt->{n_day_average} && $j >= 1) {
          my $nid = $j . '-' . $gcount;
          if (defined $n_day_open->{$nid}) {
            $n_day_open->{$nid} += $row->{count};
          }
          else {
            $n_day_open->{$nid} = $row->{count};
          }
          $c++;
          $j--;
        }
      }
    }
  }
}

#Add data to an array for plotting.
#
#Params
# ref: to hash containing data
# int: reset lables and data array
# int: (optional) average this data

sub _add_data {
  my $self = shift;
  my %arg  = @_;

  my $data  = delete $arg{data};
  my $clear = delete $arg{clear};
  my $avg   = delete $arg{average} || 0;

  my $groups     = $self->{groups};
  my $opt        = $self->{opt};
  my $graph_data = $self->{data};
  my $labels     = $self->{labels};

  if ($clear) {
    $graph_data = $self->{data} = [[@$labels]];
    $opt->{max_bugs} = 0;
  }

  for (my $i = 0; $i < @$groups; $i++) {
    my @graph_row = ();
    my $local_max = 0;
    for (my $j = $opt->{interval_count} + 1; $j > 0; $j--) {
      my $id    = $j . '-' . $i;
      my $count = 0;

      if ($avg) {
        if ($data->{$id}) {
          $count = $data->{$id}{'days'} / $data->{$id}{'count'};
        }
      }
      else {
        if (not exists $data->{$id}) {
          warn "no data found for [" . $labels->[$j - 1] . "][" . $groups->[$i] . "]\n"
            if $opt->{verbose};
        }
        $count = ($data->{$id} || 0);
      }

      push(@graph_row, $count);

      if ($count > $opt->{max_bugs}) { $opt->{max_bugs} = $count; }
    }

    push @$graph_data, \@graph_row;
  }

  return;
}

sub _pre_processing { }

sub _post_processing {
  my $self   = shift;
  my $groups = $self->{groups};

  if (!@$groups) {
    push @$groups, 'All';
  }

  # generate labels

  my $opt    = $self->{opt};
  my $date   = $self->{date};
  my $labels = $self->{labels};

  for (my $i = $opt->{interval_count} + 1; $i > 0; $i--) {
    my $refdate = $date->[$i];
    my $format  = '%d %b %Y';

    if (!$opt->{interval_period}) {
      $format = '%b';
    }
    else {
      $format = $opt->{date_format} if ($opt->{date_format});
    }

    my $t       = str2time($refdate);
    my $dt      = DateTime->from_epoch(epoch => $t);
    my $usedate = $dt->strftime($format);
    chomp $usedate;
    $usedate =~ s/ 0/ /;
    push @$labels, $usedate;

    if ($opt->{n_day_average}) {
      my $nid          = $i . '-0';
      my $n_day_open   = $self->{n_day_open};
      my $n_day_new    = $self->{n_day_new};
      my $n_day_closed = $self->{n_day_closed};
      my $n_day_divider
        = (($opt->{interval_count} + 2 - $i) > $opt->{n_day_average})
        ? $opt->{n_day_average} + 1
        : ($opt->{interval_count} + 2 - $i);
      $n_day_open->{$nid} = $n_day_open->{$nid}
        && $n_day_divider ? int($n_day_open->{$nid} / $n_day_divider) : 0;
      $n_day_new->{$nid} = $n_day_new->{$nid}
        && $n_day_divider ? int($n_day_new->{$nid} / $n_day_divider) : 0;
      $n_day_closed->{$nid} = $n_day_closed->{$nid}
        && $n_day_divider ? int($n_day_closed->{$nid} / $n_day_divider) : 0;
    }
  }
}

sub generate {
  my $self = shift;

  my $opt = $self->{opt};
  if ($opt->{bz_verified}) {
    $self->_add_data(data => $self->{new}, clear => 1);

    return;    # XXX may belong in the Bugzilla class
  }

  if ($opt->{show_new}) {
    $self->_add_data(data => $self->{new}, clear => 1);
  }

  if ($opt->{show_closed}) {
    $self->_add_data(data => $self->{closed}, clear => 1);
  }

  if ($opt->{show_open}) {
    $self->_add_data(data => $self->{open}, clear => 1);
  }

  if ($opt->{show_stalled}) {
    $self->_add_data(data => $self->{stalled}, clear => 1);
  }

  if ($opt->{show_averages}) {
    $self->_add_data(data => $self->{averages}, clear => 1, average => 1);
  }

  $self->{legend} = $self->{groups};

  # print out graph containing all requested data
  if ($opt->{show_trend}) {
    my $first     = 1;
    my $local_max = 0;

    my $title = $opt->{title} ? $opt->{title} . ": " : "";

    my @legend = ();
    Bugzilla->logger->debug(Dumper($self->{groups}));

    if ($opt->{show_new}) {
      $self->_add_data(data => $self->{new}, clear => $first);
      $first = 0;
      $title .= 'new/';
      push(@legend, 'New');
    }

    if ($opt->{show_closed}) {
      $self->_add_data(data => $self->{closed}, clear => $first);
      $first = 0;
      $title .= 'closed/';
      push(@legend, 'Closed');
    }

    if ($opt->{show_open}) {
      $self->_add_data(data => $self->{open}, clear => $first);
      $first = 0;
      $title .= 'open/';
      push(@legend, 'Open');
    }

    if ($opt->{show_stalled}) {
      $self->_add_data(data => $self->{stalled}, clear => $first);
      $first = 0;
      $title .= 'stalled/';
      push(@legend, 'Stalled');
    }

    if ( $opt->{n_day_average}
      && $opt->{show_open}
      && $opt->{graph_type} =~ /^lines/)
    {
      $self->_add_data(data => $self->{n_day_open}, clear => $first);
      $first = 0;
      my $n_day_title
        = "" . ($opt->{n_day_average} + 1) . "-" . $opt->{interval_period};
      $title .= $n_day_title . '/';
      push(@legend, $n_day_title . ' average');
    }
    $self->{legend} = \@legend;
    $self->{title} = $title;
    Bugzilla->logger->debug(Dumper($self->{legend}));
  }

  return;
}

###############################################################################
# Bugzilla ####################################################################
###############################################################################
package Bugzilla::Extension::PlotlyReports::Graphs::BugStatus::Trend;
use base 'Bugzilla::Extension::PlotlyReports::Graphs::BugStatus';

sub _init_db {
  my $self = shift;

  my $opt = $self->{opt};

  my $dbh = $self->_dbh();

  $opt->{bz_columns} = '';
  $opt->{sql_from}   = '';
  $opt->{where}      = '';
  $opt->{label}      = '';

  if ( $opt->{bz_products}
    || $opt->{bz_versions}
    || $opt->{bz_components}
    || $opt->{orgchart_by_group_member}
    || $opt->{bz_by_status}
    || $opt->{bz_by_priority}
    || $opt->{bz_by_severity})
  {
    $opt->{show_trend} = 0;
  }

  if ($opt->{bz_active_products}) {
    $opt->{bz_all_products} = 0;
  }

  if (($opt->{bz_products} || $opt->{bz_versions} || $opt->{bz_components})
    && !$opt->{orgchart_by_group_member})
  {
    $opt->{bz_columns} = "products.name";
    $opt->{label}      = 'Products';
    $opt->{sql_from} .= "products";
    $opt->{where}    .= "bugs.product_id=products.id";

    if (!$opt->{bz_all_products}) {
      $opt->{where} .= " AND products.disallownew=0";
    }

    if ($opt->{bz_versions} || $opt->{bz_components}) {
      $opt->{bz_columns} .= ", bugs.version";

      if ($opt->{bz_components}) {
        $opt->{bz_columns} .= ", components.name";
        $opt->{label} = 'Components';
        $opt->{sql_from} .= ", components";
        $opt->{where}    .= " AND bugs.component_id=components.id";
      }
    }
  }
  elsif ($opt->{bz_restrict_by_product}) {
    $opt->{sql_from} .= "products";
    $opt->{where}    .= "bugs.product_id=products.id AND (" . join(
      ' OR ',
      map {
            "products.name"
          . (($opt->{bz_restrict_by_product} =~ /\%$/) ? " LIKE " : "=")
          . $dbh->quote($_)
      } split(/[|,]+/, $opt->{bz_restrict_by_product})
    ) . ")";
  }
  elsif (!$opt->{bz_all_products}) {
    $opt->{sql_from} .= "products";
    $opt->{where}    .= "bugs.product_id=products.id AND products.disallownew=0";
  }

  if ( $opt->{bz_email}
    || $opt->{orgchart_email}
    || $opt->{orgchart_group}
    || $opt->{orgchart_all_groups})
  {
    my $group_name = $opt->{orgchart_group};
    if ($opt->{orgchart_by_group_member}) {
      $opt->{bz_columns} = "profiles.login_name";
    }
    if ($opt->{sql_from} ne '') {
      $opt->{sql_from} .= ", ";
    }
    $opt->{sql_from} .= "profiles";
    if ($opt->{where} ne '') {
      $opt->{where} .= " AND ";
    }
    if ($opt->{bz_reporter}) {
      $opt->{where} .= "bugs.reporter=profiles.userid";
    }
    else {
      $opt->{where} .= "bugs.assigned_to=profiles.userid";
    }

#    my $user_where = join(' OR ', map {"profiles.login_name='$_'"} @{$self->{emails}});
# BUGBUG this will need to be changed if other than group is enabled
#    if ($user_where ne '') {
    my $name = $dbh->quote($group_name);

    $opt->{where} .= " AND bugs.assigned_to IN (
          SELECT ugm.user_id FROM user_group_map ugm
          WHERE ugm.group_id = (SELECT id FROM groups where name = $name )
          AND grant_type = 0
      )";

#    }
  }

  if ($opt->{bz_flag}) {
    my @flags = split(/,/, $opt->{bz_flag});
    for (my $flag_id = 0; $flag_id < @flags; $flag_id++) {
      my $flag  = $flags[$flag_id];
      my $value = '';
      if ($flag =~ s/=?([?+-]+)$//) {
        $value = $1;
      }
      if ($opt->{sql_from} and $opt->{sql_from} ne '') {
        $opt->{sql_from} .= ", ";
        $opt->{where}    .= " AND ";
      }
      $opt->{sql_from} .= "flags flags_$flag_id, flagtypes flagtypes_$flag_id";
      $opt->{where}
        .= "bugs.bug_id=flags_$flag_id.bug_id AND flags_$flag_id.type_id=flagtypes_$flag_id.id AND flagtypes_$flag_id.name="
        . $dbh->quote($flag);
      if ($value) {
        $opt->{where} .= " AND ("
          . join(' OR ',
          map { "flags_$flag_id.status=" . $dbh->quote($_) } split(//, $value))
          . ")";
      }
    }
  }

  if ($opt->{bz_keyword}) {
    if ($opt->{sql_from} and $opt->{sql_from} ne '') {
      $opt->{sql_from} .= ", ";
      $opt->{where}    .= " AND ";
    }
    $opt->{sql_from} .= "keywords, keyworddefs";
    $opt->{where}
      .= "bugs.bug_id=keywords.bug_id AND keywords.keywordid=keyworddefs.id AND keyworddefs.name='$opt->{bz_keyword}'";
  }

  if ($opt->{customer}) {
    if ($opt->{where} ne '') {
      $opt->{where} .= " AND ";
    }
    $opt->{where}
      .= "(bugs.bug_id IN (SELECT DISTINCT ext_bz_bug_map.bug_id FROM ext_bz_bug_map WHERE ext_bz_bug_map.ext_bz_id=60))";
  }

  if ($opt->{bz_by_status}) {
    if ($opt->{sql_from} and $opt->{sql_from} ne '') {
      $opt->{sql_from} .= ", ";
      $opt->{where}    .= " AND ";
    }
    $opt->{bz_columns} = "bug_status.sortkey, bugs.bug_status";
    $opt->{sql_from} .= "bug_status";
    $opt->{bz_order_by} = "ORDER BY bug_status.sortkey";
    $opt->{where} .= "bugs.bug_status=bug_status.value";
    $opt->{title} = 'current status for trend' if !$opt->{title};
  }
  elsif ($opt->{bz_by_severity}) {
    if ($opt->{sql_from} and $opt->{sql_from} ne '') {
      $opt->{sql_from} .= ", ";
      $opt->{where}    .= " AND ";
    }
    $opt->{bz_columns} = "bugs.bug_severity, bug_severity.sortkey";
    $opt->{sql_from} .= "bug_severity";
    $opt->{bz_order_by} = "ORDER BY bug_severity.sortkey";
    $opt->{where} .= "bugs.bug_severity=bug_severity.value";
    $opt->{title} = 'current severity for trend' if !$opt->{title};
  }
  elsif ($opt->{bz_by_priority}) {
    if ($opt->{sql_from} and $opt->{sql_from} ne '') {
      $opt->{sql_from} .= ", ";
      $opt->{where}    .= " AND ";
    }
    $opt->{bz_columns} = "bugs.priority, priority.sortkey";
    $opt->{sql_from} .= "priority";
    $opt->{bz_order_by} = "ORDER BY priority.sortkey";
    $opt->{where} .= "bugs.priority=priority.value";
    $opt->{title} = 'current priority for trend' if !$opt->{title};
  }
}

sub _pre_processing {
  my $self = shift;
  my $opt  = $self->{opt};
  my $dbh  = $self->_dbh();

  if ($opt->{bz_restrict_by_component}) {
    my $fc  = $dbh->quote($opt->{bz_restrict_by_component});
    my $sth = $dbh->prepare("SELECT id FROM components WHERE name=$fc");
    $sth->execute();

    my $component_where = '';
    while (my ($component_id) = $sth->fetchrow_array) {
      $component_where .= ' OR ' if $component_where ne '';
      $component_where .= 'component_id=' . $component_id;
    }

    $sth->finish();

    if ($component_where eq '') {
      croak("Cannot restrict by component $opt->{bz_restrict_by_component}.");
    }

    if (defined $opt->{bz_constraint} && $opt->{bz_constraint} ne '') {
      $opt->{bz_constraint} = '(' . $opt->{bz_constraint} . ') AND ';
    }
    else {
      $opt->{bz_constraint} = '';
    }
    $opt->{bz_constraint} .= '(' . $component_where . ')';
  }
}

sub _query {
  my $self = shift;
  my $opt  = $self->{opt};
  my $dbh  = $self->_dbh;

  my $sql_from = '';
  if ($opt->{sql_from} ne '') {
    $sql_from = ", $opt->{sql_from}";
  }

  my $sql_where = '';
  if ($opt->{where} ne '') {
    $sql_where = "WHERE " . $opt->{where};
  }

  if ($opt->{bz_constraint}) {
    if ($opt->{where} eq '') {
      $sql_where = "WHERE " . $opt->{bz_constraint};
    }
    else {
      $sql_where .= " AND ($opt->{bz_constraint})";
    }
  }

  # construct from joins
  my $from = "bugs";
  foreach my $join (@{$opt->{table_join} || []}) {
    $from .= " LEFT OUTER JOIN $join ON (bugs.bug_id=$join.bug_id)";
  }

  # construct columns
  my $columns = '';
  if ($opt->{bz_columns}) {
    $columns = "$opt->{bz_columns}, ";
  }
  my $group_by = "GROUP BY $columns";
  my $order_by = '';
  if ($opt->{bz_order_by}) {
    $order_by = " $opt->{bz_order_by}, ";
  }

  # creation_ts
  $columns
    .= $dbh->sql_date_format('bugs.creation_ts', '%Y%m%d') . " AS creation, ";
  $group_by .= "creation";
  $order_by .= "creation" if $order_by;

  # closure_ts
  if (!$opt->{field_name} || $opt->{field_name} ne 'bug_status') {
    $columns .= $dbh->sql_date_format('bugs.cf_last_closed', '%Y%m%d')
      . " AS closure, (bugs.bug_status='CLOSED') AS closed, ";
    $group_by .= ", closure, closed";
    $order_by .= ", closure, closed" if $order_by;
  }

  $columns .= "COUNT(DISTINCT bugs.bug_id)";

  my $sql = <<BZ_QUERY;
        SELECT $columns
          FROM $from
               $sql_from
          $sql_where
     $group_by
     $order_by
BZ_QUERY

  if ($opt->{bz_verified}) {

    $sql = <<BZ_QUERY_VERIFIED;
            SELECT bugs_activity.added,
BZ_QUERY_VERIFIED

    $sql .= $dbh->sql_date_format('bugs_activity.bug_when', '%Y%m%d')
      . " AS status_ts,\n";

    $sql .= <<BZ_QUERY_VERIFIED;
                   COUNT(DISTINCT bugs_activity.bug_id)
              FROM bugs_activity
             WHERE bugs_activity.fieldid=2 AND
                   bugs_activity.removed='ON_QA'
BZ_QUERY_VERIFIED

    if ($opt->{bz_verified} ne '*') {
      my @verified_states = split(/[ +,]/, $opt->{bz_verified});
      $sql .= " AND (bugs_activity.added='" . $verified_states[0] . "'";
      for (my $s = 1; $s < @verified_states; $s++) {
        $sql .= " OR bugs_activity.added='" . $verified_states[$s] . "'";
      }
      $sql .= ")";
    }

    $sql .= " GROUP BY bugs_activity.added, status_ts";
  }

  Bugzilla->logger->debug($sql);
  return $sql;
}

sub _process_row {
  my ($self, $row) = @_;
  return unless $row;

  my $r = {
    by_group    => undef,
    creation_ts => undef,
    closure_ts  => undef,
    closed      => undef,
    count       => 0,
  };

  my $opt = $self->{opt};
  if ($opt->{bz_by_status} || $opt->{bz_by_severity} || $opt->{bz_by_priority}) {
    $r->{by_group}    = $row->[0];
    $r->{creation_ts} = $row->[1];
    $r->{closure_ts}  = $row->[2];
    $r->{closed}      = $row->[3];
    $r->{count}       = $row->[4];
  }
  elsif ($opt->{bz_verified}) {
    if ($opt->{bz_verified} eq '*') {
      $r->{by_group} = $row->[0];
    }
    $r->{creation_ts} = $row->[1];
    $r->{closure_ts}  = $row->[1];
    $r->{closed}      = 0;
    $r->{count}       = $row->[2];
  }
  elsif ($opt->{orgchart_all_groups}) {
    $r->{creation_ts} = $row->[0];
    $r->{closure_ts}  = $row->[1];
    $r->{closed}      = $row->[2];
    $r->{count}       = $row->[3];
  }
  elsif ($opt->{orgchart_group} && $opt->{orgchart_by_group_member}) {
    $r->{by_group}    = $row->[0];
    $r->{creation_ts} = $row->[1];
    $r->{closure_ts}  = $row->[2];
    $r->{closed}      = $row->[3];
    $r->{count}       = $row->[4];
  }
  elsif (!($opt->{bz_products} || $opt->{bz_versions} || $opt->{bz_components})) {
    $r->{creation_ts} = $row->[0];
    $r->{closure_ts}  = $row->[1];
    $r->{closed}      = $row->[2];
    $r->{count}       = $row->[3];
  }
  elsif (!($opt->{bz_versions} || $opt->{bz_components})) {
    $r->{by_group}    = $row->[0];
    $r->{creation_ts} = $row->[1];
    $r->{closure_ts}  = $row->[2];
    $r->{closed}      = $row->[3];
    $r->{count}       = $row->[4];
  }
  elsif (!$opt->{bz_components}) {
    $r->{by_group}    = $row->[0] . ' ' . $row->[1];
    $r->{creation_ts} = $row->[2];
    $r->{closure_ts}  = $row->[3];
    $r->{closed}      = $row->[4];
    $r->{count}       = $row->[5];
  }
  else {
    $r->{by_group}    = $row->[0] . ' ' . $row->[1] . ' -- ' . $row->[2];
    $r->{creation_ts} = $row->[3];
    $r->{closure_ts}  = $row->[4];
    $r->{closed}      = $row->[5];
    $r->{count}       = $row->[6];
  }

  Bugzilla::Extension::PlotlyReports::Graphs::BugStatus::_process_row($self, $r);

  return $r;
}


sub run {
  my $self = shift;
  my $opt = $self->{opt};

  $self->_init_db();
  my $dbh = $self->_dbh();
  $self->_pre_processing();

  my $sth = $dbh->prepare($self->_query);
  $sth->execute();

  while (my $db_row = $sth->fetchrow_arrayref) {
    my @row = @$db_row;
    shift(@row) if ($opt->{bz_by_status});
    $self->_process_row(\@row);
  }

  $self->_post_processing();

  return;
}

package Bugzilla::Extension::PlotlyReports::Graphs::BugStatus::Trend::History;
use base 'Bugzilla::Extension::PlotlyReports::Graphs::BugStatus::Trend';

sub _element_id {
  my $self    = shift;
  my $element = shift;
  my @array   = @{shift || []};
  my $id      = 0;

  foreach my $member (@array) {
    last if $member eq $element;
    $id++;
  }

  return !@array || $id == @array ? -1 : $id;
}

sub _init_db {
  my $self = shift;

  $self->SUPER::_init_db(@_);

  my $dbh = $self->_dbh();
  my $opt = $self->{opt};
  $opt->{show_trend} = 0;

  # check whether bz_by type is a known type
  my $fielddefs_query = <<FIELDDEFS;
        SELECT  id,
                name,
                description
          FROM  fielddefs
         WHERE  name = ?
            OR  description = ?
FIELDDEFS

  my $sth = $dbh->prepare($fielddefs_query);
  $sth->execute($opt->{bz_by}, $opt->{bz_by});

  my ($id, $name, $description) = $sth->fetchrow_array;

  die "unknown bz_by type: " . $opt->{bz_by} if !$id;

  # check whether bz_by type is supported
  $sth = $dbh->prepare("SHOW COLUMNS FROM bugs");
  $sth->execute();

  my %columns;
  while (my $row = $sth->fetchrow_hashref) {
    $columns{$row->{Field}} = 1 if $row->{Null} eq 'NO';
  }

  die "unsupported bz_by type: " . $opt->{bz_by} if !$columns{$name};

  $opt->{field_id}   = $id;
  $opt->{field_name} = $name;

  $opt->{title} = 'By ' . $description if !$opt->{title};

  # check whether multi-select table exist, if, preload sorted values
  my $groups = $self->{groups};
  @$groups = ();
  if (my @tables
    = @{$dbh->selectcol_arrayref("SHOW TABLES LIKE " . $dbh->quote($name)) || []})
  {
    my $sortkey_query = <<SORTBY;
            SELECT  value
              FROM  $name
             WHERE  isactive = 1
          ORDER BY  sortkey
SORTBY

    $sth = $dbh->prepare($sortkey_query);
    $sth->execute();

    while (my ($value) = $sth->fetchrow_array) {
      push @$groups, $value;
    }
  }
}

sub _pre_processing {
  my $self = shift;

  $self->SUPER::_pre_processing(@_);

  my $dbh = $self->_dbh();
  my $opt = $self->{opt};

  # create temporary table with historical bug activity
  my @time           = localtime(time());
  my $unique         = join('', @time) . int(rand() * 1000);
  my $activity_table = 'bugs_activity_' . $unique;
  push @{$opt->{table_join}}, $activity_table;
  $opt->{bz_columns}
    = "bugs."
    . $opt->{field_name}
    . ", $activity_table.bug_when, $activity_table.removed, $activity_table.added";

  my $temp_table_query = <<TEMP_ACTIVITY_TABLE;
            CREATE TEMPORARY TABLE $activity_table
                SELECT  bug_id,
                        GROUP_CONCAT(date_format(bug_when,'%Y%m%d') ORDER BY bug_when) AS bug_when,
                        GROUP_CONCAT(removed ORDER BY bug_when) AS removed,
                        GROUP_CONCAT(added ORDER BY bug_when) AS added
                FROM    bugs_activity
                WHERE   fieldid=$opt->{field_id}
                GROUP
                    BY  bug_id
TEMP_ACTIVITY_TABLE
  my $create_index_query = "CREATE INDEX tempindex ON $activity_table (bug_id)";

  #$timeout->conf( dbh => $dbh ); # XXX replace with DBIx::Timeout?
  $dbh->do($temp_table_query);
  $dbh->do($create_index_query);
}

sub _process_row {
  my ($self, $row) = @_;
  return unless $row;

  my $opt      = $self->{opt};
  my $groups   = $self->{groups};
  my $date     = $self->{date};
  my $new      = $self->{new};
  my $open     = $self->{open};
  my $closed   = $self->{closed};
  my $averages = $self->{averages};

  my @activity = ();
  my @ts       = ();
  if ($row->[1]) {
    my @removed = split(',', $row->[2]);
    push @ts,       $row->[4];
    push @activity, @removed && $removed[0] ? $removed[0] : 'N/A';
    my @added = split(',', $row->[3]);
    push @ts, split(',', $row->[1]);
    for (my $i = 1; $i < @ts; $i++) {
      push @activity, $i - 1 <= $#added && $added[$i - 1] ? $added[$i - 1] : 'N/A';
    }
  }
  else {
    push @ts,       $row->[4];
    push @activity, $row->[0] || 'N/A';
  }
  my $closure_ts = undef;
  my $count      = 0;
  if ($opt->{field_name} eq 'bug_status') {
    if ($activity[$#activity] eq 'CLOSED') {
      $closure_ts = $ts[$#ts];
    }
    $count = $row->[5];
  }
  else {
    $closure_ts = $row->[5] if $row->[6];
    $count      = $row->[7];
  }

  for (my $i = 1; $i <= $opt->{interval_count} + 1; $i++) {

    # ignore ticket, if created after date range end
    next if $ts[0] >= $date->[$i - 1];

    # retrieve group id of initial activity
    my $e = $self->_element_id($activity[0], $groups);
    if ($e < 0) {
      push @$groups, $activity[0];
      $e = @$groups - 1;
    }

    my $id = $i . '-' . $e;

    # count as new, if created within current date range
    if ($ts[0] >= $date->[$i]) {
      $new->{$id} = ($new->{$id} || 0) + $count;
    }

    # get activity at date range end
    my $j = 0;
    $j++ while $j < $#ts && $ts[$j + 1] < $date->[$i - 1];

    # retrieve group id of given activity
    $e = $self->_element_id($activity[$j], $groups);
    if ($e < 0) {
      push @$groups, $activity[$j];
      $e = @$groups - 1;
    }

    $id = $i . '-' . $e;

    # count as closed, if closed (final) within current date range
    if ($closure_ts && $closure_ts >= $date->[$i] && $closure_ts < $date->[$i - 1])
    {
      $closed->{$id} = ($closed->{$id} || 0) + $count;

      if ($opt->{show_averages}) {
        $ts[0] =~ m/(....)(..)(..)/;
        my $year1  = $1;
        my $month1 = $2;
        my $day1   = $3;
        $ts[$j] =~ m/(....)(..)(..)/;
        my $year2  = $1;
        my $month2 = $2;
        my $day2   = $3;

        $averages->{$id}{'days'} = ($averages->{$id}{'days'} || 0)
          + Delta_Days($year1, $month1, $day1, $year2, $month2, $day2);
        $averages->{$id}{'count'} = ($averages->{$id}{'count'} || 0) + $count;
      }
    }

    # count as open, if not closed (final) by date range end
    if (!$closure_ts || $closure_ts >= $date->[$i - 1] || $opt->{bz_closed_as_open})
    {
      $open->{$id} = ($open->{$id} || 0) + $count;
    }
  }
}

1;

__END__

=head1 Name

Bugzilla::Extension::PlotlyReports::Graphs::BugStatus

=head1 Description

This module generates reports of the current open defects for a product or
products based on bug status.

=head1 METHODS

=head2 new

=over

=item B<Description>

Creates an object to generate a graph with.

  my %opt = (totals => 1, hide_empty => 0);
  my $gf  =
    Bugzilla::Extension::PlotlyReports::Graphs::BugStatus->new({%opt});

  $gt->run();
  $gt->generate();

=back

=head2 run

=over

=item B<Description>

Run the queries to gather the bug data.

=item B<Params>

None.

=item B<Returns>

Nothing.

=back

=head2 generate

=over

=item B<Description>

Process the bug data and generate the results.

=item B<Params>

None.

=item B<Returns>

Nothing.

=back

=cut
