# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension);

use Bugzilla::Bug;
use Bugzilla::Config::Common;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Field;
use Bugzilla::Flag;
use Bugzilla::Util;
use Bugzilla::Status;
use Bugzilla::Keyword;
use Bugzilla::Token;
use Bugzilla::User::Session;
use Bugzilla::Mailer;

use File::Basename;
use List::MoreUtils qw(insert_after any first_index uniq none);
use Text::Diff;
use Graph::Easy;
use IPC::Run qw(run binary);
use Net::LDAP;

use Storable qw(dclone);
use Data::Dumper;

use Bugzilla::Extension::RedHat::DocTypeText;
use JSON qw(from_json);
use List::Compare;
use Bugzilla::Extension::RedHat::Schema;

our $VERSION = 0.3;

## REDHAT EXTENSION BEGIN 822622
use constant REL_ECS_NOTIFY => 64;
## REDHAT EXTENSION END 822622
## REDHAT EXTENSION BEGIN 235855
use constant REL_ARCH_NOTIFY => 65;
## REDHAT EXTENSION END 235855
## REDHAT EXTENSION START 1003675
use constant REL_ALWAYS_CC => 66;
## REDHAT EXTENSION END 1003675

use constant LOCALCONFIG_VARS => (
  {name => 'test_password', default => '',},
  {name => 'LDAPMXpass',    default => '',},
);

# Hooks in to other modules.
BEGIN {
  *Bugzilla::Product::allows_on_dev            = \&_allows_on_dev;
  *Bugzilla::Product::allows_no_clear_acks     = \&_allows_no_clear_acks;
  *Bugzilla::Component::always_cc              = \&_always_cc;
  *Bugzilla::Product::set_allows_on_dev        = \&_set_allows_on_dev;
  *Bugzilla::Product::set_allows_no_clear_acks = \&_set_allows_no_clear_acks;
  *Bugzilla::Component::set_always_cc          = \&_set_always_cc;
  *Bugzilla::all_doc_texts                     = \&_all_doc_texts;
  *Bugzilla::Bug::mandatory_fields             = \&_mandatory_fields;
  *Bugzilla::User::can_see_change              = \&_user_can_see_change;
  *Bugzilla::Bug::can_have_cf                  = \&_bug_can_have_cf;
  *Bugzilla::Bug::can_edit_desc                = \&_bug_can_edit_desc;
  *Bugzilla::Bug::update_comment               = \&_bug_update_comment;
  *Bugzilla::Bug::edit_comment                 = \&_bug_edit_comment;
  *Bugzilla::Comment::activity                 = \&_comment_activity;
  *Bugzilla::User::can_edit_keywords           = \&_user_can_edit_keywords;
  *Bugzilla::User::can_whine                   = \&_user_can_whine;
  *Bugzilla::User::can_whine_at_others         = \&_user_can_whine_at_others;
  *Bugzilla::User::can_minor_update            = \&_user_can_minor_update;
  *Bugzilla::User::can_edit_comments           = \&_user_can_edit_comments;
  *Bugzilla::Field::acl_descr                  = \&_field_acl_descr;
  *Bugzilla::Field::classes                    = \&_field_classes;
  *Bugzilla::User::can_delete                  = \&_user_can_delete;
  *Bugzilla::User::delete_me                   = \&_user_delete_me;
  *Bugzilla::User::is_bug_updated              = \&_user_is_bug_updated;
  *Bugzilla::User::requested_on_bug            = \&_user_requested_on_bug;
  *Bugzilla::User::is_partner_user             = \&_user_is_partner_user;
  *Bugzilla::User::manage_user_access          = \&_user_manage_user_access;

  ## REDHAT EXTENSION START 1824881
  *Bugzilla::Bug::edit_private_groups        = \&_bug_edit_private_groups;
  *Bugzilla::Comment::private_groups         = \&_comment_private_groups;
  *Bugzilla::Comment::private_group_ids      = \&_comment_private_group_ids;
  *Bugzilla::Comment::user_in_private_groups = \&_comment_user_in_private_groups;
  *Bugzilla::Comment::set_private_groups     = \&_comment_set_private_groups;
  *Bugzilla::non_insider_groups              = \&_non_insider_groups;
  ## REDHAT EXTENSION END 1824881
}

#################
# Database schema
#################

sub db_schema_abstract_schema {
  redhat_db_schema_abstract_schema(@_);
  return;
}

##########################################
# Database updates performed in checksetup
##########################################

sub install_update_db {
  redhat_install_update_db(@_);
  return;
}

sub install_before_final_checks {
  redhat_install_before_final_checks(@_);
  return;
}

################
# Installation #
################

sub install_filesystem {
  my ($self, $args) = @_;
  my $files = $args->{'files'};

  my $extensionsdir = bz_locations()->{'extensionsdir'};

  my @files = qw(
    bz-expire-cookies.pl
    bz-schema-audit.pl
    changes.pl
    group-member-audit
    idps.pl
    jobdelta.pl
    jwts.pl
    ldap-audit
    no_attach_data.pl
    pushd.pl
    rebuild_fts_index.sql
    release.pl
    sitemap.pl
    unshare_old_queries.pl
  );

  foreach my $file (@files) {
    my $scriptname = $extensionsdir . "/RedHat/bin/$file";
    $files->{$scriptname} = {perms => Bugzilla::Install::Filesystem::WS_EXECUTE};
  }

  return;
}

sub bug_fields {
  my ($self, $args) = @_;
  my $fields = $args->{fields};
  push(@$fields, 'target_release');

  return;
}

## REDHAT EXTENSION BEGIN 1626576
sub _bug_can_have_cf {
  my ($bug, $cf) = @_;

  my $field = Bugzilla::Field->new({name => $cf});
  return ($field->is_visible_on_bug($bug));
}
## REDHAT EXTENSION END 1626576

## REDHAT EXTENSION BEGIN 1404885
sub _bug_can_edit_desc {
  my ($bug, $user) = @_;

  my $can = 0;

  # A user can edit the description when:
  # 1: they are the reporter
  # 2: they are in the edit group for the bugs product
  if ($bug->reporter->id == $user->id) {
    $can = 1;
  }
  elsif ($user->in_group("editbugs", $bug->product_id)) {
    $can = 1;
  }

  return $can;
}
## REDHAT EXTENSION END 1404885

## REDHAT EXTENSION START 420461
sub _bug_update_comment {
  my ($self, $comment_id, $new_comment, $log_activity, $timestamp, $security)
    = @_;

  # Some validation checks.
  if ($self->{'error'}) {
    ThrowCodeError("bug_error", {bug => $self});
  }
  detaint_natural($comment_id)
    || ThrowCodeError('bad_arg',
    {argument => 'comment_id', function => 'update_comment'});

  # The comment ID must belong to this bug.
  my ($current_comment_obj) = grep($_->id == $comment_id, @{$self->comments});
  $current_comment_obj
    || ThrowCodeError('bad_arg',
    {argument => 'comment_id', function => 'update_comment'});

  # If the new comment is undefined, then there is nothing to update.
  # To delete a comment, an empty string should be passed.
  return unless defined $new_comment;
  $new_comment =~ s/\s*$//s;       # Remove trailing whitespaces.
  $new_comment =~ s/\r\n?/\n/g;    # Handle Windows and Mac-style line endings.
  trick_taint($new_comment);

  # We assume _check_comment() has already been called earlier.
  Bugzilla->dbh->do('UPDATE longdescs SET thetext = ? WHERE comment_id = ?',
    undef, ($new_comment, $comment_id));
  $self->_sync_fulltext();

  # Log a diff of the old and new comment to the longdescs activity table
  if ($log_activity) {
    my $dbh         = Bugzilla->dbh;
    my $old_comment = $current_comment_obj->body;
    $timestamp ||= $dbh->selectrow_array("SELECT NOW()");
    $dbh->do(
      "INSERT INTO longdescs_activity "
        . "(comment_id, who, change_when, old_comment, security) "
        . "VALUES (?, ?, ?, ?, ?)",
      undef,
      ($comment_id, Bugzilla->user->id, $timestamp, $old_comment, $security)
    );
  }

  # Update the comment object with this new text.
  $current_comment_obj->{'thetext'} = $new_comment;

  return;
}
## REDHAT EXTENSION END 420461

## REDHAT EXTENSION START 420461
sub _bug_edit_comment {
  my ($self, $id, $comment, $security, $count) = @_;
  $comment = $self->_check_comment($comment);
  return if $comment eq '';

  # So we really want to edit comment. Make sure we are allowed to do so.
  # 1. Is the user allowed to even add comments to this bug
  # 2. Is the user allowed to edit previous comments
  my $privs;
  $self->check_can_change_field('longdesc', 0, 1, \$privs)
    || ThrowUserError('illegal_change', {field => 'longdesc', privs => $privs});

  my $can = 0;

  $can = 1 if (Bugzilla->user->can_edit_comments());

  if ($count == 0 && $self->can_edit_desc(Bugzilla->user)) {
    $can = 1;
  }

  $can || ThrowUserError('illegal_change', {field => 'longdesc'});

  $self->{edited_comments} ||= [];
  push(
    @{$self->{edited_comments}},
    {
      id       => $id,
      comment  => $comment->{thetext},
      security => $security,
      count    => $count
    }
  );

  return;
}
## REDHAT EXTENSION END 420461

## REDHAT EXTENSION START 1824881
sub _comment_private_groups {
  my ($self) = @_;

  unless ($self->{private_groups}) {
    my $dbh = Bugzilla->dbh;

    my $group_ids
      = $dbh->selectcol_arrayref(
      'SELECT group_id FROM comment_private_groups WHERE comment_id = ?',
      undef, $self->id);

    my $groups;
    $groups = Bugzilla::Group->new_from_list($group_ids) if (@$group_ids);
    $self->{private_groups} = $groups // [];
  }

  return $self->{private_groups};
}

sub _comment_private_group_ids {
  my ($self) = @_;

  unless ($self->{private_group_ids}) {
    my $dbh = Bugzilla->dbh;

    my $group_ids
      = $dbh->selectcol_arrayref(
      'SELECT group_id FROM comment_private_groups WHERE comment_id = ?',
      undef, $self->id);

    $self->{private_group_ids} = $group_ids // [];
  }

  return $self->{private_group_ids};
}

sub _comment_user_in_private_groups {
  my ($self, $the_user) = @_;

  my $can = 0;

  if (@{$self->private_group_ids}) {
    $can = any { $the_user->in_group_id($_) } @{$self->private_group_ids};
  }

  return $can;
}

sub _bug_edit_private_groups {
  my ($self, $comment_id, $groups) = @_;

  my $dbh = Bugzilla->dbh;

  my $user = Bugzilla->user;
  return unless ($user->is_insider);

  my $comment = Bugzilla::Comment->check({id => $comment_id});

  # So we really want to edit comment. Make sure we are allowed to do so.
  # 1. Is the user allowed to even add comments to this bug
  # 2. Is the user allowed to edit previous comments
  my $privs;
  $self->check_can_change_field('longdesc', 0, 1, \$privs)
    || ThrowUserError('illegal_change', {field => 'longdesc', privs => $privs});

  $self->{edited_comment_private_groups} ||= [];

  my ($added, $removed) = $comment->set_private_groups($groups);

  if (@$added || @$removed) {
    push(
      @{$self->{edited_comment_private_groups}},
      {
        comment_id => $comment->id,
        added      => $added,
        removed    => $removed,
        count      => $comment->count,
      }
    );
  }

  return;
}
## REDHAT EXTENSION END 1824881

## REDHAT EXTENSION START 420461
sub _comment_activity {
  my ($self, $activity_sort_order) = @_;

  return $self->{'activity'} if $self->{'activity'};

  my $dbh = Bugzilla->dbh;
  my $query
    = 'SELECT longdescs_activity.comment_id AS id, profiles.userid, '
    . $dbh->sql_date_format('longdescs_activity.change_when', '%Y.%m.%d %H:%i:%s')
    . '
                        AS time, longdescs_activity.old_comment AS old, security
                   FROM longdescs_activity
             INNER JOIN profiles
                     ON profiles.userid = longdescs_activity.who
                  WHERE longdescs_activity.comment_id = ?';
  $query .= " ORDER BY longdescs_activity.change_when DESC";
  my $sth = $dbh->prepare($query);
  $sth->execute($self->id);

  # We are shifting each comment activity body 1 back. The reason this
  # has to be done is that the longdescs_activity table stores the comment
  # body that the comment was before the edit, not the actual new version
  # of the comment.
  my @activity;
  my $new_comment;
  my $last_old_comment;
  my $count          = 0;
  my $is_secure      = 0;
  my $can_see_secure = Bugzilla->user->can_edit_comments();

  while (my $change_ref = $sth->fetchrow_hashref()) {
    my %change = %$change_ref;
    $change{'author'} = new Bugzilla::User($change{'userid'});
    if ($count == 0) {
      $change{new} = $self->body;
    }
    else {
      $change{new} = $new_comment;
    }
    $new_comment      = $change{old};
    $last_old_comment = $change{old};

    # Hide values that predate the latest security change
    unless ($can_see_secure) {
      if ($is_secure) {
        $change{old}      = 'REDACTED';
        $change{new}      = 'REDACTED';
        $last_old_comment = 'REDACTED';
      }
      elsif ($change{security}) {
        $is_secure        = 1;
        $change{old}      = 'REDACTED';
        $last_old_comment = 'REDACTED';
      }
    }

    push(@activity, \%change);
    $count++;
  }

  return [] if !@activity;

  # Store the original comment as the first or last entry
  # depending on sort order
  push(
    @activity,
    {
      author   => $self->author,
      body     => $last_old_comment,
      time     => $self->creation_ts,
      original => 1
    }
  );

  $activity_sort_order = $activity_sort_order
    || Bugzilla->user->settings->{'comment_sort_order'}->{'value'};

  if ($activity_sort_order eq "oldest_to_newest") {
    @activity = reverse @activity;
  }

  $self->{'activity'} = \@activity;

  return $self->{'activity'};
}
## REDHAT EXTENSION END 420461

sub _user_can_see_change {
  my ($user, $field_name, $change) = @_;

  my $visible = 1;

  if ($field_name =~ m/^cf_/) {
    $visible = $user->can_view_cf($field_name);
  }
  elsif ($field_name eq 'ext_bz_bug_map.ext_bz_bug_id') {
    my $etid = $change->{'object_id'};
    $visible = $user->can_see_ext_bz($etid);
  }
  elsif ($field_name eq 'flagtypes.name') {
    if (defined $change->{'object_id'} && $change->{'object_id'} ne '') {
      my $flag = Bugzilla::FlagType->new($change->{'object_id'});
      $visible = $user->can_see_flag($flag);
    }
    else {
      #deleted flag?
      $visible = 0;
    }
  }
  elsif ($field_name eq 'rh_rule') {

    # If multiple rules make exactly the same changes they get concatenated ...
    my @ids   = split ' ', $change->{'new'};
    my $rules = Bugzilla::Extension::RuleEngine::Rule->new_from_list(\@ids);
    $visible = any { $user->can_see_rule($_) } @$rules;
  }

  return ($visible);
}

sub _user_can_edit_comments {
  my $self = shift;

  if (!defined $self->{'can_edit_comments'}) {
    $self->{'can_edit_comments'}
      = ($self->in_group('admin') || $self->in_group('edit_comment'));
  }

  return ($self->{'can_edit_comments'});
}

sub _user_can_edit_keywords {
  my $self = shift;

  if (!defined $self->{'can_edit_keywords'}) {
    $self->{'can_edit_keywords'}
      = ($self->in_group('admin') || $self->in_group('editkeywords'));
  }

  return ($self->{'can_edit_keywords'});
}

sub _user_can_minor_update {
  my $self = shift;

  if (!defined $self->{'can_minor_update'}) {
    $self->{'can_minor_update'} = 1;
  }

  return ($self->{'can_minor_update'});
}

sub _user_can_whine {
  my $self = shift;

  if (!defined $self->{'can_whine'}) {
    $self->{'can_whine'}
      = ($self->in_group('admin') || $self->in_group('bz_canusewhines'));
  }

  return ($self->{'can_whine'});
}

sub _user_can_whine_at_others {
  my $self = shift;

  if (!defined $self->{'can_whine_at_others'}) {
    $self->{'can_whine_at_others'}
      = ($self->in_group('admin') || $self->in_group('bz_canusewhineatothers'));
  }

  return ($self->{'can_whine_at_others'});
}

sub _field_acl_descr {
  my ($self, $user) = @_;

  my $txt = '';

  if ($self->user_can_view($user)) {
    my $grps = $self->view_groups();
    if ($grps) {
      $txt .= "\n\nView Groups: ";
      $txt .= join(', ', map { $grps->{$_}->name } keys(%$grps));
    }
  }

  if ($self->user_can_edit($user)) {
    my $grps = $self->edit_groups();
    if ($grps) {
      $txt = "\n" if ($txt eq '');
      $txt .= "\nEdit Groups: ";
      $txt .= join(', ', map { $grps->{$_}->name } keys(%$grps));
    }
  }

  if ($self->user_can_admin($user)) {
    my $grps = $self->admin_groups();
    if ($grps) {
      $txt = "\n" if ($txt eq '');
      $txt .= "\nAdmin Groups: ";
      $txt .= join(', ', map { $grps->{$_}->name } keys(%$grps));
    }
  }

  return ($txt);
}

# This is here because it needs the Group->category customisation
sub _field_classes {
  my $self = shift;

  unless ($self->{classes}) {
    $self->{classes} = [];
    my @cats;

    my $grps = $self->view_groups();

    foreach my $grp (keys %{$grps}) {
      push(@cats, $grps->{$grp}->category);
    }

    if (any { $_ eq 'Partner' } @cats) {
      push(@{$self->{classes}}, 'bz_partner_data');
    }
    elsif (any { $_ eq 'Red Hat' } @cats) {
      push(@{$self->{classes}}, 'bz_private_data');
    }
  }

  return ($self->{classes});
}

sub bug_filter_change {
  my ($self, $args) = @_;

  my $user      = $args->{user};
  my $field     = $args->{field};
  my $added     = $args->{added};
  my $removed   = $args->{removed};
  my $visible   = $args->{visible};
  my $object_id = $args->{object_id};

  $user = defined $user ? $user : Bugzilla->user;

  my $field_object = Bugzilla::Field->new({name => $field});

  unless ($field_object->user_can_view($user)) {
    $$visible = 0;
  }

  ## REDHAT EXTENSION START 847166
  if ($field eq 'flagtypes.name' and $object_id) {
    my $flag          = new Bugzilla::FlagType($object_id);
    my $view_group_id = defined $flag ? $flag->view_group_id : 0;
    if ($view_group_id and not $user->in_group_id($view_group_id)) {
      $$visible = 0;
    }
  }

  ## REDHAT EXTENSION BEGIN 847166 1039815
  if ($field_object->type == Bugzilla::Constants::FIELD_TYPE_MULTI_SELECT) {
    my $values = $user->see_field_values($field_object);

    # not defined = no restrictions, scalar = some values
    if (defined $values) {
      my $partner = $removed || $added;
      if (none { $_->name eq $partner } @$values) {
        $$visible = 0;
      }
    }
  }
  ## REDHAT EXTENSION END 847166 1039815

  ## REDHAT EXTENSION START 915533
  # Some See Also changes should not be seen
  if ($field eq 'see_also') {
    my $regexpes = {redhat => '/rs/cases/[0-9]+/attachments/[a-f0-9\-]+$',};

    my $see_also = $removed || $added;
    while (my ($group, $regexp) = each %$regexpes) {
      if (!Bugzilla->user->in_group($group) && $see_also =~ $regexp) {
        $$visible = 0;
      }
    }
  }
  ## REDHAT EXTENSION END 915533

  return;
}

sub bug_filter_fields {
  my ($self, $args) = @_;

  my $fields = $args->{fields};
  my @filtered_fields;

  foreach my $field (@$fields) {
    if ($field->type == Bugzilla::Constants::FIELD_TYPE_MULTI_SELECT) {

      # Some fields (like partner, are restricted on what you can see)
      my $values = Bugzilla->user->see_field_values($field);

      # not defined = no restrictions, scalar = some values
      if (not defined $values or scalar(@$values)) {
        push @filtered_fields, $field;
      }
    }
    ## REDHAT EXTENSION END 1039815
    else {
      push(@filtered_fields, $field);
    }
  }

  @$fields = @filtered_fields;

  return;
}

sub buglist_columns {
  my ($self, $args) = @_;

  my $columns = $args->{'columns'};
  my $dbh     = Bugzilla->dbh;

  state $pg_suffix
    = lc(Bugzilla->localconfig->{db_driver}) eq 'pg' ? '::text' : '';

  ## REDHAT EXTENSION BEGIN 447765 876015
  if (Bugzilla->user->id) {
    $columns->{assigned_to}
      = {name => "map_assigned_to.login_name", title => "Assignee"};
    $columns->{reporter} = {name => "map_reporter.login_name", title => "Reporter"};
    $columns->{qa_contact}
      = {name => "map_qa_contact.login_name", title => "QA Contact"};
    $columns->{docs_contact}
      = {name => "map_docs_contact.login_name", title => "Docs Contact"};
  }
  else {
    $columns->{assigned_to}
      = {name => "map_assigned_to.realname", title => "Assignee"};
    $columns->{reporter} = {name => "map_reporter.realname", title => "Reporter"};
    $columns->{qa_contact}
      = {name => "map_qa_contact.realname", title => "QA Contact"};
    $columns->{docs_contact}
      = {name => "map_docs_contact.realname", title => "Docs Contact"};
  }
  ## REDHAT EXTENSION END 447765 876015

  $columns->{alias}{name}
    = Bugzilla->dbh->sql_group_concat('DISTINCT map_alias.alias' . $pg_suffix);
  $columns->{target_release}{name} = Bugzilla->dbh->sql_group_concat(
    'DISTINCT map_target_release.value' . $pg_suffix);

  ## REDHAT EXTENSION BEGIN 813638
  $columns->{'blocked'}{name}
    = Bugzilla->dbh->sql_group_concat('DISTINCT map_blocks.blocked' . $pg_suffix);
  $columns->{'dependson'}{name} = Bugzilla->dbh->sql_group_concat(
    'DISTINCT map_dependencies.dependson' . $pg_suffix);
  ## REDHAT EXTENSION END 813638
  # This exists for backwards compatibility
  # It can be removed when *.runQuery RPC calls are removed
  $columns->{'blockedby'}{name} = Bugzilla->dbh->sql_group_concat(
    'DISTINCT map_blockedby.blocked' . $pg_suffix);

  ## REDHAT EXTENSION START 536717
  $columns->{blocks_all}    = {title => 'Blocks (full)',     name => 'NULL'};
  $columns->{dependson_all} = {title => 'Depends On (full)', name => 'NULL'};
  ## REDHAT EXTENSION END 536717

  ## REDHAT EXTENSION START 1261259 1320419
  if (Bugzilla->user->id) {
    $columns->{flags_all} = {
      title => 'Flags (full)',
      name  => $dbh->sql_group_concat(
        $dbh->sql_string_concat(
          'map_flagtypes.name', 'map_flags.status',
          "COALESCE(map_flag_requestee.login_name, '')"
        ),
        undef, undef,
        'map_flagtypes.sortkey, map_flagtypes.name, map_flag_requestee.login_name'
      ),
    };
  }
  else {
    $columns->{flags_all} = {title => 'Flags (full)', name => 'NULL'};

  }
  ## REDHAT EXTENSION END 1261259 1320419

  return;
}

sub buglist_column_joins {
  my ($self, $args) = @_;
  my $joins = $args->{'column_joins'};
  my $user  = $args->{'user'};

  $joins->{'flagtypes.name'}{then_to}{extra}
    = ["(map_flagtypes.view_group_id IS NULL OR map_flagtypes.view_group_id IN ("
      . $user->groups_as_string . "))"
    ];

  $joins->{alias} = {table => 'bugs_aliases', as => 'map_alias',};
  $joins->{target_release}
    = {table => 'bugs_release', as => 'map_target_release',};

  ## RED HAT EXTENSION BEGIN 813638
  $joins->{blocked}
    = {table => 'dependencies', to => 'dependson', as => 'map_blocks',};
  $joins->{dependson}
    = {table => 'dependencies', to => 'blocked', as => 'map_dependencies',};
  $joins->{blocks_all}
    = {table => 'dependencies', to => 'dependson', as => 'map_blocks_all',};
  $joins->{dependson_all}
    = {table => 'dependencies', to => 'blocked', as => 'map_dependencies_all',};

  ## RED HAT EXTENSION END 813638
  # This exists for backwards compatibility
  # It can be removed when *.runQuery RPC calls are removed
  $joins->{blockedby}
    = {table => 'dependencies', to => 'dependson', as => 'map_blockedby',};

  $joins->{flags_all} = {
    as      => 'map_flags',
    table   => 'flags',
    extra   => ['map_flags.attach_id IS NULL'],
    then_to => [
      {
        to    => 'id',
        as    => 'map_flagtypes',
        from  => 'map_flags.type_id',
        table => 'flagtypes',
        extra => [
          "(map_flagtypes.view_group_id IS NULL OR map_flagtypes.view_group_id IN ("
            . $user->groups_as_string . "))"
        ]
      },
      {
        to    => 'userid',
        as    => 'map_flag_requestee',
        from  => 'map_flags.requestee_id',
        table => 'profiles'
      },
    ]
  };

  return;
}

sub product_end_of_create {
  my ($self, $args) = @_;

  my $product = $args->{'product'};

  # Add every new product to the insidergroup group automatically
  my $private_group
    = Bugzilla::Group->new({name => Bugzilla->params->{'insidergroup'}});
  if ($private_group) {
    $product->set_group_controls($private_group,
      {membercontrol => CONTROLMAPSHOWN, othercontrol => CONTROLMAPSHOWN});
    $product->update;
  }

  return;
}

sub bug_check_can_change_field {
  my ($self, $args) = @_;

  my ($bug, $field, $new_value, $old_value, $priv_results)
    = @$args{qw(bug field new_value old_value priv_results)};

  my $user    = Bugzilla->user;
  my $product = $bug->product_obj();


  ## REDHAT EXTENSION START 1441696
  if (!defined $product) {
    ThrowCodeError('missing_product_object');
  }
  ## REDHAT EXTENSION END 1441696

  ## REDHAT EXTENSION START 406141 822804
# Disallow a bug from being reopened if currently CLOSED/ERRATA unless user is in 'admin' group
  if ($field =~ /^(?:bug_status|resolution|dup_id)$/) {
    if ( $product->name =~ /Red Hat Enterprise Linux/
      && $bug->bug_status eq 'CLOSED'
      && $bug->resolution eq 'ERRATA'
      && (any { $_->ext_bz_id == 139 } @{$bug->external_bugs}))
    {
      if (not $user->in_group('admin')) {
        push(@$priv_results, PRIVILEGES_REQUIRED_EMPOWERED);
        return;
      }
    }
  }

  ## REDHAT EXTENSION BEGIN 913098
# Allow CLOSED status on RHEL products if user is a RH employee OR in the rhel_beta group
# If we are not allowed to use the CLOSED state, then don't display resolution or dupe id.
  if ($field =~ /^(?:resolution|dup_id|bug_status)$/) {
    unless ($user->is_insider() or $user->in_group('rhel_beta')) {
      if ($product->name =~ /^Red Hat Enterprise Linux/) {
        if ($field eq 'bug_status') {

          # Must be redhat or rhel_beta to change state to closed.
          if ($new_value eq 'CLOSED') {
            push(@$priv_results, PRIVILEGES_REQUIRED_EMPOWERED);
          }
        }
        else {
          # Not allowed to close the bug, don't show the duplicate related fields
          push(@$priv_results, PRIVILEGES_REQUIRED_EMPOWERED);
        }
      }
    }
  }
  ## REDHAT EXTENSION END 913098

  # Disallow priority change if not in the setpriority group
  if (  $field eq "priority"
    and $new_value ne $old_value
    and ($bug->classification eq 'Red Hat' and not $user->in_group('setpriority')))
  {
    push(@$priv_results, PRIVILEGES_REQUIRED_EMPOWERED);
    return;
  }
  ## REDHAT EXTENSION END 406141 822804 495985 1872544

  ## REDHAT EXTENSION START 502604
# Disallow a bug from being put into VERIFIED state if RHEL product and not in 'qe_staff' group
# Use the privilege error code of 100 or greater for customized error messages.
  if ($field eq 'bug_status') {
    if ($product->name =~ /^Red Hat Enterprise Linux/ && $new_value eq 'VERIFIED') {
      if (not($user->in_group('qe_staff') || $user->in_group('security'))) {

        # Special case value to show a different message in the template
        push(@$priv_results, 100);
        return;
      }
    }
  }
  ## REDHAT EXTENSION END 502604

  ## REDHAT EXTENSION START 955365 1326559
  # Only selected products can be in the ON_DEV status
  if ($field eq 'bug_status' and $new_value eq 'ON_DEV') {
    unless ($product->allows_on_dev) {
      push(@$priv_results, PRIVILEGES_REQUIRED_EMPOWERED);
      return;
    }
  }
  ## REDHAT EXTENSION END 955365

  ## REDHAT EXTENSION BEGIN 707252
  # If this is a fedora bug, allow anyone to update the 'blocks' field
  if ( $user->id
    && $args->{'field'} eq 'blocked'
    && $args->{'bug'}->product eq 'Fedora')
  {
    push(@$priv_results, PRIVILEGES_REQUIRED_NONE);
    return;
  }
  ## REDHAT EXTENSION END 707252

  ## REDHAT EXTENSION BEGIN 1698290
  ## For these two fields we have to validate the supplied args.
  ## Existing args will also be set but have to be ignored.
  if ($field eq 'cf_partner' || $field eq 'cf_verified') {
    my $OK = 1;
    unless ($user->is_insider()) {
      my $cgi      = Bugzilla->cgi;
      my @supplied = $cgi->param($field);

      foreach my $value (@supplied) {
        $value = lc($value);
        my $group = Bugzilla::Group->new({name => $value});
        if ($group && !$user->in_group($value)) {
          $OK = 0;
          last;
        }
      }
    }

    push(@$priv_results, PRIVILEGES_REQUIRED_NONE) if ($OK);
  }
  ## REDHAT EXTENSION END 1698290

  return;
}

sub bug_end_of_create {
  my ($self, $args) = @_;

  my $bug       = $args->{'bug'};
  my $timestamp = $args->{'timestamp'};

  # What groups is this bug in
  my $groups = $bug->groups_in;

  ## REDHAT EXTENSION BEGIN 815549
  # If the bug is created with the security group, add the CC: and keyword
  my $security_bug = (grep { $_->name eq 'security' } @$groups) ? 1 : 0;
  if ($security_bug) {
    $bug->add_cc('security-response-team@redhat.com');
    $bug->modify_keywords('Security', 'add');
    $bug->update();
  }
  ## REDHAT EXTENSION END 815549

  ## REDHAT EXTENSION START 798022
  $self->_changes_for_ecs($args);
  ## REDHAT EXTENSION END 798022

  ## REDHAT EXTENSION START 829154
  # If the product must have a group and one is not given, that is an error
  if ($bug->product_obj->always_private and scalar(@{$bug->groups_in}) == 0) {
    ThrowUserError('group_not_supplied', {bug => $bug});
  }
  ## REDHAT EXTENSION END 829154

  ## REDHAT EXTENSION BEGIN 855475
  $self->_restrict_sensitive_bugs($args, 0);
  ## REDHAT EXTENSION END 855475

  return;
}

sub _non_insider_groups {

  # TODO horribly postgresql specific
  my $excluded_categories = "'"
    . join(q{', '},
    split(/,\s*/, Bugzilla->params->{'noninsider_catagories_exclude'}))
    . "'";

  unless (Bugzilla->request_cache->{_non_insider_groups}) {
    my $SQL = <<EOS;
SELECT id FROM groups
WHERE id NOT IN (
  WITH RECURSIVE ancestors AS (
    SELECT member_id
      FROM group_group_map
    WHERE grantor_id = (SELECT id FROM groups WHERE name = ?)
    UNION
      SELECT e.member_id
      FROM   group_group_map e
      INNER JOIN ancestors s ON s.member_id = e.grantor_id
  ) SELECT * FROM ancestors
)
AND category NOT IN ($excluded_categories)
AND name <> ?
EOS

    my $dbh = Bugzilla->dbh;

    my $group_ids = $dbh->selectcol_arrayref(
      $SQL, undef,
      Bugzilla->params->{'insidergroup'},
      Bugzilla->params->{'insidergroup'}
    );

    Bugzilla->request_cache->{_non_insider_groups}
      = Bugzilla::Group->new_from_list($group_ids);
  }

  return Bugzilla->request_cache->{_non_insider_groups};
}

sub bug_before_set_all {
  my ($self,   $args) = @_;
  my ($params, $bug)  = @$args{qw(params bug)};

  my $input = Bugzilla->input_params;
  my $user  = Bugzilla->user;

  ## REDHAT EXTENSION START 406141 439413
  # Disallow certain product changes FROM a product if not a Red Hatter
  if (  $params->{product}
    and $params->{product} ne $bug->product
    and not Bugzilla->user->is_insider())
  {
    if ( $bug->{product} =~ /Red Hat Enterprise Linux/
      or $bug->{product} eq 'Red Hat Cluster Suite'
      or $bug->{product} eq 'Red Hat Hardware Certification')
    {
      ThrowUserError("product_change_denied");
    }
  }
  ## REDHAT EXTENSION END 406141 439413

  ## REDHAT EXTENSION START 798022
  # If the 'Doc Text' (cf_release_notes) field is the default for the
  # type (whitespace excepted), then pretend it doesn't exist

  # If called from process_bug.cgi, cf_release_notes is an array
  # From RPC, it is a string
  my $doc_type = $params->{cf_doc_type};
  my $doc_text = $params->{cf_release_notes};
  $doc_type = $doc_type->[0] if ref($doc_type) eq 'ARRAY';
  $doc_text = $doc_text->[0] if ref($doc_text) eq 'ARRAY';

  if ($doc_type and length($doc_text)) {
    my $doc_type_text
      = Bugzilla::Extension::RedHat::DocTypeText->new({name => $doc_type});
    if (_nows($doc_text) eq _nows($doc_type_text->{value})) {
      delete $params->{cf_release_notes};
    }
  }
  ## REDHAT EXTENSION END 798022

  ## REDHAT EXTENSION START 1824881
  if ($user->is_insider) {
    foreach my $key (keys %$input) {
      if ($key =~ /defined_extra_private_groups_(\d+)$/) {
        my $comment_id = $1;
        my $com_groups = [];
        if (exists $input->{"extra_private_groups_$comment_id"}) {
          $com_groups = $input->{"extra_private_groups_$comment_id"};
          $com_groups = [$com_groups] if (ref($com_groups) ne 'ARRAY');
        }
        $bug->edit_private_groups($comment_id, $com_groups);
      }
    }
  }
  ## REDHAT EXTENSION END 1824881

  return;
}

sub _comment_set_private_groups {
  my ($self, $groups) = @_;

  my $dbh = Bugzilla->dbh;
  my (@added, @removed, @new_groups, @cur_groups);

  @new_groups = map { Bugzilla::Group->check({name => $_})->id } @$groups
    if (@$groups);

  my $pcg = $self->private_groups;

  @cur_groups = map { $_->id } @{$pcg} if (@{$pcg});

  my $lc  = List::Compare->new(\@cur_groups, \@new_groups);
  my @rem = $lc->get_unique;
  my @add = $lc->get_complement;

  if (@add) {
    foreach my $gid (@add) {
      my $group = Bugzilla::Group->new($gid);
      $dbh->do(
        "INSERT INTO comment_private_groups (comment_id, group_id) VALUES (?,?)",
        undef, $self->id, $group->id);
      push(@added, $group->name);
    }
  }

  if (@rem) {
    foreach my $gid (@rem) {
      my $group = Bugzilla::Group->new($gid);
      $dbh->do(
        "DELETE FROM comment_private_groups WHERE comment_id = ? AND group_id = ?",
        undef, ($self->id, $group->id));
      push(@removed, $group->name);
    }
  }

  return (\@added, \@removed);
}

sub object_end_of_create {
  my ($self, $args) = @_;

  my $class  = $args->{'class'};
  my $object = $args->{'object'};

  if ($class->isa('Bugzilla::Comment')) {
    my $user = Bugzilla->user;
    if ($user->is_insider) {
      my $input = Bugzilla->input_params;
      if (exists $input->{'extra_private_groups_new'}) {
        my $new_com_groups = $input->{'extra_private_groups_new'};
        $new_com_groups = [$new_com_groups] if (ref($new_com_groups) ne 'ARRAY');
        $object->set_private_groups($new_com_groups);
      }
    }
  }

  return;
}

## REDHAT EXTENSION BEGIN 855475
# This sub should only be called if the bug is new,
# or the product or component has changed.
# If you update the product or components in this sub, you must also update
# the watchers extension too (search for 855475)
sub _restrict_sensitive_bugs {
  my ($self, $args, $log_change) = @_;

  my $bug       = $args->{bug};
  my $timestamp = $args->{timestamp};

  # If the bug is already private, do nothing.
  return if (scalar(@{$bug->groups_in}) > 0);

  my @comps;

  # This code only applies to RHEL 5/6/7 bugs.
  @comps = qw(kernel distribution kernel-pegas kernel-alt)
    if ($bug->product =~ /^Red Hat Enterprise Linux (?:5|6|7|8)$/);

  ## RED HAT EXTENSION START 1451244
  my @entitlements_prods = (
    'Red Hat Enterprise Linux 6',
    'Red Hat Enterprise Linux 7',
    'Red Hat Enterprise Linux 8',
    'Red Hat Enterprise Virtualization Manager',
    'Red Hat OpenStack',
    'OpenShift Container Platform',
    'Red Hat Satellite 5',
    'Red Hat Satellite 6',
    'Red Hat Certification Program',
    'Red Hat Certificate System',
    'Red Hat Directory Server',
    'Red Hat Gluster Storage',
    'Red Hat Ceph Storage',
    'Red Hat Quickstart Cloud Installer'
  );

  push(@comps, 'entitlements')
    if (any { $_ eq $bug->product } @entitlements_prods);
  ## RED HAT EXTENSION END 1451244

  return unless @comps;

  # Check to see if the bug has any components which are required to be private.
  my $bug_components = $bug->all_components;
  foreach my $private_component (@comps) {
    if (any { $private_component eq $_ } @$bug_components) {

      # Add the bug to the private group
      my $group = Bugzilla::Group->new({name => Bugzilla->params->{'insidergroup'}});

      my $sth = Bugzilla->dbh->prepare(
        'INSERT INTO bug_group_map (bug_id, group_id) VALUES (?,?)');
      $sth->execute($bug->id, $group->id);

      # If we are doing an update, we need to log the change
      if ($log_change) {

        # We will record this change as been done by bugzilla@redhat.com
        # If we cannot find this user, use the user that did this change
        # as a backup
        my $user_id = Bugzilla::User::login_to_id(Bugzilla->params->{system_user});
        $user_id ||= Bugzilla->user->id;

        LogActivityEntry($bug->id, 'bug_group', '', Bugzilla->params->{'insidergroup'},
          $user_id, $timestamp);
      }
      delete $bug->{groups_in};
      return;
    }
  }

  return;
}
## REDHAT EXTENSION END 855475

sub bug_start_of_update {
  my ($self, $args) = @_;
  my $bug      = $args->{bug};
  my $old_bug  = $args->{old_bug};
  my $changes  = $args->{changes};
  my $delta_ts = $args->{timestamp};
  my $dbh      = Bugzilla->dbh;

  ## RED HAT EXTENSION START 1174396
  if ($bug->product_obj->name eq 'Fedora') {
    if ( $old_bug->status->name ne $bug->status->name
      && $old_bug->status->name eq 'CLOSED'
      && $old_bug->resolution eq 'EOL')
    {
      my $version = Bugzilla::Version->check(
        {name => $bug->version, product => $bug->product_obj});

      ThrowUserError('inactive_version', {bug => $bug}) unless ($version->is_active);
    }

    if ( $old_bug->status->name ne $bug->status->name
      && $bug->bug_status eq 'CLOSED'
      && $bug->resolution eq 'EOL')
    {
      Bugzilla->user->in_group('fedora_pm')
        || ThrowUserError("resolution_denied",
        {bug => $bug, group => "fedora_pm", resolution => $bug->resolution});
    }
  }
  ## RED HAT EXTENSION END 1174396

  ## REDHAT EXTENSION START 470051 903197
  # If this bug has a blocker and is a duplicate, then the duplicate bug
  # should have the same blocker(s) added
  if (
    $bug->dup_id
    && ( $old_bug->status->name ne $bug->status->name
      || $old_bug->resolution ne $bug->resolution
      || $old_bug->dup_id ne $bug->dup_id)
    )
  {
    my $dup_id  = $bug->dup_id;
    my $dup_bug = $bug->new($dup_id);

    my ($removed, $added) = diff_arrays($dup_bug->blocked, $bug->blocked);

    # We cannot add a blocker the user cannot see
    $added = Bugzilla->user->visible_bugs($added);

    # Neither can we block the bug against itself, or a bug in its
    # dependency tree. Use SQL to avoid permission issues (plus it
    # is quicker)
    my %cant_block = ($dup_id => 1);
    my %to_check   = ($dup_id => 1);

    my $query = "SELECT dependson FROM dependencies WHERE blocked = ?";

    while (scalar keys %to_check) {
      foreach my $block_bug_id (keys %to_check) {

        # We don't need to check this one again
        delete $to_check{$block_bug_id};

        # Get a list of bugs that depend on this one
        my $dep_bug_ids = $dbh->selectcol_arrayref($query, undef, $block_bug_id);

        foreach my $dep_bug_id (@$dep_bug_ids) {

          # If we haven't seen this bug before, add it to the list
          # of bugs we cannot block AND check for this bugs
          # dependencies
          if (none { $_ == $dep_bug_id } keys %cant_block) {
            $to_check{$dep_bug_id}   = 1;
            $cant_block{$dep_bug_id} = 1;
          }
        }
      }
    }

    # Remove bugs from the dependency tree
    @$added = grep { !exists $cant_block{$_} } @$added;

    if (scalar(@$added) and $dup_bug->check_can_change_field('blocked', 0, 1)) {
      my @dependson = @{$dup_bug->dependson};
      my @blocked   = (@{$dup_bug->blocked}, @$added);
      $dup_bug->set_dependencies(\@dependson, \@blocked);

      ## REDHAT EXTENSION START 861276
      $dup_bug->{_skip_extra_change} = 1;
      ## REDHAT EXTENSION END 861276

      my $dup_changes = $dup_bug->update();
      $bug->{_dup_blocked_changes} = $dup_changes;
    }
  }
  ## REDHAT EXTENSION END 470051 903197

  ## REDHAT EXTENSION BEGIN 212492
  $self->_check_closed_duplicate($args);
  ## REDHAT EXTENSION END 212492

  ## REDHAT EXTENSION START 829154
  # If the product must have a group and one is not given, that is an error
  if ($bug->product_obj->always_private and scalar(@{$bug->groups_in}) == 0) {
    ThrowUserError('group_not_supplied', {bug => $bug});
  }
  ## REDHAT EXTENSION END 829154

  ## REDHAT EXTENSION BEGIN 554957
  # Get the users merge preferences settings.
  my $merge_behaviour = Bugzilla->user->setting('merge_bug_data');

  if ( (my $target_id = $bug->dup_id)
    && defined($bug->{_dup_for_update})
    && ($merge_behaviour ne 'Never'))
  {
    my $target_bug = $bug->{_dup_for_update};

    my $source_id  = $bug->id;
    my $source_bug = $bug;

    my $source_fields = _select_fields($source_bug);
    my $target_fields = _select_fields($target_bug);

    my %merged_fields;

    _merge_fields(\%merged_fields, $source_bug, $target_bug, $source_fields,
      $target_fields);
    _merge_flags(\%merged_fields, $source_bug, $target_bug, $source_fields,
      $target_fields);

    # Is there any data that even needs to be merged?
    my $needs_merge = scalar keys %merged_fields;

    my $perform_merge = Bugzilla->cgi->param('merge');
    my $is_browser    = (Bugzilla->usage_mode == USAGE_MODE_BROWSER);

    # If the user has configured to always merge, or has confirmed the prompt
    if (
      (
        defined($perform_merge) && ($perform_merge eq 'Merge changes') && $needs_merge
      )
      || ($merge_behaviour eq 'Always' && $needs_merge)
      )
    {
      # User has clicked the 'Merge changes' button, or are always merging changes.
      # Apply the merge
      _apply_new_fields($target_bug, \%merged_fields);
      _apply_new_flags($target_bug, \%merged_fields);
    }

    # If the user wants to be prompted to merge changes.
    elsif (($merge_behaviour eq 'Prompt')
      && $is_browser
      && $needs_merge
      && (!defined($perform_merge) || $perform_merge ne 'Do not merge changes'))
    {
      # Show the changes that will be made and ask for confirmation.
      _display_merge_warning_changes($source_id, $target_id, $source_fields,
        $target_fields, \%merged_fields);
    }
  }
  ## REDHAT EXTENSION END 554957

  ## REDHAT EXTENSION START 982140 1000549
  if (  ($changes->{bug_status} or $changes->{target_milestone})
    and $bug->bug_status eq 'ON_QA'
    and $bug->cf_type eq 'Bug'
    and $bug->classification eq 'JBoss'
    and ($bug->target_milestone eq '---' or $bug->target_milestone eq ''))
  {
    # Check that there is more than one milestone available
    my $milestones = $bug->product_obj->milestones;
    if (any { $_->name ne '---' } @$milestones) {
      ThrowUserError('jboss_no_milestone');
    }
  }
  ## REDHAT EXTENSION END 982140 1000549


  ## REDHAT EXTENSION START 1306103
  # If the account for the  new assignee for the assigned_to,
  # qa_contact or docs_contact field is disabled,
  # disallow the update.
  #
  # Exception: Allow the update if the assignee is the default for this field.

  if (defined($changes->{assigned_to})) {
    my $default_user = $bug->component_obj->default_assignee;
    if (@{$bug->rh_sub_component_objs}) {
      $default_user = $bug->rh_sub_component_objs->[0]->default_assignee;
    }
    $self->_check_account_enabled($changes->{assigned_to}, $bug, $default_user);
  }
  if (defined($changes->{qa_contact})) {
    my $default_user = $bug->component_obj->default_qa_contact;
    if (@{$bug->rh_sub_component_objs}) {
      $default_user = $bug->rh_sub_component_objs->[0]->default_qa_contact;
    }
    $self->_check_account_enabled($changes->{qa_contact}, $bug, $default_user);
  }
  if (defined($changes->{docs_contact})) {
    my $default_user = $bug->component_obj->default_docs_contact;
    if (@{$bug->rh_sub_component_objs}) {
      $default_user = $bug->rh_sub_component_objs->[0]->default_docs_contact;
    }
    $self->_check_account_enabled($changes->{docs_contact}, $bug, $default_user);
  }
  ## REDHAT EXTENSION END 1306103

  ## REDHAT EXTENSION START 420461
  my $edited_comments = 0;
  foreach my $comment (@{$bug->{edited_comments} || []}) {
    $edited_comments = 1;
    $changes->{'longdesc'} = [$comment->{count}, 'updated'];
    $bug->update_comment($comment->{id}, $comment->{comment}, 1, $delta_ts,
      $comment->{security});
  }
  delete $bug->{edited_comments};    # delete in case $bug->update is called again.
  ## REDHAT EXTENSION START 420461

  ## REDHAT EXTENSION START 1824881
  my $edited_comment_private_groups = 0;
  foreach my $comment (@{$bug->{edited_comment_private_groups} || []}) {
    $changes->{'longdesc'} = [$comment->{count}, 'comment private groups updated'];
  }
  delete $bug->{edited_comment_private_groups};    # delete in case $bug->update is called again.
  ## REDHAT EXTENSION START 1824881

  return;
}

## REDHAT EXTENSION START 1306103
sub _check_account_enabled {
  my ($self, $changes, $bug, $default_user) = @_;

  my (undef, $to) = @$changes;
  return unless defined $to;                       # Assignee is being set to nothing.

  my $to_user = Bugzilla::User->new({id => $to});

  unless (
       $to_user->is_enabled
    || $to_user->login eq 'nobody@redhat.com'
    || (defined($default_user) && ($to_user->login eq $default_user->login))
    || ( $to_user->login =~ /\@redhat.com$/
      && $to_user->email_enabled()
      && _validate_redhat_addr($to_user->login))
    )
  {
    ThrowUserError('cant_assign_disabled_user', {to => $to_user});
  }

  return;
}
## REDHAT EXTENSION END 1306103


sub bug_end_of_update {
  my ($self, $args) = @_;
  my $dbh  = Bugzilla->dbh;
  my $user = Bugzilla->user;

  my ($bug, $old_bug, $timestamp, $changes)
    = @$args{qw(bug old_bug timestamp changes)};

  ## REDHAT EXTENSION BEGIN 815549 1057613
  # If the bug has the security group, add the CC: and keyword if needed
  if (any { $_->name eq 'security' } @{$bug->groups_in}) {

    # We will record this change as been done by bugzilla@redhat.com
    # If we cannot find this user, use the user that did this change
    # as a backup
    my $user_id = Bugzilla::User::login_to_id(Bugzilla->params->{system_user});
    $user_id ||= $user->id;
    unless (any { $_->name eq 'Security' } @{$bug->keyword_objects}) {

      # Don't do anything if the Security keyword doesn't exist.
      my $keyword = Bugzilla::Keyword->new({name => 'Security'});
      if ($keyword) {
        $dbh->do('INSERT INTO keywords (bug_id, keywordid) VALUES (?,?)',
          undef, $bug->id, $keyword->id);

        if ($changes->{'keywords'}) {
          $changes->{'keywords'}->[1] .= ', Security';
        }
        else {
          $changes->{'keywords'} = ['', 'Security'];
        }
        LogActivityEntry($bug->id, 'keywords', '', 'Security', $user_id, $timestamp);
      }
    }

    unless (
      (!$bug->cc || any { $_ eq 'security-response-team@redhat.com' } @{$bug->cc})
      || (!$bug->cc_users
        || any { $_->login eq 'security-response-team@redhat.com' } @{$bug->cc_users})
      )
    {
      # Don't do anything if the User doesn't exist
      my $srt_user
        = Bugzilla::User->new({name => 'security-response-team@redhat.com'});
      if ($srt_user) {
        $dbh->do('INSERT INTO cc (bug_id, who) VALUES (?,?)',
          undef, $bug->id, $srt_user->id);

        if ($changes->{'cc'}) {
          $changes->{'cc'}->[1] .= ', security-response-team@redhat.com';
        }
        else {
          $changes->{'cc'} = ['', 'security-response-team@redhat.com'];
        }
        LogActivityEntry($bug->id, 'cc', '', 'security-response-team@redhat.com',
          $user_id, $timestamp);
      }
    }
  }
  ## REDHAT EXTENSION END 815549 1057613

  ## REDHAT EXTENSION START 1005702
  # We have to do this at the end because changes->target_release is
  # set by update itself
  if (  ($changes->{bug_status} or $changes->{target_release})
    and $bug->bug_status eq 'ON_QA'
    and $bug->product eq 'RHQ Project'
    and (none { $_ ne '---' } @{$bug->target_release}))
  {
    # Check that there is more than one release available
    my $releases = $bug->product_obj->releases;
    if (any { $_->name ne '---' } @$releases) {
      ThrowUserError('rhq_no_release');
    }
  }
  ## REDHAT EXTENSION END 1005702

  ## REDHAT EXTENSION START 581952
  if ($changes->{'bug_status'}) {

    # If the bug has been closed then update the cf_last_closed
    # value to the current timestamp if cf_last_closed exists
    if (!$bug->status->is_open && Bugzilla::Field->new({name => 'cf_last_closed'}))
    {
      my $current_last_closed = $bug->cf_last_closed;
      $current_last_closed = '' if not defined $current_last_closed;

      $dbh->do("UPDATE bugs SET cf_last_closed = ? WHERE bug_id = ?",
        undef, $timestamp, $bug->id);
      LogActivityEntry($bug->id, 'cf_last_closed', $current_last_closed, $timestamp,
        $user->id, $timestamp);
    }
  }
  ## REDHAT EXTENSION END 581952

  ## REDHAT EXTENSION START 433348
  # Add Reopened keyword to bug if the old status was closed and new status isn't
  if ($changes->{'bug_status'}) {
    if (!$old_bug->status->is_open
      && $bug->status->is_open
      && none { $_->name eq 'Reopened' } @{$bug->keyword_objects})
    {
      # Don't do anything if the Reopened keyword doesn't exist.
      my $keyword = Bugzilla::Keyword->new({name => 'Reopened'});
      if ($keyword) {
        $dbh->do('INSERT INTO keywords (bug_id, keywordid) VALUES (?,?)',
          undef, $bug->id, $keyword->id);

        if ($changes->{'keywords'}) {

          # Hack, we have to alter the current keywords activity entry to
          # add the Reopened keyword since other changes were made.
          my $field = Bugzilla::Field->new({name => 'keywords'});
          my $new_added
            = join(", ", (split(/[,\s]/, $changes->{'keywords'}->[1]), 'Reopened'));
          $changes->{'keywords'}->[1] = $new_added;
          $dbh->do(
            "UPDATE bugs_activity SET added = ? WHERE bug_id = ?
                              AND who = ? AND fieldid = ? and bug_when = ?", undef,
            $new_added, $bug->id, $user->id, $field->id, $timestamp
          );
        }
        else {
          $changes->{'keywords'} = ['', 'Reopened'];
          LogActivityEntry($bug->id, 'keywords', '', 'Reopened', $user->id, $timestamp);
        }
      }
    }
  }
  ## REDHAT EXTENSION END 433348

  ## REDHAT EXTENSION BEGIN 488370
  # If the component has changed, and any of the CDW ack flags is '+' then set it
  # to '?' . Except if the old component is distribution
  if (
    $changes->{component}
    && !(
      $bug->product_obj->allows_no_clear_acks
      or any { $_ eq 'distribution' } @{$old_bug->all_components}
    )
    )
  {

    # Get a list of qa_ack+, pm_ack+ or dev_ack+ flags for this bug
    # We use SQL to get around permission restrictions
    my $query = "
            SELECT f.id AS flag_id, t.name AS flag_type_name
            FROM flags f
                JOIN flagtypes t ON f.type_id = t.id
            WHERE t.name IN ('qa_ack', 'pm_ack', 'devel_ack')
                AND f.status = '+'
                AND f.bug_id = ?
        ";

    my $rows = $dbh->selectall_arrayref($query, undef, $bug->id);

    foreach my $row (@$rows) {
      my $flag_id        = $row->[0];
      my $flag_type_name = $row->[1];

      # Make sure the change was not part of this request
      if (defined $changes->{'flagtypes.name'}) {
        if (index($changes->{'flagtypes.name'}[1], $flag_type_name . '+') != -1) {

          # The flag was intentionally changed as part of this
          # request, therefore we shouldn't change it
          next;
        }
      }

      # We need to change this flag from + to ?
      # As above, use SQL to get around permission issues
      $dbh->do("UPDATE flags SET status = ? WHERE id = ?", undef, '?', $flag_id);

      # We will record this change as been done by bugzilla@redhat.com
      # If we cannot find this user, use the user that did this change
      # as a backup
      my $user_id = Bugzilla::User::login_to_id(Bugzilla->params->{system_user});
      $user_id ||= $user->id;

      # Log the change
      LogActivityEntry(
        $bug->id,           'flagtypes.name',
        "$flag_type_name+", "$flag_type_name?",
        $user_id,           $timestamp,
        undef,              undef,
        $bug,               $old_bug
      );
    }
  }
  ## REDHAT EXTENSION END 488370

  ## REDHAT EXTENSION START 565827 826305
  if ($changes->{'bug_status'}
    and
    (any { $_ eq $old_bug->status->name } ('ON_QA', 'VERIFIED', 'RELEASE_PENDING'))
    and $bug->status->name eq 'ASSIGNED'
    and none { $_ eq 'FailedQA' } @{$bug->cf_verified})
  {
    # Can the user set the 'FailedQA' value
    # not defined = no restrictions, scalar = some values
    my $field = Bugzilla::Field->new({name => 'cf_verified'});
    if ($field && any { $_->name eq 'FailedQA' } @{$field->legal_values}) {
      $dbh->do("INSERT INTO bug_cf_verified (bug_id, value) VALUES (?, ?)",
        undef, $bug->id, 'FailedQA');

      # Log the change
      LogActivityEntry($bug->id, 'cf_verified', '', 'FailedQA', $user->id,
        $timestamp);
    }
  }
  ## REDHAT EXTENSION END 565827 826305

  ## REDHAT EXTENSION START 798022
  $self->_changes_for_ecs($args);
  ## REDHAT EXTENSION END 798022

  ## REDHAT EXTENSION BEGIN 855475
  # Only make a change if product or component changes.
  if ($changes->{product} || $changes->{component}) {
    $self->_restrict_sensitive_bugs($args, 1);
  }
  ## REDHAT EXTENSION END 855475

  return;
}

## REDHAT EXTENSION START 798022
## Remove all whitespace from a string
sub _nows {
  my $text = shift;
  return '' if not defined $text;
  $text =~ s/\s//g;
  return $text;
}

sub _changes_for_ecs {
  my ($self, $args) = @_;
  my $bug       = $args->{bug};
  my $timestamp = $args->{timestamp};
  my $old_bug   = $args->{old_bug};     # undef if called from bug_end_of_create

  my $dbh  = Bugzilla->dbh;
  my $user = Bugzilla->user;
  my $doc_type_text_new
    = Bugzilla::Extension::RedHat::DocTypeText->new({name => $bug->cf_doc_type});
  my $doc_type_text_old = undef;
  $doc_type_text_old
    = Bugzilla::Extension::RedHat::DocTypeText->new(
    {name => $old_bug->cf_doc_type})
    if (defined($old_bug) && $old_bug->cf_doc_type);

  # Holds what changes we will make. The changes are made at the end. Makes
  # the code easier to manage. Keys are flag, cf_doc_type and cf_release_notes
  my %changes = ();

  # Determine if the text has changed. It may match the new one if the
  # user has Javascript or the old one if they have no Javascript, or are
  # making an RPC call
  my $default_text = $bug->cf_release_notes =~ /^\s*$/ ? 1 : 0;
  if (defined($doc_type_text_new)
    && (_nows($bug->cf_release_notes) eq _nows($doc_type_text_new->value)))
  {
    $default_text = 1;
  }
  if (defined($doc_type_text_old)
    && (_nows($bug->cf_release_notes) eq _nows($doc_type_text_old->value)))
  {
    $default_text = 1;
  }

  # Bug 1167890 - ECS flag change logic.
  my ($ecs_flag) = grep { $_->name eq 'requires_doc_text' } @{$bug->flags};

  # We need to be careful since the logged in user might not have
  # the necessary privileges to see this flag or the flag might not be visible
  # on the bug's product
  my $flags = Bugzilla::FlagType::match(
    {name => 'requires_doc_text', product_id => $bug->product_id});

  my $flag_type = @$flags ? $flags->[0] : undef;

  # Bail out if this flag doesn't exist on the bug's product.
  return () unless ($flag_type);

  my $flag;
  if ($bug && $flag_type) {
    $flag = Bugzilla::Flag->new({
      condition => 'bug_id = ? AND type_id = ?',
      values    => [$bug->id, $flag_type->id]
    });
  }
  my $desired_flag_status = defined $ecs_flag ? $ecs_flag->status : '';
  my $current_flag_status = defined $flag     ? $flag->status     : '';

  # Because the flag change happens before this code is run you need to check
  # the changes to see if it's been changed already
  if (exists $args->{changes}->{'flagtypes.name'}) {
    my @flags_rem   = split(/,\s*/, $args->{changes}->{'flagtypes.name'}->[0]);
    my @flags_added = split(/,\s*/, $args->{changes}->{'flagtypes.name'}->[1]);

    foreach my $change (@flags_added) {
      if ($change =~ /^requires_doc_text([\?\+\-])$/) {
        $desired_flag_status = $1;
        last;
      }
    }

    foreach my $change (@flags_rem) {
      if ($change =~ /^requires_doc_text([\?\+\-])$/) {
        $current_flag_status = $1;
        last;
      }
    }
  }

  my $oldflag;
  my $new_type = $bug->cf_doc_type;
  my $old_type;
  if ($old_bug) {
    ($oldflag) = grep { $_->name eq 'requires_doc_text' } @{$old_bug->flags};
    $old_type = $old_bug->cf_doc_type;
  }
  my $old_flag_status = defined $oldflag ? $oldflag->status : '';

  # RHBZ 1717818 Use current value if old value isn't visible???
  $old_flag_status = $current_flag_status
    if ($old_bug && !$user->can_see_flag($flag_type));

  # If we are using the default text, determine whether the doc type (and
  # therefore doc text) should change. The rules of what the type is are
  # defined in bug 798022 (See 'Drop-Down List Details' in the PDF)
  if ($default_text && !defined($old_bug)) {

    # Get all the keywords (if any) for this bug
    my @keywords = map { $_->name } @{$bug->keyword_objects};
    my $type     = '';

    if (any { $_ eq 'SecurityTracking' } @keywords) {
      $changes{flag} = '-';
      $type = 'No Doc Update';
    }
    elsif (any { $_ eq 'TechPreview' } @keywords) {
      $type = 'Technology Preview';
    }
    elsif (any { $_ eq 'Rebase' } @keywords) {
      $type = 'Rebase: Bug Fixes and Enhancements';
    }
    elsif (any {/^(FutureFeature|Improvement|SubFeature)$/} @keywords) {
      $type = 'Enhancement';
    }
    else {
      $type = 'If docs needed, set a value';
    }

    if ($type ne $bug->cf_doc_type) {
      $changes{cf_doc_type}      = $type;
      $changes{cf_release_notes} = '';
    }
  }
  elsif (
    (not defined $old_bug) or    # If it is a new bug not $default_text, we know it has been changed
    (_nows($old_bug->cf_release_notes) ne _nows($bug->cf_release_notes))
    or ($old_bug->cf_doc_type ne $bug->cf_doc_type)
    or ($current_flag_status ne $old_flag_status)
    )
  {
    # Not the default text, but the text has changed. Lets see if the
    # status of the requires_doc_text flag needs to change (See 'Automatic
    # State Transitions of the requires_doc_text Flag' in the PDF)

    my $user_is_ecs     = $user->in_group('ecs');
    my $new_flag_status = '?';

    # If NACKing or don't want docs update, make sure all fields agree
    if (
      (
        ($old_flag_status ne $current_flag_status) && ($desired_flag_status eq '-')
      )
      || ($new_type ne $old_type && $new_type eq 'No Doc Update')
      )
    {
      $new_flag_status           = '-';
      $changes{cf_doc_type}      = 'No Doc Update' if ($old_type ne 'No Doc Update');
      $changes{cf_release_notes} = ''
        unless ($old_bug && $old_bug->cf_release_notes eq '');
    }

    # If the user is ECS and the flag was not changed, the flag to +
    elsif ($user_is_ecs && ($desired_flag_status eq $current_flag_status)) {
      $new_flag_status = '+';
    }

    # If the user is not ECS, and they want to set the flag to - or ?, let them.
    elsif (!$user_is_ecs
      && (($desired_flag_status eq '-') || $desired_flag_status eq '?'))
    {
      $new_flag_status = $desired_flag_status;
    }

    # If the text has changed, and the user is not ECS.
    elsif (!$user_is_ecs && ($current_flag_status ne '?')) {

      # Then we need to mark the doc text for review.
      $new_flag_status = '?';
    }

    # If the user is ECS, let them set it to whatever they want.
    elsif ($user_is_ecs) {
      $new_flag_status = $desired_flag_status;
    }

    if ($current_flag_status ne $new_flag_status) {
      $changes{flag} = $new_flag_status;
    }
  }

  # Now make the changes
  if (scalar keys %changes) {

    # We will record this change as been done by bugzilla@redhat.com
    # If we cannot find this user, use the user that did this change
    # as a backup
    my $bz_user_id = Bugzilla::User::login_to_id(Bugzilla->params->{system_user});
    $bz_user_id ||= $user->id;

    if (my $new_flag_status = delete $changes{flag}) {

      # We need to be careful since the logged in user might not have
      # the necessary privileges to see this flag
      my $doc_text_flag_type = Bugzilla::FlagType->new({name => 'requires_doc_text'});
      my $doc_text_flag      = Bugzilla::Flag->new({
        condition => 'bug_id = ? AND type_id = ?',
        values    => [$bug->id, $doc_text_flag_type->id]
      });

      if (defined $doc_text_flag and $doc_text_flag->id) {
        if ($doc_text_flag->status ne $new_flag_status) {

          # Only do an update if the status has changed.
          my $query
            = "UPDATE flags SET status = ?, setter_id =?, modification_date = ? WHERE id = ?";
          $dbh->do($query, undef, $new_flag_status, $bz_user_id, $timestamp,
            $doc_text_flag->id);
        }
      }
      else {
        # We need to create a new flag
        my $query = "
                    INSERT INTO flags (
                        type_id, bug_id, setter_id,
                        status, creation_date, modification_date
                    )
                    VALUES (?, ?, ?, ?, ?, ?)
                ";
        $dbh->do($query, undef, $doc_text_flag_type->id, $bug->id, $bz_user_id,
          $new_flag_status, $timestamp, $timestamp);
      }

      # Log the change, if a change was made and it isn't a new bug
      if (defined $old_bug
        and (not $doc_text_flag or $doc_text_flag->status ne $new_flag_status))
      {
        LogActivityEntry(
          $bug->id, 'flagtypes.name',
          (defined $doc_text_flag ? "requires_doc_text" . $doc_text_flag->status : ''),
          "requires_doc_text$new_flag_status",
          $bz_user_id, $timestamp, undef, undef, $bug, $old_bug
        );
      }
    }

    # Update the cf_doc_type and cf_release_notes fields (if specified)
    foreach my $field (keys %changes) {
      my $query = "UPDATE bugs SET $field = ? WHERE bug_id = ?";
      $dbh->do($query, undef, $changes{$field}, $bug->id);

      # Log the change, except if this is a new bug
      if (defined $old_bug) {
        LogActivityEntry($bug->id, $field, $bug->$field, $changes{$field}, $bz_user_id,
          $timestamp);
      }
      else {
        my $field_obj = new Bugzilla::Field({name => $field});
        $bug->set_custom_field($field_obj, $changes{$field});
      }
    }
  }

  return;
}
## REDHAT EXTENSION END 798022

sub bug_format_comment {
  my ($self, $args) = @_;

  my $regexes = $args->{'regexes'};

## REDHAT EXTENSION START 627570
  # Linkify links that look like JIRA issues
  my $replace_jira_links = sub {
    my $pre   = $_[0]->{matches}->[0];
    my $match = html_quote($_[0]->{matches}->[1]);
    return qq{$pre<a href="https://issues.jboss.org/browse/$match">$match</a>};
  };

  push(
    @$regexes,
    {
      match   => qr/(jira(?:\W*issue)?\W*)(?-i)([A-Z0-9]{2,}\-\d+)/i,
      replace => $replace_jira_links
    }
  );
## REDHAT EXTENSION END 627570

  return;
}

sub webservice {
  my ($self, $args) = @_;

  my $dispatch = $args->{dispatch};

  $dispatch->{'RedHat'}    = "Bugzilla::Extension::RedHat::WebService::Bugzilla";
  $dispatch->{'Flag'}      = "Bugzilla::Extension::RedHat::WebService::Flag";
  $dispatch->{'Component'} = "Bugzilla::Extension::RedHat::WebService::Component";
  $dispatch->{'Field'}     = "Bugzilla::Extension::RedHat::WebService::Field";

  return;
}

sub webservice_error_codes {
  my ($self, $args) = @_;

  my $error_map = $args->{error_map};
  $error_map->{jboss_no_milestone} = 123;

  return;
}

sub config {
  my ($self, $args) = @_;

  my $config = $args->{config};
  $config->{RedHat} = "Bugzilla::Extension::RedHat::Config";

  return;
}

sub config_add_panels {
  my ($self, $args) = @_;

  my $modules = $args->{panel_modules};

  return;
}

sub config_modify_panels {
  my ($self, $args) = @_;

  my $panels = $args->{panels};

  ## REDHAT EXTENSION BEGIN 406301
  # Log4perl param
  push(
    @{$panels->{admin}{params}},
    {
      name    => 'log4perl_config',
      type    => 'l',
      default => Bugzilla::Constants::LOG4PERL_DEFAULT_CONFIG,
    }
  );
  ## REDHAT EXTENSION END 406301

  push(
    @{$panels->{advanced}{params}},
    {
      name    => 'login_lockout_interval',
      type    => 't',
      default => Bugzilla::Constants::LOGIN_LOCKOUT_INTERVAL,
    },
    {
      name    => 'max_login_attempts',
      type    => 't',
      default => Bugzilla::Constants::MAX_LOGIN_ATTEMPTS,
    }
  );

  ## REDHAT EXTENSION BEGIN 711031
  # Allow the categories of flag types that are valid
  push(
    @{$panels->{advanced}{params}},
    {name => 'flag_type_categories', default => '', type => 't',}
  );
  ## REDHAT EXTENSION END 711031

  ## REDHAT EXTENSION BEGIN 751352
  # Allow the categories of flag types that are valid
  push(
    @{$panels->{advanced}{params}},
    {name => 'group_categories', default => '', type => 't',}
  );
  ## REDHAT EXTENSION END 751352

  insert_after { $_->{name} eq 'usetargetmilestone' }
  {name => 'usetargetrelease', type => 'b', default => 0} =>
    @{$panels->{bugfields}{params}};

  ## REDHAT EXTENSION START 876015
  insert_after { $_->{name} eq 'useqacontact' }
  {name => 'usedocscontact', type => 'b', default => 0} =>
    @{$panels->{bugfields}{params}};
  ## REDHAT EXTENSION END 876015

  ## REDHAT EXTENSION START 583503
  push(
    @{$panels->{core}{params}},
    {name => 'force_secure_cookies', type => 'b', default => '0',}
  );
  ## REDHAT EXTENSION END 582503

  ## REDHAT EXTENSION START 212492
  push(
    @{$panels->{bugchange}{params}},
    {name => 'duplicate_warn_override', type => 'b', default => 1}
  );
  ## REDHAT EXTENSION END 242492

  push(
    @{$panels->{core}{params}},
    {name => 'utf8',         type => 'b', default => '0', checker => \&check_utf8},
    {name => 'shutdownhtml', type => 'l', default => ''},
    {name => 'announcehtml', type => 'l', default => ''},
    {name => 'proxy_url',    type => 't', default => ''},
    {name => 'inbound_proxies',  type => 't', default => ''},
    {name => 'rpc_log_args',     type => 'b', default => '0',},
    {name => 'rh_radius_groups', type => 't', default => '',},
    {name => 'bugzilla_term',    type => 't', default => 'Red Hat Bugzilla'},
    {name => 'system_user',      type => 't', default => 'bugzilla@redhat.com'},
    {
      name    => 'favicon',
      type    => 't',
      default => 'extensions/RedHat/web/css/favicons/production.ico',
    },
## REDHAT EXTENSION START 1821023
    {name => 'nobody_group',        type => 't', default => 'nobody',},
    {name => 'nobody_user',         type => 't', default => 'nobody@example.com',},
    {name => 'spam_classification', type => 't', default => 'Retired',},
    {name => 'spam_component',      type => 't', default => 'Can',},
    {name => 'spam_product',        type => 't', default => 'Spam Sinkhole',},
## REDHAT EXTENSION END 1821023
    {
      name    => 'noninsider_catagories_exclude',
      type    => 't',
      default => q{Admin, Deprecated, Red Hat},
    },
  );


  ## REDHAT EXTENSION START 590893
  push(
    @{$panels->{query}{params}},
    {name => 'long_query_timeout', type => 't', default => '0',},
    ## REDHAT EXTENSION END 590893
    ## REDHAT EXTENSION START 523845
    {
      name => 'expired_search_time',
      type => 's',
      choices =>
        ['1 Month', '3 Months', '6 Months', '9 Months', '12 Months', 'Never'],
      default => 'Never'
    }
    ## REDHAT EXTENSION END 523845
  );

  ## REDHAT EXTENSION START 821671
  push(
    @{$panels->{shadowdb}{params}},
    {name => 'second_shadowdbhost', type => 't', default => '',},

    {
      name    => 'second_shadowdbport',
      type    => 't',
      default => '3306',
      checker => \&check_numeric,
    },

    {name => 'second_shadowdbsock', type => 't', default => '',},

    # This entry must be _after_ the shadowdb{host,port,sock} settings so that
    # they can be used in the validation here
    {
      name    => 'second_shadowdb',
      type    => 't',
      default => '',
      checker => \&check_second_shadowdb
    }
  );
  ## REDHAT EXTENSION END 821671

  ## REDHAT EXTENSION START 840329
  push(
    @{$panels->{general}{params}},
    {
      name     => 'mail_errors_to',
      type     => 't',
      no_reset => '1',
      default  => '',
      checker  => \&check_email
    }
  );
  ## REDHAT EXTENSION END 840329

  ## REDHAT EXTENSION START 1490644
  push(
    @{$panels->{ldap}{params}},
    {
      name     => 'LDAPMXserver',
      type     => 't',
      no_reset => '1',
      default  => 'ldap.corp.redhat.com',
    },
    {
      name     => 'LDAPMXBaseDN',
      type     => 't',
      no_reset => '1',
      default  => 'uid=bugzilla,ou=serviceaccounts,dc=redhat,dc=com',
    },
    {
      name     => 'LDAPMXSearchDN',
      type     => 't',
      no_reset => '1',
      default  => 'ou=mx,dc=redhat,dc=com',
    },
    {
      name     => 'LDAPMXmailattribute',
      type     => 't',
      no_reset => '1',
      default  => 'sendmailMTAAliasValue',
    },
  );
  ## REDHAT EXTENSION END 1490644

  return;
}

sub localconfig_symbols {
  my ($self, $args) = @_;
  my $read_symbols       = $args->{read_symbols};
  my @local_read_symbols = map($_->{name}, LOCALCONFIG_VARS);
  push(@$read_symbols, @local_read_symbols);
  return;
}

sub localconfig_vars {
  my ($self, $args) = @_;
  my $localconfig_vars = $args->{localconfig_vars};
  push(@$localconfig_vars, LOCALCONFIG_VARS);
  return;
}

sub bug_url_sub_classes {
  my ($self, $args) = @_;
  my $sub_classes = $args->{sub_classes};

  ## REDHAT EXTENSION START 915533
  push @$sub_classes, 'Bugzilla::Extension::RedHat::BugUrl::SFDC';
  ## REDHAT EXTENSION END 915533

  return;
}

sub search_operator_field_override {
  my ($self, $args) = @_;

  my $operators = $args->{'operators'};

  $operators->{alias} = {
    _non_changed => sub { _alias_nonchanged(@_) }
  };
  $operators->{target_release} = {
    _non_changed => sub { _target_release_nonchanged(@_) }
  };

  return;
}

sub webservice_before_call {
  my ($self, $args) = @_;

  ## REDHAT EXTENSION START 1090710
# We want to log: user id, user e-mail, XML|JSON, method called, COOKIE|TOKEN (or BOTH|NONE)
# Get user id and e-mail
  my ($user_id, $user_email)
    = Bugzilla->user->id ? (Bugzilla->user->id, Bugzilla->user->name) : (0, '');

  # Get connection method
  my $script
    = basename($0) eq 'xmlrpc.cgi'  ? 'XML'
    : basename($0) eq 'jsonrpc.cgi' ? 'JSON'
    :                                 'UNKNOWN';

  # Method called
  my $method = $args->{full_method};

  # Find out the authentication method
  my $using_token  = Bugzilla->user->authorizer->login_token;
  my $using_cookie = Bugzilla->cgi->cookie('Bugzilla_login')
    && Bugzilla->cgi->cookie('Bugzilla_logincookie');
  my $type
    = $using_token
    ? $using_cookie
      ? 'BOTH'
      : 'TOKEN'
    : $using_cookie ? 'COOKIE'
    :                 'NONE';

  Bugzilla->logger->info(
    "__RPCLOG [$user_id] [$user_email] [$script] [$method] [$type]");
  ## REDHAT EXTENSION END 1090710

  return;
}

## REDHAT EXTENSION 1183474 BEGIN

sub rpc_sanitize_hash {
  my ($self, $hash) = @_;

  my $params = dclone($hash);

  $params->{Bugzilla_api_token} = '**REDACTED**'
    if defined $params->{Bugzilla_api_token};
  $params->{api_token}        = '**REDACTED**' if defined $params->{api_token};
  $params->{Bugzilla_api_key} = '**REDACTED**'
    if defined $params->{Bugzilla_api_key};
  $params->{api_key}        = '**REDACTED**' if defined $params->{api_key};
  $params->{Bugzilla_token} = '**REDACTED**' if defined $params->{Bugzilla_token};
  $params->{token}          = '**REDACTED**' if defined $params->{token};
  $params->{Bugzilla_password} = '**REDACTED**'
    if defined $params->{Bugzilla_password};
  $params->{password} = '**REDACTED**' if defined $params->{password};

  return $params;
}

sub rpc_log_arguments {
  my ($self, $args) = @_;

  return if (Bugzilla->params->{rpc_log_args} != 1);

  # Don't put newlines into the logfile.
  local $Data::Dumper::Indent = 0;

  my $method = $args->{method};

  if ($method eq 'Bug.add_attachment') {

    # Don't display attachment data in the log file.
    Bugzilla->logger->info("RPC CALL: bug.add_attachment");
  }
  elsif (ref($args->{params}) eq 'HASH') {
    my $params = $self->rpc_sanitize_hash($args->{params});
    Bugzilla->logger->info("RPC CALL: $method " . Dumper($params));
  }
  elsif (ref($args->{params}) eq 'ARRAY') {
    foreach my $item (@{$args->{params}}) {
      if (ref($item) eq 'HASH') {
        my $params = $self->rpc_sanitize_hash($item);
        Bugzilla->logger->info("RPC CALL: $method " . Dumper($params));
      }
    }
  }

  return;
}
## REDHAT EXTENSION 1183474 END

## REDHAT EXTENSION START 481660 987269
# For these two fields, we don't want to use Bugzilla::Search::build_subselect
# While it gives us some improvements for = queries, it has a massive loss
# for != queries (where it has bug_id IN (a, very, very, very, long, list) )
sub _alias_nonchanged {
  my ($invocant, $args) = @_;

  $args->{full_field} = "bugs_aliases.alias";
  $invocant->_do_operator_function($args);
  $args->{term} = qq{
        bugs.bug_id IN (
            SELECT bugs_aliases.bug_id
            FROM bugs_aliases
            WHERE $args->{term}
        )
    };

  return;
}
## REDHAT EXTENSION END 481660

sub _target_release_nonchanged {
  my ($invocant, $args) = @_;

  $args->{full_field} = "bugs_release.value";
  $invocant->_do_operator_function($args);
  $args->{term} = qq{
        bugs.bug_id IN (
            SELECT bugs_release.bug_id
            FROM bugs_release
            WHERE $args->{term}
        )
    };

  return;
}
## REDHAT EXTENSION END 987269

sub object_before_set {
  my ($self, $args) = @_;
  my $dbh = Bugzilla->dbh;

  my $object     = $args->{object};
  my $field_name = $args->{field};
  my $values     = $args->{value};

  return unless $object->isa('Bugzilla::Bug');

  my $user  = Bugzilla->user;
  my $field = new Bugzilla::Field({name => $field_name});

  ## REDHAT EXTENSION BEGIN 1039815
  # This checks that no matter what other checks we have, a user cannot add
  # or remove a multi select field value they cannot see. This check applies
  # to all mutli select values
  if (defined $field
    && $field->type == Bugzilla::Constants::FIELD_TYPE_MULTI_SELECT)
  {
    # Get the current values. We need to use SQL to get the
    # value to get around restrictions
    my $query          = "SELECT value FROM bug_$field_name WHERE bug_id = ?";
    my $current_values = $dbh->selectcol_arrayref($query, undef, $object->id);

    # Get list of legal values
    my $field_values = $field->legal_values;
    my ($removed, $added) = diff_arrays($current_values, $values);

    foreach my $new_value (@$added) {
      unless (any { lc $_->name eq lc $new_value } @$field_values) {
        ThrowUserError('field_cannot_add', {field => $field, value => $new_value});
      }
    }
    foreach my $old_value (@$removed) {
      unless (any { lc $_->name eq lc $old_value } @$field_values) {

        # We don't show the name to reveal a value they can't see
        ThrowUserError('field_cannot_remove', {field => $field});
      }
    }
  }
  ## REDHAT EXTENSION END 1039815

  return 1;
}

## REDHAT EXTENSION BEGIN 693936
# Remove aliases when removing bugs
# This is a Red Hat extension to support multiple aliases per bug
sub bug_remove_from_db {
  my ($self, $args) = @_;
  my $bug_id = $args->{bug_id};

  my $dbh = Bugzilla->dbh;

  my $query = "DELETE FROM bugs_aliases WHERE bug_id = ?";
  $dbh->do($query, undef, $bug_id);

  ## REDHAT EXTENSION START 420461
  my $comment_ids = $dbh->selectcol_arrayref(
    "SELECT comment_id FROM longdescs
                                  WHERE bug_id = ?", undef, $bug_id
  );
  if (scalar(@$comment_ids)) {
    $dbh->do("DELETE FROM longdescs_activity WHERE "
        . $dbh->sql_in('comment_id', $comment_ids));
  }
  ## REDHAT EXTENSION END 420461

  return;
}
## REDHAT EXTENSION END 693936

sub bugmail_recipients {
  my ($self, $args) = @_;
  my $recipients = $args->{recipients};
  my $diffs      = $args->{diffs};
  my $bug        = $args->{bug};

  my $dbh = Bugzilla->dbh;

  ## REDHAT EXTENSION BEGIN 235855
  my $ccarch = Bugzilla->params->{rh_auto_cc_rules} || [];
  foreach my $rule (@$ccarch) {
    if (any { $bug->{product} eq $_ } @{$rule->{product}}) {
      if (any { $bug->{rep_platform} eq $_ } @{$rule->{platform}}) {
        foreach my $email (@{$rule->{email}}) {
          my $user_id = Bugzilla::User::login_to_id($email);
          $recipients->{$user_id}->{+REL_ARCH_NOTIFY} = 1;
        }
      }
    }
  }
  ## REDHAT EXTENSION END 235855

  ## REDHAT EXTENSION BEGIN 1003675 1326559
  foreach my $comp (@{$bug->all_component_objs}) {
    next unless ($comp);
    if ($comp->always_cc) {
      foreach my $email (split(',\s*', $comp->always_cc)) {
        my $user_id = Bugzilla::User::login_to_id($email);
        $recipients->{$user_id}->{+REL_ALWAYS_CC} = 1 if ($user_id);
      }
    }
  }
  ## REDHAT EXTENSION END 1003675 1326559

  ## REDHAT EXTENSION START 798022 856375
  # Only send emails if:
  #   doc_text has changed AND the product is 'Red Hat' or 'JBoss'
  if (($bug->classification eq 'Red Hat' || $bug->classification eq 'JBoss')
    && any { $_->{field_name} eq 'cf_release_notes' } @$diffs)
  {
    # All members of the ecs_notify group get an e-mail
    # Start by getting a list of groups which inherit ecs_notify
    my $ecs_notify = new Bugzilla::Group({name => 'ecs_notify'});
    my $group_ids  = Bugzilla::Group->flatten_group_membership($ecs_notify->id);

    foreach my $group_id (@$group_ids) {

      # Get each member of the group
      my $group   = new Bugzilla::Group($group_id);
      my $members = $group->members_non_inherited;
      foreach my $user (@$members) {
        $recipients->{$user->id}->{+REL_ECS_NOTIFY} = 1;
      }
    }
  }

  ## REDHAT EXTENSION END 798022 856375

  return;
}

sub bugmail_relationships {
  my ($self, $args) = @_;
  my $relationships = $args->{relationships};

  ## REDHAT EXTENSION BEGIN 235855
  $relationships->{+REL_ARCH_NOTIFY} = 'ArchNotify';
  ## REDHAT EXTENSION END 235855
  ## REDHAT EXTENSION START 1003675
  $relationships->{+REL_ALWAYS_CC} = 'Always';
  ## REDHAT EXTENSION END 1003675

  return;
}

## REDHAT EXTENSION BEGIN 212492

sub _check_closed_duplicate {
  my ($self, $args) = @_;

  # If the bug is NOT being closed as a duplicate, return.
  if ($args->{bug}->resolution ne 'DUPLICATE') {
    return;
  }

  # If the system has turned these warnings off, return.
  unless (Bugzilla->params->{duplicate_warn_override}) {
    return;
  }

  # If the user has turned these warnings off, return.
  unless (Bugzilla->user->setting('warn_duplicate') eq 'on') {
    return;
  }

  # Skip this check if the request is coming from RPC.
  if (Bugzilla->usage_mode != USAGE_MODE_BROWSER) {
    return;
  }

  # If the user has confirmed that they know what they are doing, return.
  my $input = Bugzilla->input_params;
  if ($input->{confirm_duplicate_bug}) {
    return;
  }

  # Don't run if we are modifying multiple bugs
  if (not $input->{id}) {
    return;
  }

  # Don't run if the duplicate has not changed
  if (not defined $args->{bug}->{_dup_for_update}) {
    return;
  }

  $self->_check_for_duplicate_mismatch($args);

  return;
}

sub _check_for_duplicate_mismatch {
  my ($self, $args, $action) = @_;

  my $bug     = $args->{bug};
  my $dupe_of = $args->{bug}->{_dup_for_update};

  if ($bug->product_id != $dupe_of->product_id) {
    $self->_alert_duplicate_mismatch($bug, $dupe_of);
  }

  my ($removed) = diff_arrays($bug->all_versions, $dupe_of->all_versions);
  if (scalar(@$removed) > 0) {
    $self->_alert_duplicate_mismatch($bug, $dupe_of);
  }

  ($removed) = diff_arrays($bug->all_components, $dupe_of->all_components);
  if (scalar(@$removed) > 0) {
    $self->_alert_duplicate_mismatch($bug, $dupe_of);
  }

  # No mismatch found, nothing to do here.

  return;
}

sub _alert_duplicate_mismatch {
  my ($self, $bug, $dupe_of) = @_;

  my %vars = (bug => $bug, dupe => $dupe_of,);

  my $template = Bugzilla->template;
  print Bugzilla->cgi->header();
  $template->process("bug/process/warn_duplicate.html.tmpl", \%vars)
    || ThrowTemplateError($template->error());
  exit;
}

## REDHAT EXTENSION END 212492

sub bug_columns {
  my ($self, $args) = @_;
  my $columns = $args->{'columns'};

  ## REDHAT EXTENSION START 876015
  push(@$columns, "docs_contact");
  ## REDHAT EXTENSION END 876015

  return;
}

sub bug_prefetch {
  my ($self, $args) = @_;
  my $fields    = $args->{'fields'};
  my $in_clause = $args->{'in_clause'};

  ## REDHAT EXTENSION START 965089
  $fields->{alias} = {
    value => 'alias',
    sql   => qq{
            SELECT bug_id, alias
            FROM bugs_aliases
            WHERE bug_id $in_clause
            ORDER BY alias
        },
  };
  $fields->{component} = {
    value => 'all_components',
    sql   => qq{
            SELECT b.bug_id, c.name AS value
            FROM bugs b JOIN components c ON b.component_id = c.id
            WHERE b.bug_id $in_clause
            UNION ALL
            SELECT bug_id, c.name AS value
            FROM  rh_multiple_bug_components mbc
            JOIN components c ON (c.id = mbc.component_id)
            WHERE bug_id $in_clause
            ORDER BY value
        },
  };
  $fields->{target_release} = {
    value => 'target_release',
    sql   => qq{
            SELECT bug_id, value
            FROM bugs_release
            WHERE bug_id $in_clause
            ORDER BY value
        },
  };
  $fields->{version} = {
    value => 'all_versions',
    sql   => qq{
            SELECT bug_id, version AS value
            FROM bugs
            WHERE bug_id $in_clause
            UNION ALL
            SELECT mbv.bug_id, v.value AS value
            FROM rh_multiple_bug_versions mbv
            JOIN versions v ON (v.id = mbv.version_id)
            WHERE bug_id $in_clause
            ORDER BY value
        },
  };
  ## REDHAT EXTENSION END 965089

  return;
}

## REDHAT EXTENSION BEGIN 554957

sub _display_merge_warning_changes {
  my ($source_id, $target_id, $source_fields, $target_fields, $merged_fields)
    = @_;

  # Convert the internal field names into what the user is shown
  # eg cf_internal_whiteboard becomes 'Internal Whiteboard'
  my %field_names;
  foreach my $field_name (keys %{$source_fields->{fields}}) {
    my $field = Bugzilla::Field->new({name => $field_name});

    if (defined($field)) {

      # Use the long form
      $field_names{$field_name} = $field->description;
    }
    else {
      # Can't find an object with the given name
      $field_names{$field_name} = $field_name;
    }
  }

  my $template_args = {
    source_bug_id => $source_id,
    target_bug_id => $target_id,

    merged_flags => $merged_fields->{flags},

    source_fields => $source_fields->{fields},
    target_fields => $target_fields->{fields},
    merged_fields => $merged_fields->{fields},

    field_names => \%field_names,
  };

  # If the priority and status need to be merged, add them in to the fields.
  if (defined($merged_fields->{priority})) {
    $template_args->{source_fields}->{priority} = $source_fields->{priority};
    $template_args->{target_fields}->{priority} = $target_fields->{priority};
    $template_args->{merged_fields}->{priority} = $merged_fields->{priority};
    $template_args->{field_names}->{priority}   = 'Priority';
  }

  if (defined($merged_fields->{severity})) {
    $template_args->{source_fields}->{severity} = $source_fields->{severity};
    $template_args->{target_fields}->{severity} = $target_fields->{severity};
    $template_args->{merged_fields}->{severity} = $merged_fields->{severity};
    $template_args->{field_names}->{severity}   = 'Severity';
  }

  my $template = Bugzilla->template;
  $template->process("bug/process/warn_merge.html.tmpl", $template_args)
    || ThrowTemplateError($template->error());

  exit;
}

# Build up a list of fields that will be merged.
# Returns a hash of field_name => [ $values, ... ]

sub _select_fields {
  my ($bug) = @_;

  # need a list of fields that we are going to examine,

  # cf_pm_score - This is an integer and can't be merged by the current code.

  my @merge_fields = qw(
    cf_internal_whiteboard cf_devel_whiteboard cf_qa_whiteboard
    cf_partner cf_verified
    dependson blocks keywords status_whiteboard
  );

  my %results;

  my %fields;
  foreach my $field (@merge_fields) {
    next unless $bug->can($field);
    $fields{$field} = $bug->$field;
  }
  $results{fields} = \%fields;

  # Get the priority and severity.

  $results{priority} = $bug->priority;
  $results{severity} = $bug->bug_severity;

  # Build up a list of flags to be merged across
  my %flags;

  foreach my $flag (@{$bug->flags}) {
    my $flag_name = $flag->type->name;
    push @{$flags{$flag->type->id}}, _flag_to_hash($flag);
  }

  $results{flags} = \%flags;

  $results{bug_id} = $bug->id;

  return \%results;
}

# Merge duplicate bug fields together for either
# review or application.
#
# The target bug may not support all the flags or fields that we
# are trying to merge onto it.  In that case those fields will be
# dropped.
#
# Params:
#   $target_bug  - The bug that we will be merging the fields on to.
#   $target_data - The fields from the duplicate bug.
#   $source_data - The fields from the target bug.
#
#  Returns a hashref similar to _select_fields, but containing the
#  data that has been merged.

sub _merge_fields {
  my ($merge_results, $source_bug, $target_bug, $source_data, $target_data) = @_;

  my $source_fields = $source_data->{fields};
  my $target_fields = $target_data->{fields};

  my $target_bug_id = $target_data->{bug_id};
  my $source_bug_id = $source_data->{bug_id};

  my %merge_fields;

  foreach my $field (keys %$source_fields) {
    next unless $target_bug->can($field);

    my $source_field = $source_fields->{$field};
    my $target_field = $target_fields->{$field};

    if ($field eq 'keywords') {
      $source_field = [map { $_->name } @{$source_bug->keyword_objects}];
      $target_field = [map { $_->name } @{$target_bug->keyword_objects}];
    }

    # If this field type is an arrayref
    if (ref($source_field) eq 'ARRAY') {

      # Do nothing if there's no data.
      next if scalar(@{$source_field}) == 0;

      if ($field eq 'dependson' || $field eq 'blocks') {

 # A Bug cannot block or depend on itself.  If the source bug is a dependency or blocker
 # of the target bug, we must remove it from the merged array
        @$source_field = grep { $_ != $target_bug_id } @$source_field;
      }

      push my @ary, uniq @$source_field, @$target_field;

      # Don't make any changes if there is nothing to change
      next if scalar(@ary) == scalar(@$target_field);

      $merge_fields{$field} = \@ary;
    }

    # this field type is a string.
    else {
      next if length($source_field) == 0;

      if (length($target_field) == 0) {
        $merge_fields{$field} = $source_field;
      }
      else {
        next if index($target_field, $source_field) != -1;
        my $merged_str = $target_field . ' | ' . $source_field;
        $merge_fields{$field} = $merged_str if length($merged_str);
      }
    }
  }

  $merge_results->{fields} = \%merge_fields if scalar keys %merge_fields;

  # Merge priority and severity.

  if ($source_data->{severity} ne $target_data->{severity}) {
    my $result = _merge_priority(
      'bug_severity',
      $target_data->{severity},
      $source_data->{severity}
    );

    # Only perform the merge if they are different
    $merge_results->{severity} = $result if $result ne $target_data->{severity};
  }

  if ($source_data->{priority} ne $target_data->{priority}) {
    my $result = _merge_priority(
      'priority',
      $target_data->{priority},
      $source_data->{priority}
    );

    # Only perform the merge if they are different
    $merge_results->{priority} = $result if $result ne $target_data->{priority};
  }

  return;
}

sub _merge_flags {
  my ($merge_results, $source_bug, $target_bug, $source_data, $target_data) = @_;

  # Merge the flags.  The bug being cloned to must support the new
  # FlagType before we can assign it.  If it doesn't support the new
  # FlagType, then we drop it on the floor.

  my $target_flags = $target_data->{flags} || {};
  my $source_flags = $source_data->{flags} || {};

  my $flag_types = $target_bug->flag_types;

  my @merged_flag_list;
  my %merge_fields;

  # For each flag type
  while (my ($key, $value) = each(%$source_flags)) {

    # $key is the type_id of the flag list
    # $value is an arrayref of flags of this type.

    my ($flag_type) = grep { $_->id == $key } @$flag_types;
    unless (defined($flag_type)) {

      # This flag type is not supported. drop it on the floor.
      next;
    }

    # Get the target flags.
    my $target_flag_list = $target_flags->{$key};

    # we need to compare every source flag against every flag on the target.
    #
    # Case one - Target does not have any flags of this type already (easy)
    #            Add all the flags to the bug
    if (!defined($target_flag_list) || scalar(@$target_flag_list) == 0) {
      foreach my $f (@$value) {
        if (_can_set_flag($flag_type, $f)) {
          push @merged_flag_list, {source => $f, merged => $f};
        }
      }
    }

    # Case Two - Target does have flags of this time
    # If these flags are multiplicable then we need to determine if the
    # requestee is the same persion.
    # If requestee is different, append flag.
    # If requestee is the same, do nothing.

    elsif (defined($target_flag_list) && $flag_type->is_multiplicable) {

      my @local_target_list = @$target_flag_list;

      foreach my $single_flag (@$value) {
        my $matched;
        for (my $i = 0; $i < scalar(@local_target_list); $i++) {
          if ($local_target_list[$i]->{requestee_id} eq $single_flag->{requestee_id}) {
            $matched = delete $local_target_list[$i];
            last;
          }
        }

        # If $matched has anything in it, we need to merge the flag.
        if (defined($matched)) {

          # Don't do merging if status is the same
          if ($matched->{status} ne $single_flag->{status}) {
            my $merged = _select_flag($matched, $single_flag);
            if (_can_set_flag($flag_type, $merged)) {
              push @merged_flag_list,
                {source => $single_flag, merged => $merged, target => $matched};
            }
          }
        }
        else {
          # Not a match - add it to the merged list directly
          if (_can_set_flag($flag_type, $single_flag)) {
            push @merged_flag_list, {source => $single_flag, merged => $single_flag};
          }
        }
      }
    }

    # Case Three - Target has a non-multiplicable flag.
    # Flag will need to be merged according to precedence.
    else {
      # If the flags have the same status, don't do anything.
      if ($target_flag_list->[0]->{status} ne $value->[0]->{status}) {

        # Not a multiplicable flag, so each array element will only have one element.
        my $flg = _select_flag($target_flag_list->[0], $value->[0]);
        if (_can_set_flag($flag_type, $flg)) {
          push @merged_flag_list,
            {source => $value->[0], target => $target_flag_list->[0], merged => $flg};
        }
      }
    }
  }

  # If we have merged some flags, apply them to the results list.
  $merge_results->{flags} = \@merged_flag_list if scalar @merged_flag_list;

  return;
}


# This sorts out which priority or severity takes precedence.
sub _merge_priority {
  my ($field_name, $target_value, $source_value) = @_;

  my $valid_values = get_legal_field_values($field_name);

  # push the unspecified value on to the end of the array
  my $unspec = undef;
  $unspec = shift @$valid_values if $valid_values->[0] eq 'unspecified';
  if (defined($unspec)) {
    push @$valid_values, $unspec;
  }

  # Find the index of each of the priority strings
  # these are sorted in order of precedence.
  my $source_idx = first_index { $_ eq $source_value } @$valid_values;
  my $target_idx = first_index { $_ eq $target_value } @$valid_values;

  my $greater_idx = $target_idx > $source_idx ? $source_idx : $target_idx;

  # And return whichever is the highest.
  my $target_priority = $valid_values->[$greater_idx];

  return $target_priority;
}

# Merge the two flag arrays
# Highest value flag will override lower value flags in the event of a clash.
# - > + > ? > ' '
# Return a newly merged array of flags.

sub _select_flag {
  my ($f1, $f2) = @_;

  my $f1_st = $f1->{status};
  my $f2_st = $f2->{status};

  # - > + > ? > ' '
  if ($f1_st eq '-' or $f2_st eq '-') {
    return $f1_st eq '-' ? $f1 : $f2;
  }
  elsif ($f1_st eq '+' or $f2_st eq '+') {
    return $f1_st eq '+' ? $f1 : $f2;
  }
  elsif ($f1_st eq '?' or $f2_st eq '?') {
    return $f1_st eq '?' ? $f1 : $f2;
  }
  return $f1;
}

# Helper function to tell us if a user can request or set a flag
sub _can_set_flag {
  my ($flag_type, $flag) = @_;

  if ($flag->{status} eq '?') {
    return Bugzilla->user->can_request_flag($flag_type);
  }

  return Bugzilla->user->can_set_flag($flag_type);
}


# Given the list of merged fields, apply them to the bug.

sub _apply_new_fields {
  my ($target_bug, $data) = @_;

  # For every field in the bug, set the new arguments.
  # If the field begins with 'cf_' then it's a custom field.

  my $fields        = $data->{fields};
  my @custom_fields = Bugzilla->active_custom_fields;

  # This will set all the custom fields.
  foreach my $field (keys %$fields) {
    next unless defined $fields->{$field};
    next unless $field =~ m/^cf_/;

    my $field_data = $fields->{$field};

    if (ref($field_data) eq 'ARRAY') {
      next if scalar(@{$field_data}) == 0;
    }
    else {
      next if length($field_data) == 0;
    }

    my ($custom_field) = grep { $_->name eq $field } @custom_fields;
    $target_bug->set_custom_field($custom_field, $field_data);
  }

  if (defined($fields->{status_whiteboard})) {
    $target_bug->set_status_whiteboard($fields->{status_whiteboard});
  }

  if (defined($fields->{keywords})) {
    $target_bug->modify_keywords($fields->{keywords}, 'set');
  }

  if (defined($fields->{dependson})) {
    $target_bug->set_all({dependson => {add => $fields->{dependson}}});
  }

  if (defined($fields->{blocked})) {
    $target_bug->set_all({blocked => {add => $fields->{blocked}}});
  }

  # If bug severity or priority exist, set them.
  if (defined($data->{severity})) {
    $target_bug->set_severity($data->{severity});
  }

  if (defined($data->{priority})) {
    $target_bug->set_priority($data->{priority});
  }

  return $target_bug;
}

# Given a list of merged flags, apply them to the bug

sub _apply_new_flags {
  my ($target_bug, $data) = @_;

  my $flags = $data->{flags};

  my @new_flags;

  foreach my $flag (@$flags) {
    my $new_flag = {
      status    => $flag->{merged}{status},
      setter_id => $flag->{merged}{setter_id},
      type_id   => $flag->{merged}{type_id},
    };

    # Add the requestee_id if we have one.
    $new_flag->{requestee} = $flag->{source}{requestee}
      if defined $flag->{source}{requestee};

    # If the target bug already has this flag, set the flag_id so we
    # update the existing flag instead of setting a new one.
    $new_flag->{id} = $flag->{target}->{flag_id}
      if defined $flag->{target}->{flag_id};

    push @new_flags, $new_flag;
  }

  $target_bug->set_flags([], \@new_flags);

  return;
}

# Convert a flag object into a hashref.

sub _flag_to_hash {
  my $flag = shift;

  my $flag_args = {
    name             => $flag->type->name,
    status           => $flag->status,
    type_id          => $flag->type->id,
    setter_id        => $flag->setter_id,
    flag_id          => $flag->id,
    is_multiplicable => $flag->type->is_multiplicable,
  };

  if (defined($flag->requestee_id)) {
    $flag_args->{requestee_id} = $flag->requestee_id;
    $flag_args->{requestee}    = $flag->requestee->login;
  }

  return $flag_args;
}

## REDHAT EXTENSION END 554957

# Add hooks to the admin menu, that could change visiblity to the menu
sub user_can_administer {
  my ($self, $args) = @_;
  my $can_administer = $args->{can_administer};
  my $user           = Bugzilla->user;

  return;
}

## REDHAT EXTENSION BEGIN 1326559
sub _allows_on_dev        { return $_[0]->{'allows_on_dev'}; }
sub _allows_no_clear_acks { return $_[0]->{'allows_no_clear_acks'}; }
sub _always_cc            { return $_[0]->{'always_cc'}; }
sub _set_allows_on_dev    { $_[0]->set('allows_on_dev', $_[1]); return; }

sub _set_allows_no_clear_acks {
  $_[0]->set('allows_no_clear_acks', $_[1]);
  return;
}
sub _set_always_cc { $_[0]->set('always_cc', $_[1]); return; }
## REDHAT EXTENSION END 1326559

#########
# Pages #
#########

## REDHAT EXTENSION START 1324287

sub page_before_template {
  my ($self, $args) = @_;

  my $page = $args->{page_id};
  my $vars = $args->{vars};

  if ($page =~ m{^redhat/doctypetext/list\.}) {
    _page_doctypetext_list($vars);
  }
  elsif ($page =~ m{^redhat/doctypetext/edit\.}) {
    _page_doctypetext_edit($vars);
  }
  elsif ($page =~ m{^redhat/doctypetext/confirm-delete\.}) {
    _page_doctypetext_confirm_delete($vars);
  }
  ## REDHAT EXTENSION START 1456689
  elsif ($page =~ m{^redhat/groupgraphs/select\.}) {
    _page_groupgraphs_select($vars);
  }
  ## REDHAT EXTENSION END 1456689
  elsif ($page =~ m{^redhat/admin_user_sessions\.}) {
    _page_admin_user_sessions($vars);
  }
  elsif ($page =~ m{^redhat/ban_apikey\.}) {
    _page_ban_apikeys($vars);
  }
  elsif ($page =~ m{^redhat/admin_user_access\.}) {
    _page_admin_user_access($vars);
  }

  return;
}


sub _page_admin_user_sessions {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  my $user = Bugzilla->user;
  $user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "edit", object => "usersessions"});

  my $who    = $input->{'who'};
  my $action = $input->{'action'} || 'list';

  if ($action eq 'list') {
    $vars->{'token'} = issue_session_token('admin_user_sessions');

    my $who_user = Bugzilla::User->new({name => $who});

    my $sessions = Bugzilla::User::Session->match({userid => $who_user->id});

    $vars->{sessions} = $sessions;
    $vars->{who}      = $who_user;

    my $api_keys = Bugzilla::User::APIKey->match({user_id => $who_user->id});
    $vars->{api_keys} = $api_keys;

  }

  return;
}

sub _page_admin_user_access {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  my $user = Bugzilla->user;
  $user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "edit", object => "admin_user_access"});

  my $who = $input->{'who'};

  my $who_user = Bugzilla::User->check({name => $who});
  $vars->{who} = $who_user;

  return;
}

sub _page_ban_apikeys {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  my $user = Bugzilla->user;
  $user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "edit", object => "usersessions"});

  my $thekey = $input->{'thekey'};
  my $token  = $input->{'token'};

  if ($thekey) {
    check_token_data($token, 'admin_ban_apikeys');

    my $crypted_key = bz_crypt($thekey, Bugzilla->localconfig->{'site_wide_secret'},
      PASSWORD_DIGEST_ALGORITHM);

    my $api_key = Bugzilla::User::APIKey->new({name => $crypted_key});

    if (!$api_key || $api_key->api_key ne $crypted_key) {
      ThrowUserError("api_key_not_valid");
    }

    delete_token($token);

    $api_key->set_banned(1);
    $api_key->update();
    $vars->{'api_key'} = $api_key;
    $vars->{'thekey'}  = $thekey;

    $vars->{'message'} = 'api_key_banned';
  }

  $vars->{'token'} = issue_session_token('admin_ban_apikeys');

  return;
}

sub _page_doctypetext_list {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "doctypetext"});

  # often used variables
  my $token  = $input->{'token'};
  my $action = $input->{'action'} || 'list';
  my $id     = $input->{'doc_type_text_id'};

  detaint_natural($id);


  if ($action eq 'add') {
    check_token_data($token, 'add_doctypetext');
    my $args
      = {cf_doc_type_id => $input->{cf_doc_type_id}, value => $input->{value},};

    my $doctypetext = Bugzilla::Extension::RedHat::DocTypeText->create($args);
    delete_token($token);

    $vars->{'message'}     = 'doc_type_text_created';
    $vars->{'doctypetext'} = $doctypetext;
  }
  elsif ($action eq 'update') {
    check_token_data($token, 'edit_doctypetext');

    $id || ThrowCodeError('param_required', {param => 'id'});

    my $doctypetext = Bugzilla::Extension::RedHat::DocTypeText->new($id);
    $doctypetext->set_cf_doc_type_id($input->{'cf_doc_type_id'});
    $doctypetext->set_value($input->{'value'});
    $doctypetext->update();
    delete_token($token);
    $vars->{'message'}     = 'doc_type_text_updated';
    $vars->{'doctypetext'} = $doctypetext;
  }
  elsif ($action eq 'delete') {
    $id || ThrowCodeError('param_required', {param => 'id'});

    my $doctypetext = Bugzilla::Extension::RedHat::DocTypeText->new($id);

    check_token_data($token, 'delete_doctypetext');
    $doctypetext->remove_from_db();
    delete_token($token);

    $vars->{'message'}     = 'doc_type_text_deleted';
    $vars->{'doctypetext'} = $doctypetext;
  }

  $vars->{'doctypetexts'} = Bugzilla::Extension::RedHat::DocTypeText->match();
  $vars->{'doc_types'}
    = Bugzilla::Field->new({name => 'cf_doc_type'})->legal_values();

  return;
}

sub _available_doc_types {
  my $id = shift || 0;
  my @available;
  my $dbh = Bugzilla->dbh;
  my $dts = Bugzilla::Field->new({name => 'cf_doc_type'})->legal_values();
  foreach my $dt (@$dts) {

    # is anyone other than me using it?
    my $used
      = $dbh->selectrow_array(
      "SELECT id FROM doc_type_text WHERE id <> ? AND cf_doc_type_id = ?",
      undef, $id, $dt->id);
    push(@available, $dt) unless ($used);
  }

  return (\@available);
}


sub _page_doctypetext_edit {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "doctypetext"});

  my $action = $input->{'action'} || '';

  if ($action eq 'new') {
    $vars->{'action'}    = 'add';
    $vars->{'token'}     = issue_session_token('add_doctypetext');
    $vars->{'doc_types'} = _available_doc_types();
    return;
  }

  my $id = $input->{'doc_type_text_id'};
  detaint_natural($id);
  $id || ThrowCodeError('param_required', {param => 'doc_type_text_id'});
  $vars->{'doctypetext'} = Bugzilla::Extension::RedHat::DocTypeText->new($id);
  $vars->{'doc_types'}   = _available_doc_types($id);
  $vars->{'action'}      = 'update';
  $vars->{'token'}       = issue_session_token('edit_doctypetext');

  return;
}

sub _page_doctypetext_confirm_delete {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "doctypetext"});

  my $id = $input->{'doc_type_text_id'};
  detaint_natural($id);
  $id || ThrowCodeError('param_required', {param => 'doc_type_text_id'});

  $vars->{'token'}       = issue_session_token('delete_doctypetext');
  $vars->{'doctypetext'} = Bugzilla::Extension::RedHat::DocTypeText->new($id);
  my $field = Bugzilla::Field->new({name => 'cf_doc_type'});
  $vars->{'doc_type'} = Bugzilla::Field::Choice->type($field)
    ->new($vars->{'doctypetext'}->{cf_doc_type_id})->name;

  return;
}

sub _all_doc_texts {
  my @docs;
  my $field = Bugzilla::Field->new({name => 'cf_doc_type'});
  foreach my $doc (@{Bugzilla::Extension::RedHat::DocTypeText->match()}) {
    $doc->{key}
      = Bugzilla::Field::Choice->type($field)->new($doc->{cf_doc_type_id})->name;
    push(@docs, $doc);
  }

  return (@docs);
}

## REDHAT EXTENSION END 1324287

## REDHAT EXTENSION START 1456689
sub _page_groupgraphs_select {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->is_insider() || ThrowUserError(
    "auth_failure",
    {
      group  => Bugzilla->params->{'insidergroup'},
      action => "view",
      object => "groupgraphs"
    }
  );

  $vars->{graph} = "";

  if ($input->{'graph'} && $input->{'graph'} ne "") {
    my $graph  = $input->{'graph'};
    my $format = lc($input->{'format'} // 'dot');
    my $wide   = $input->{'wide'} // 0;
    my $admin  = $input->{'admin'} // 0;
    my $svg    = _gen_graph($graph, $format, $wide, $admin);
    $vars->{svg}    = $svg;
    $vars->{graph}  = $graph;
    $vars->{format} = uc($format);
    $vars->{wide}   = $wide;
    $vars->{admin}  = $admin;
  }

  return;
}

sub _gen_graph {
  my ($type, $format, $wide, $admin) = @_;

  my $graph = Graph::Easy->new({timeout => 500000});

  my %groups = (
    'Red Hat'    => {'bg' => '#ff323c',    edge => '#A30000', active => 'magenta',},
    'Partner'    => {'bg' => 'lightgreen', edge => 'green',   active => 'green',},
    'Admin'      => {'bg' => 'orange',     edge => 'orange',  active => 'white',},
    'Deprecated' => {'bg' => '#FFBF00',    edge => 'blue',    active => 'red',},
    'Public'     => {'bg' => 'lightblue',  edge => 'black',   active => 'grey',},
    'Agile' => {'bg' => 'green', edge => 'darkgreen', active => 'darkgreen',},
  );


  foreach my $key (keys %groups) {
    $groups{$key}->{node} = $graph->add_group($key);
    $groups{$key}->{node}->set_attribute('fill', $groups{$key}->{bg});
  }

  my $svg;
  if ($type eq 'membership') {
    $svg = _group_inheritance($graph, $format, $wide, $admin, \%groups);
  }
  elsif ($type eq 'group_prod_map') {
    $svg = _group_product_map($graph, $format, $wide, $admin, \%groups);
  }
  elsif ($type eq 'group_rights') {
    _group_inheritance($graph, $format, $wide, $admin, \%groups, 1, 1);
    $svg = _group_rights($graph, $format, $wide, $admin, \%groups);
  }
  else {
    ThrowUserError("invalid_groupgraph_type", {type => $type});
  }

  return ($svg);
}

sub _group_inheritance {
  my ($graph, $format, $wide, $admin, $groups, $nosvg, $simple) = @_;

  my @skips = qw/view_nonsecure_bugs edit_all_bugs view_all_bugs/;
  push(@skips, 'admin') unless ($admin);

  my $query = <<SQL;
SELECT groups.id, groups.name, groups.category, groups.description
FROM groups
SQL

  my $dbh = Bugzilla->dbh();

  my $sth = $dbh->prepare($query);

  $sth->execute;
  foreach my $row (@{$sth->fetchall_arrayref}) {

    my $edge;
    my $name = $row->[1];

    next if ($simple && any { $_ eq $name } @skips);

    my $category = $row->[2] // 'Public';
    $groups->{$category}->{node}->add_node($name);

    $graph->node($name)->set_attribute('title', $row->[3]) if ($graph->node($name));

  }

  $query = <<SQL;
SELECT groups.id, groups.name, groups.isactive, grantor_id, grant_type, g2.name, g2.category, groups.category,
    (SELECT count(distinct product_id) FROM group_control_map gmc LEFT JOIN products p ON gmc.product_id = p.id WHERE groups.id = gmc.group_id AND (membercontrol >= 1 OR othercontrol >= 1) AND p.isactive = 1),
    (SELECT count(distinct product_id) FROM group_control_map gmc2 LEFT JOIN products p2 ON gmc2.product_id = p2.id WHERE gmc2.group_id = grantor_id AND p2.isactive = 1)
FROM groups
LEFT JOIN group_group_map ON groups.id = group_group_map.member_id
LEFT JOIN groups g2 ON group_group_map.grantor_id = g2.id
ORDER BY groups.id, g2.id, grant_type
SQL

  $sth = $dbh->prepare($query);
  $sth->execute;
  foreach my $row (@{$sth->fetchall_arrayref}) {

    my $edge;
    my $name1 = $row->[1];
    my $name2;
    my $category = $row->[7] // '';

    next if ($simple && any { $_ eq $name1 } @skips);

    my $node = $graph->add_node($name1);

    $node->set_attribute('fill', $groups->{$category}->{active}) if ($row->[8] > 0);

    if (defined($row->[5]) && $row->[5] ne '') {
      $name2 = $row->[5];
    }

    if ($name2) {
      my $comment = $node->get_attribute('comment') // "";
      $comment .= "$name2#";
      $node->set_attribute('comment', $comment);

      next if ($simple && any { $_ eq $name2 } @skips);

      $edge = $graph->add_edge_once($name1, $name2)
        if ($admin || $category !~ m/(?:Admin)/i);

      if ($edge) {
        $edge->set_attribute('color', $groups->{$category}->{'edge'});
      }
    }
  }

  return () if ($nosvg);

  $graph->randomize();
  my $svg = _gen_svg($graph, $format, $wide);
  return ($svg);
}

sub _group_rights {
  my ($graph, $format, $wide, $admin, $groups) = @_;

  my $query = <<SQL;
SELECT flagtypes.id, flagtypes.name, grant_group.name, req_group.name, view_group.name, products.name, grant_group.category, req_group.category, view_group.category, products.classification_id
FROM flagtypes
    LEFT JOIN groups AS grant_group ON flagtypes.grant_group_id = grant_group.id
    LEFT JOIN groups AS req_group ON flagtypes.request_group_id = req_group.id
    LEFT JOIN groups AS view_group ON flagtypes.view_group_id = view_group.id
    LEFT JOIN flaginclusions ON flagtypes.id = flaginclusions.type_id
    LEFT JOIN flagexclusions ON flagtypes.id = flagexclusions.type_id
    LEFT JOIN products ON flaginclusions.product_id = products.id
    WHERE flagtypes.is_active = 1
    ORDER BY flagtypes.name
SQL

  my $dbh = Bugzilla->dbh();

  my $graph2 = Graph::Easy->new({timeout => 500000});

  my $sth = $dbh->prepare($query);
  $sth->execute;
  my %labels;

  foreach my $row (@{$sth->fetchall_arrayref}) {

    # can grant
    if (defined($row->[2])) {
      push(@{$labels{$row->[2]}->{'Grant flags'}}, $row->[1]);
    }

    # can req
    if (defined($row->[3])) {
      push(@{$labels{$row->[3]}->{'Request flags'}}, $row->[1]);
    }

    # can view
    if (defined($row->[4])) {
      push(@{$labels{$row->[4]}->{'View flags'}}, $row->[1]);
    }
  }

  foreach my $cf (@{Bugzilla->fields({obsolete => 0, custom => 1})}) {
    my $grps = $cf->admin_groups();
    foreach my $grp (keys %{$grps}) {
      push(@{$labels{$grps->{$grp}->name}->{"Can admin CF"}}, $cf->name);
    }

    $grps = $cf->edit_groups();
    foreach my $grp (keys %{$grps}) {
      push(@{$labels{$grps->{$grp}->name}->{"Can edit CF"}}, $cf->name);
    }

    $grps = $cf->view_groups();
    foreach my $grp (keys %{$grps}) {
      push(@{$labels{$grps->{$grp}->name}->{"Can view CF"}}, $cf->name);
    }
  }

  foreach my $node ($graph->nodes()) {
    my $comment = $node->get_attribute('comment');
    if ($comment) {
      my $name = $node->name();
      @{$labels{$name}->{'Member of'}} = sort(split('#', $comment));
    }
  }

  $query = <<SQL;
SELECT p.name, g.name
FROM group_control_map gcm
    LEFT JOIN products p ON gcm.product_id = p.id
    LEFT JOIN groups g ON g.id = gcm.group_id
WHERE gcm.editcomponents = 1 AND p.isactive = 1
ORDER BY p.name
SQL

  $sth = $dbh->prepare($query);
  $sth->execute;
  foreach my $row (@{$sth->fetchall_arrayref}) {
    push(@{$labels{$row->[1]}->{"Edit Products"}}, $row->[0]);
  }

  $query = <<SQL;
SELECT p.name, g.name
FROM group_control_map gcm
    LEFT JOIN products p ON gcm.product_id = p.id
    LEFT JOIN groups g ON g.id = gcm.group_id
WHERE (gcm.membercontrol >= 1 OR gcm.othercontrol >= 1)
AND p.isactive = 1
SQL

  $query .= q{ AND g.name <> 'admin'} unless ($admin);

  $query .= <<SQL;

ORDER BY p.name
SQL

  $sth = $dbh->prepare($query);
  $sth->execute;

  foreach my $row (@{$sth->fetchall_arrayref}) {
    push(@{$labels{$row->[1]}->{"Setable on Products"}}, $row->[0]);
  }

  my @types = (
    'Grant flags',
    'Request flags',
    'View flags',
    'Can admin CF',
    'Can edit CF',
    'Can view CF',
    'Edit Products',
    'Member of',
    'Setable on Products'
  );
  foreach my $grp (keys %labels) {
    my $node  = $graph->add_node($grp);
    my $label = "$grp\\l";
    foreach my $type (@types) {
      if (defined $labels{$grp}->{$type}) {

        # BUGBUG you need a '|' here, but it will blow up if you try and set it here
        # instead you have to replace it after the graphviz document has been generated
        $label
          .= "PANTS{$type\\l" . join("\\r", uniq(@{$labels{$grp}->{$type}})) . "\\r}";
      }
    }
    $label .= '';
    $node->set_attribute('label', "$label");
  }


  $graph->randomize();

  my $svg = _gen_svg($graph, $format, $wide);

  return $svg;
}

sub _group_product_map {
  my ($graph, $format, $wide, $admin, $groups) = @_;

  my $query = <<SQL;
SELECT p.name, g.name, c.name
FROM group_control_map gcm
    LEFT JOIN products p ON gcm.product_id = p.id
    LEFT JOIN groups g ON g.id = gcm.group_id
    LEFT JOIN classifications c ON p.classification_id = c.id
WHERE g.category in ('Red Hat')
AND (membercontrol >= 1 OR othercontrol >= 1)
SQL

  my $graph4     = Graph::Easy->new({timeout => 500000});
  my $prod_group = $graph4->add_group('Products');
  my $grp_group  = $graph4->add_group('Groups');

  my %grp_cols = (
    redhat             => 'darkred',
    program_management => 'red',
    security           => 'lightgreen',
    qe_staff           => 'blue'
  );

  my $dbh = Bugzilla->dbh();

  my $sth = $dbh->prepare($query);
  $sth->execute;

  foreach my $row (@{$sth->fetchall_arrayref}) {
    my $cl_grp = $graph4->add_group($row->[2]);
    my $pname  = $row->[0];
    $cl_grp->add_node($pname);
    my ($g, $p, $edge) = $graph4->add_edge_once($row->[1], $pname);

    unless ($grp_cols{$row->[1]}) {
      my $col = sprintf("rgb(%s,%s,%s)", int rand(256), int rand(256), int rand(256));
      my $col_name = Graph::Easy->color_name(Graph::Easy->color_as_hex($col));
      $grp_cols{$row->[1]} = $col_name;
    }
    $edge->set_attribute('color', $grp_cols{$row->[1]});
  }

  my $svg = _gen_svg($graph4, $format, $wide);
  return $svg;
}

sub _gen_svg {
  my ($gr, $format, $wide) = @_;

  my $graphviz = $gr->as_graphviz();

  # BUGBUG see PANTS above
  $graphviz =~ s/PANTS/|/g;

  # The graph module only makes boxes, but we want records.
  $graphviz =~ s/shape=box/shape=record/;

  my $input = Bugzilla->input_params;
  if ($input->{download}) {
    print Bugzilla->cgi->header(
      -type                => qq{application/octet-stream; name="groups-$format.gv"},
      -content_disposition => qq{attachment; filename="groups-$format.gv"},
    );
    disable_utf8();
    print $graphviz;
    exit;
  }

  my $conc = $wide ? 'false' : 'true';
  trick_taint($format);

  my $svg;
  run [
    "/usr/bin/$format", '-Tsvg', '-Gsplines=polyline', '-Gcenter=1',
    "-Gconcentrate=$conc"
    ],
    \$graphviz, ">", binary(), \$svg;

  return ($svg);
}

## REDHAT EXTENSION END 1456689

## REDHAT EXTENSION START 1490644
sub password_change_request {
  my ($self, $args) = @_;
  my $user = $args->{'user'};

  return () unless ($user && $user->id);

  my $host     = Bugzilla->params->{LDAPMXserver};
  my $basedn   = Bugzilla->params->{LDAPMXBaseDN};
  my $pass     = Bugzilla->localconfig->{LDAPMXpass};
  my $searchdn = Bugzilla->params->{LDAPMXSearchDN};
  my $attr     = Bugzilla->params->{LDAPMXmailattribute};

  my $ldap = new Net::LDAP($host, scheme => 'ldaps', onerror => 'die');
  $ldap->bind(dn => $basedn, password => $pass);

  my $login = $user->login;

  # handle jfearn+test@redhat.com
  if ($login =~ m/([^@+]+)\+([^@]+)@([^@]+)$/) {
    $login = "$1\@$3";
  }

  my $search
    = $ldap->search(base => $searchdn, filter => 'rhatEmailAddress=' . $login);

  $search->code && die $search->error;
  foreach my $entry ($search->entries) {

    my @addrs = $entry->get_value($attr);
    foreach my $addr (@addrs) {
      if ( $addr =~ /^(.*)\@post-office.corp.redhat.com$/
        || $addr =~ /^(.*)\@listman.util.phx.redhat.com$/)
      {
        ThrowUserError('forgotten_password_change_disabled',
          {account => $login, contact => Bugzilla->params->{maintainer}});
      }
    }
  }

  return;
}
## REDHAT EXTENSION END 1490644

# RED HAT EXTENSION START 1690644
# When creating a 'user' object, validate the redhat mail
sub object_before_create {
  my ($self, $args) = @_;
  my $class  = $args->{'class'};
  my $params = $args->{'params'};

  if ($class->isa('Bugzilla::User')) {
    my $valid = _validate_redhat_addr($params->{'login_name'});

    ThrowUserError('cannot_create_invalid_redhat_mail',
      {account => $params->{'login_name'}})
      unless ($valid);
  }

  return;
}

sub object_end_of_update {

  my ($self, $args) = @_;

  my ($object, $old_object, $changes) = @$args{qw(object old_object changes)};

  # Note that this is a made-up class, for this example.
  if ($object->isa('Bugzilla::User')) {

    my $valid = _validate_redhat_addr($object->login);

    if (defined $changes->{'name'}) {
      my ($old, $new) = @{$changes->{'name'}};
      print "The name field changed from $old to $new!";
    }

    if (!$valid && Bugzilla->user->login ne Bugzilla->params->{system_user}) {
      my ($old_dt, $new_dt);
      ($old_dt, $new_dt) = @{$changes->{'disabledtext'}}
        if (defined $changes->{'disabledtext'});
      my ($old_dm, $new_dm);
      ($old_dm, $new_dm) = @{$changes->{'disable_mail'}}
        if (defined $changes->{'disable_mail'});

      if ($object->is_enabled || $object->email_enabled) {
        ThrowUserError('cannot_update_invalid_redhat_mail',
          {account => $object->login});
      }
    }
  }

  return;
}

sub _validate_redhat_addr {
  my $login = shift;

  return (1) unless ($login =~ /redhat.com$/);

  return (0) unless ($login =~ /\@redhat.com$/);

  return (1) if ($ENV{PROJECT} && $ENV{PROJECT} eq 'tests');

  # handle jfearn+test@redhat.com
  if ($login =~ m/([^@+]+)\+([^@]+)@([^@]+)$/) {
    $login = "$1\@$3";
  }

  my $valid = 0;

  my $host     = Bugzilla->params->{LDAPMXserver};
  my $basedn   = Bugzilla->params->{LDAPMXBaseDN};
  my $pass     = Bugzilla->localconfig->{LDAPMXpass};
  my $searchdn = Bugzilla->params->{LDAPMXSearchDN};
  my $attr     = Bugzilla->params->{LDAPMXmailattribute};

  my $ldap = new Net::LDAP($host, scheme => 'ldaps', onerror => 'die');

  eval { $ldap->bind(dn => $basedn, password => $pass); };

  if ($@) {
    ThrowCodeError('ldap_cannot_bind', {msg => $@});
  }

  my $search
    = $ldap->search(base => $searchdn, filter => 'rhatEmailAddress=' . $login);

  $search->code && die $search->error;
  foreach my $entry ($search->entries) {

    my @addrs = $entry->get_value('rhatEmailAddress');
    foreach my $addr (@addrs) {

      if ($addr eq $login) {
        $valid = 1;
        last;
      }
    }
  }

  return ($valid);
}

# RED HAT EXTENSION END 1690644

## REDHAT EXTENSION START 1312306
sub update_product_add_vars {
  my ($self, $args) = @_;
  my $vars = $args->{vars};
  $self->_update_editcomponent_groups($args);

  return;
}

sub _update_editcomponent_groups {
  my ($self, $args) = @_;

  my $product = $args->{product};
  my $changes = $args->{changes};
  my $vars    = $args->{vars};
  my $dbh     = Bugzilla->dbh;
  my (@removed, @added);

  my $egps_in = $vars->{editcomponents_groups};
  $egps_in = [$egps_in] if (ref($egps_in) ne 'ARRAY');

  $dbh->bz_start_transaction();

  my $gc     = $product->group_controls;
  my @gc_ids = grep { $gc->{$_}{editcomponents} } keys %$gc;
  my @ec_ids;
  @ec_ids = map { Bugzilla::Group->check({name => $_})->id } @$egps_in
    if ($egps_in->[0]);
  my $lc  = List::Compare->new(\@gc_ids, \@ec_ids);
  my @rem = $lc->get_unique;
  my @add = $lc->get_complement;

  if (@add) {
    foreach my $gid (@add) {
      my $group = Bugzilla::Group->new($gid);
      if ($gc->{$group->id}) {
        $dbh->do(
          "UPDATE group_control_map SET editcomponents = 1 WHERE product_id = ? AND group_id = ?",
          undef,
          ($product->id, $group->id)
        );
      }
      else {
        $dbh->do(
          "INSERT INTO group_control_map (product_id, group_id, editcomponents) VALUES (?,?,?)",
          undef,
          ($product->id, $group->id, 1)
        );
      }
      push(@added, $group->name);
    }
  }

  if (@rem) {
    foreach my $gid (@rem) {
      my $group = Bugzilla::Group->new($gid);
      $dbh->do(
        "UPDATE group_control_map SET editcomponents = 0 WHERE product_id = ? AND group_id = ?",
        undef,
        ($product->id, $group->id)
      );
      push(@removed, $group->name);
    }
  }

  delete($product->{group_controls}) if (@add || @rem);

  $dbh->bz_commit_transaction();

  $changes->{editcomponents_groups} = [join(', ', @removed), join(', ', @added)];

  return;
}
## REDHAT EXTENSION END 1312306

## REDHAT EXTENSION START 1323778
sub _mandatory_fields {
  my $invocant = shift;
  my $params   = shift;

  my @mandatory_fields;
  foreach my $field (
    @{Bugzilla->fields({is_mandatory => 1, enter_bug => 1, obsolete => 0})})
  {
    if ($field->is_mandatory && $field->is_visible_on_bug($params // $invocant)) {
      push(@mandatory_fields, $field);
    }
  }
  return (@mandatory_fields);
}
## REDHAT EXTENSION END 1323778

## REDHAT EXTENSION START 1569389
sub template_before_create {
  my ($self, $args) = @_;

  # This is how we add dynamic filters
  $args->{config}->{FILTERS}->{filter_relevance} = [
    sub {
      my ($context, @args) = @_;
      return sub {
        my $txt = shift;

        # don't process strings that don't match
        if ($txt && $txt ne '') {
          $txt = sprintf("%2.5f", $txt);
        }
        return ($txt);
      }
    },
    1
  ];

  return;
}

## REDHAT EXTENSION END 1569389

# Can we delete a user?
# Stolen from editusers.cgi
sub _user_can_delete {
  my $self = shift;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "edit", object => "users"});

  my $dbh = Bugzilla->dbh;
  my ($ok_to_delete, $not_ok_to_delete);

  # Find other cross references.
  $not_ok_to_delete->{'attachments'}
    = $dbh->selectrow_array(
    'SELECT COUNT(*) FROM attachments WHERE submitter_id = ?',
    undef, $self->id);

  $ok_to_delete->{'assignee_or_qa'} = $dbh->selectrow_array(
    qq{SELECT COUNT(*)
           FROM bugs
           WHERE assigned_to = ? OR qa_contact = ? OR docs_contact = ?}, undef,
    ($self->id, $self->id, $self->id)
  );

  $not_ok_to_delete->{'reporter'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM bugs WHERE reporter = ?',
    undef, $self->id);

  $ok_to_delete->{'cc'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM cc WHERE who = ?',
    undef, $self->id);

  $not_ok_to_delete->{'bugs_activity'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM bugs_activity WHERE who = ?',
    undef, $self->id);

  $ok_to_delete->{'component_cc'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM component_cc WHERE user_id = ?',
    undef, $self->id);

  $ok_to_delete->{'email_setting'}
    = $dbh->selectrow_array(
    'SELECT COUNT(*) FROM email_setting WHERE user_id = ?',
    undef, $self->id);

  $ok_to_delete->{'flags_requestee'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM flags WHERE requestee_id = ?',
    undef, $self->id);

  $not_ok_to_delete->{'flags_setter'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM flags WHERE setter_id = ?',
    undef, $self->id);

  $not_ok_to_delete->{'longdescs'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM longdescs WHERE who = ?',
    undef, $self->id);

  my $namedquery_ids
    = $dbh->selectcol_arrayref('SELECT id FROM namedqueries WHERE userid = ?',
    undef, $self->id);

  $ok_to_delete->{'namedqueries'} = scalar(@$namedquery_ids);

  if (scalar(@$namedquery_ids)) {
    $ok_to_delete->{'namedquery_group_map'}
      = $dbh->selectrow_array(
          'SELECT COUNT(*) FROM namedquery_group_map WHERE namedquery_id IN' . ' ('
        . join(', ', @$namedquery_ids)
        . ')');
  }
  else {
    $ok_to_delete->{'namedquery_group_map'} = 0;
  }

  $ok_to_delete->{'profile_setting'}
    = $dbh->selectrow_array(
    'SELECT COUNT(*) FROM profile_setting WHERE user_id = ?',
    undef, $self->id);

  $not_ok_to_delete->{'profiles_activity'}
    = $dbh->selectrow_array(
    'SELECT COUNT(*) FROM profiles_activity WHERE who = ? AND userid != ?',
    undef, ($self->id, $self->id));

  $ok_to_delete->{'quips'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM quips WHERE userid = ?',
    undef, $self->id);

  $ok_to_delete->{'series'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM series WHERE creator = ?',
    undef, $self->id);

  $ok_to_delete->{'watched'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM watch WHERE watched = ?',
    undef, $self->id);

  $ok_to_delete->{'watcher'}
    = $dbh->selectrow_array('SELECT COUNT(*) FROM watch WHERE watcher = ?',
    undef, $self->id);

  $ok_to_delete->{'whine_events'}
    = $dbh->selectrow_array(
    'SELECT COUNT(*) FROM whine_events WHERE owner_userid = ?',
    undef, $self->id);

  $ok_to_delete->{'whine_schedules'} = $dbh->selectrow_array(
    qq{SELECT COUNT(distinct eventid)
           FROM whine_schedules
           WHERE mailto = ?
           AND mailto_type = ?
          }, undef, ($self->id, MAILTO_USER)
  );

  $not_ok_to_delete->{'product_responsibilities'}
    = scalar @{$self->product_responsibilities};
  $ok_to_delete->{'has_audit_entries'} = $self->has_audit_entries;

  my $can = 1;

  foreach my $key (keys %$not_ok_to_delete) {
    if ($not_ok_to_delete->{$key} > 0) {
      $can = 0;
      last;
    }
  }

  return ({
    can              => $can,
    ok_to_delete     => $ok_to_delete,
    not_ok_to_delete => $not_ok_to_delete,
    name             => $self->login,
    id               => $self->id,
  });
}

sub _user_delete_me {
  my $self = shift;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "delete", object => "users"});

  my $dbh = Bugzilla->dbh;

  my $deletable = $self->can_delete();
  if ($deletable->{can} != 1) {
    ThrowUserError("cannot_delete_user",
      {object => $self, stats => Dumper($deletable)});
  }

  # Lock tables during the check+removal session.
  # XXX: if there was some change on these tables after the deletion
  #      confirmation checks, we may do something here we haven't warned
  #      about.
  $dbh->bz_start_transaction();

  Bugzilla->logout_user($self);

  # Get the timestamp for LogActivityEntry.
  my $timestamp = $dbh->selectrow_array('SELECT NOW()');

  # When we update a bug_activity entry, we update the bug timestamp, too.
  my $sth_set_bug_timestamp
    = $dbh->prepare('UPDATE bugs SET delta_ts = ? WHERE bug_id = ?');

  # Flags
  my $flag_ids
    = $dbh->selectcol_arrayref('SELECT id FROM flags WHERE requestee_id = ?',
    undef, $self->id);

  my $flags = Bugzilla::Flag->new_from_list($flag_ids);

  $dbh->do(
    'UPDATE flags SET requestee_id = NULL, modification_date = ?
              WHERE requestee_id = ?', undef, ($timestamp, $self->id)
  );

  my %updatedbugs;

  # We want to remove the requestee but leave the requester alone,
  # so we have to log these changes manually.
  my %bugs;
  push(@{$bugs{$_->bug_id}->{$_->attach_id || 0}}, $_) foreach @$flags;
  foreach my $bug_id (keys %bugs) {
    my $bug = Bugzilla::Bug->new($bug_id);
    foreach my $attach_id (keys %{$bugs{$bug_id}}) {
      my @old_summaries = Bugzilla::Flag->snapshot($bugs{$bug_id}->{$attach_id});
      $_->_set_requestee() foreach @{$bugs{$bug_id}->{$attach_id}};
      my @new_summaries = Bugzilla::Flag->snapshot($bugs{$bug_id}->{$attach_id});
      my ($removed, $added)
        = Bugzilla::Flag->update_activity(\@old_summaries, \@new_summaries);
      LogActivityEntry(
        $bug_id,   'flagtypes.name', $removed, $added,
        $user->id, $timestamp,       undef,    $attach_id,
        $bug
      );
    }
    $sth_set_bug_timestamp->execute($timestamp, $bug_id);
    $updatedbugs{$bug_id} = 1;
  }

  # Get every bug the user has a role on
  my $bug_ids = $dbh->selectall_arrayref(
    'SELECT DISTINCT bug_id
         FROM bugs
         WHERE  assigned_to = ?
            OR   qa_contact = ?
            OR docs_contact = ?
    UNION
    SELECT DISTINCT bug_id FROM cc WHERE who = ?', undef, $self->id, $self->id,
    $self->id, $self->id
  );

  my $bugs = Bugzilla::Bug->new_from_list($bug_ids);

  # Now update them all
  foreach my $bug (@$bugs) {
    $bug->reset_assigned_to()  if ($bug->{'assigned_to'} == $self->id);
    $bug->reset_qa_contact()   if ($bug->{'qa_contact'} == $self->id);
    $bug->reset_docs_contact() if ($bug->{'docs_contact'} == $self->id);
    $bug->remove_cc($self);
    $bug->update;
  }

  # Finally, remove the user account itself.
  $dbh->do('DELETE FROM profiles WHERE userid = ?', undef, $self->id);

  eval { $dbh->bz_commit_transaction(); };
  if ($@) {
    ThrowUserError("couldnot_delete_user", {object => $self, error => $@});

  }

  ## Can we update the flag in the same place as the other changes?
  foreach (keys(%updatedbugs)) {
    Bugzilla::BugMail::Send($_, {'changer' => $user});
  }

  return {$self->login => 'deleted'};
}

## REDHAT EXTENSION START 1772732
sub _user_is_bug_updated {
  my ($self, $bug_id) = @_;

  my $dbh = Bugzilla->dbh;

  my $sql = <<EOS;
SELECT count(*)
FROM bug_user_last_visit
LEFT JOIN bugs on bugs.bug_id = bug_user_last_visit.bug_id
WHERE bug_user_last_visit.user_id = ?
AND bug_user_last_visit.bug_id = ?
AND bugs.delta_ts > bug_user_last_visit.last_visit_ts
EOS

  my $updated = $dbh->selectrow_array($sql, undef, $self->id, $bug_id);

  return $updated;
}

sub _user_requested_on_bug {
  my ($self, $bug) = @_;

  my $dbh = Bugzilla->dbh;

  my $sql = <<EOS;
SELECT count(*)
FROM flags
WHERE requestee_id = ?
AND bug_id = ?
AND status = '?'
EOS

  my $count = $dbh->selectrow_array($sql, undef, $self->id, $bug->{id});

  return $count if ($count);

  my $the_bug;

  if (ref($bug) eq 'Bugzilla::Bug') {
    $the_bug = $bug;
  }
  else {
    $the_bug = Bugzilla::Bug->new($bug->{id});
  }

  foreach my $flag (@{$the_bug->flags}) {
    if ($flag->name =~ m/^approvals_request_(.+)$/) {
      $count++ if ($flag->user_can_edit($self));
    }
  }

  return $count;
}
## REDHAT EXTENSION END 1772732

sub check_login_name {
  my ($self, $args) = @_;
  my $invocant = $args->{invocant};
  my $name     = $args->{name};

  # Check the name if it's a new user, or if we're changing the name.
  if (!ref($invocant) || lc($invocant->login) ne lc($name)) {
    ## RED HAT EXTENSION START 1358115
    # can't move in to or out of redhat'
    if ( ref($invocant)
      && $invocant->login ne ""
      && (lc($invocant->login) ne lc($name))
      && ($name =~ m/\@redhat.com$/ || $invocant->login =~ m/\@redhat.com$/)
      && ($name !~ m/\@redhat.com$/ || $invocant->login !~ m/\@redhat.com$/))
    {
      ThrowUserError('constrain_redhat_accounts');
    }

    # Don't do any more processing of @redhat.com accounts
    return () if ($name =~ m/\@redhat.com$/);

    # Can't move out of a partner group
    if ( ref($invocant)
      && $invocant->login ne ""
      && (lc($invocant->login) ne lc($name))
      && ($invocant->is_partner_user()))
    {
      my $OK = 0;
      foreach my $group (@{$invocant->groups}) {
        if ($group->category eq 'Partner' && $group->user_regexp) {
          my $regexp = $group->user_regexp;
          if ($regexp ne '' and $name =~ /$regexp/i) {
            $OK = 1;
            last;
          }
        }
      }
      ThrowUserError('constrain_partner_accounts') unless ($OK);
    }

    # Can't move in to a partner group
    if ( ref($invocant)
      && $invocant->login ne ""
      && (lc($invocant->login) ne lc($name))
      && (!$invocant->is_partner_user()))
    {
      foreach my $group (Bugzilla::Group->get_all()) {
        if ($group->category eq 'Partner' && $group->user_regexp) {
          my $regexp = $group->user_regexp;
          if ($regexp ne '' and $name =~ /$regexp/i) {
            ThrowUserError('constrain_partner_accounts');
          }
        }
      }
    }
    ## RED HAT EXTENSION END 1358115
  }

  return;
}

## REDHAT EXTENSION BEGIN 1051965
sub _user_is_partner_user {
  my $self = shift;
  return 0 if $self->is_insider();

  foreach my $group (@{$self->groups}) {
    return 1 if $group->category && $group->category eq 'Partner';
  }

  return 0;
}
## REDHAT EXTENSION END 1051965

sub set_groups_to_object {
  my ($self, $args) = @_;

  my $user  = $args->{user};
  my $group = $args->{group};

  ## REDHAT EXTENSION START 1271052 - Only admins may promote to admin groups
  if (($group->category eq 'Admin') && (!Bugzilla->user->in_group("admin"))) {
    ThrowUserError('not_admin', {group => $group->name});
  }
  ## REDHAT EXTENSION END 1271052
  ## REDHAT EXTENSION START 1358117
  if ( ($group->category eq 'Admin' || $group->category eq 'Red Hat')
    && ($user->login !~ m/\@redhat\.com$/))
  {
    ThrowUserError('protect_internal_groups', {group => $group->name});
  }
  if ($group->category eq 'Partner') {
    my $regexp = $group->user_regexp;
    my $wf     = $group->group_workflow;

    $regexp = $wf->regex if ($wf && $wf->regex && $wf->regex ne '');

    # Have to allow redhat for onsite partner engineers
    if ( $regexp
      && $regexp ne ''
      && ($user->login !~ m/$regexp/ && $user->login !~ m/\@redhat\.com$/))
    {
      ThrowUserError('protect_partner_groups',
        {group => $group->name, regexp => $regexp});
    }
  }
  ## REDHAT EXTENSION END 1358117

  return;
}

sub can_change_email {
  my ($self, $args) = @_;

  my $can = $args->{can};

  my $user = Bugzilla->user;
  if ($user->login =~ /\@redhat\.com$/ || $user->is_partner_user) {
    ${$args->{can}} = 0;
  }

  return;
}

## REDHAT EXTENSION START 1821023

sub job_map {
  my ($self, $args) = @_;

  my $job_map = $args->{job_map};

  # This adds the named class (an instance of TheSchwartz::Worker) as a handler
  $job_map->{'user_term'} = 'Bugzilla::Extension::RedHat::Job::Termination';

  return;
}

sub _user_manage_user_access {
  my $self   = shift;
  my $params = shift;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "manage_user_access"});

  defined($params->{names})
    || ThrowCodeError('param_required', {param => 'names'});


  defined($params->{action})
    || ThrowCodeError('param_required', {param => 'action'});

  my $action  = delete $params->{action};
  my $to      = delete $params->{replacement};
  my $message = delete $params->{message}
    // "$action performed by " . Bugzilla->user->identity();

  my $revoke  = 1;
  my $ban     = 0;
  my $disable = 1;
  my $close   = 0;
  my $gdpr    = 0;
  my $spam    = 0;

  my $nobody_user_name = Bugzilla->params->{'nobody_user'};

  if ($action eq 'remove') {
    $revoke = 0;
    $ban    = 1;
    $close  = 1;
  }
  elsif ($action eq 'redact') {
    $revoke = 0;
    $ban    = 1;
    $close  = 1;
    $gdpr   = 1;
  }
  elsif ($action eq 'repudiate') {
    $revoke = 0;
    $ban    = 1;
    $close  = 1;
    $spam   = 1;
  }
  elsif ($action eq 'reconfirm') {
    $disable = 0;
  }
  elsif ($action eq 'revoke') {
    ## NOP
  }
  else {
    ThrowUserError('no_such_action', {action => $action});
  }

  my $to_user;

  if ($to) {
    eval { $to_user = Bugzilla::User->new({name => $to}); };
  }

  my $dbh = Bugzilla->dbh;
  my %changes;

  my @user_objects;

  foreach my $login (@{$params->{names}}) {
    my $other_user;
    eval { $other_user = Bugzilla::User->new({name => $login}); };

    if ($@) {
      $changes{$login}->{messages}
        = ["Error loading account for login $login, no actions taken.\nERROR: $@"];
    }
    elsif ($other_user) {
      push(@user_objects, $other_user);
    }
    else {
      $changes{$login}->{messages}
        = ["No account was found for login $login, no actions taken."];
    }
  }

  foreach my $other_user (@user_objects) {
    my $changed = 0;
    my @messages;
    my $already_disabled = $other_user->email_disabled;

    $dbh->bz_start_transaction;

    my %counts = (
      bugs         => 0,
      spam_bugs    => 0,
      sessions     => 0,
      keys_banned  => 0,
      keys_revoked => 0,
      components   => 0,
      msgs         => [],
    );

    if ($to && !$to_user) {
      push(@messages,
        "No account was found for the replacement login '$to', roles and bugs will be assigned to $nobody_user_name"
      );
    }

    my $ticket = $params->{ticket} // $other_user->id;

    $other_user->set_password(generate_random_password(64, 1));

    my $sessions = Bugzilla::User::Session->match({userid => $other_user->id});

    foreach my $session (@$sessions) {
      $session->remove_from_db;
      $counts{sessions}++;
    }

    my $api_keys = Bugzilla::User::APIKey->match({user_id => $other_user->id});
    foreach my $apikey (@$api_keys) {
      if ($revoke) {
        $apikey->set_revoked(1);
        $counts{keys_revoked}++;
      }

      if ($ban) {
        $apikey->set_banned(1);
        $counts{keys_banned}++;
      }

      $apikey->update();
    }

    if ($disable) {
      $other_user->set_disabledtext($message);
      $other_user->set_disable_mail(1);
    }

    if ($close) {

      # remove all DIRECT groups and bless.
      # It will not remove REGEX groups, this is expected.
      $other_user->set_groups({set => []});
      $other_user->set_bless_groups({set => []});

      # remove all product responsibilities
      foreach my $record (@{$other_user->product_responsibilities}) {

        # Skip Fedora. TODO this should not be hard coded
        next if ($record->{product}->classification_id == 2);

        foreach my $comp (@{$record->{components}}) {
          if ($comp->default_assignee->id == $other_user->id) {
            $comp->set_default_assignee($to_user ? $to : $nobody_user_name);
          }

          if ( $comp->default_qa_contact
            && $comp->default_qa_contact->id == $other_user->id)
          {
            $comp->set_default_qa_contact($to_user ? $to : '');
          }

          if ( $comp->default_docs_contact
            && $comp->default_docs_contact->id == $other_user->id)
          {
            $comp->set_default_docs_contact($to_user ? $to : '');
          }

          my @current_cc = split ',\s*', $comp->cc_list_string;
          if (any { $_ eq $other_user->login } @current_cc) {
            my @initial_cc = grep { $_ ne $other_user->login } @current_cc;
            $comp->set_cc_list(\@initial_cc);
          }
          my $returned_changes = $comp->update();
          $counts{components}++ if ($returned_changes);
        }
      }

      # Sub-components
      foreach my $record (@{$other_user->sub_component_responsibilities}) {

        # Skip Fedora. TODO this should not be hard coded
        next if ($record->{product}->classification_id == 2);

        foreach my $comp (@{$record->{components}}) {
          foreach my $sub_comp (@{$comp->{sub_components}}) {

            if ($sub_comp->default_assignee->id == $other_user->id) {
              $sub_comp->set_default_assignee($to_user ? $to : $nobody_user_name);
            }

            if ( $sub_comp->default_qa_contact
              && $sub_comp->default_qa_contact->id == $other_user->id)
            {
              $sub_comp->set_default_qa_contact($to_user ? $to : '');
            }

            if ( $sub_comp->default_docs_contact
              && $sub_comp->default_docs_contact->id == $other_user->id)
            {
              $sub_comp->set_default_docs_contact($to_user ? $to : '');
            }

            my @current_cc = split ',\s*', $sub_comp->cc_list_string;
            if (any { $_ eq $other_user->login } @current_cc) {
              my @initial_cc = grep { $_ ne $other_user->login } @current_cc;
              $sub_comp->set_cc_list(\@initial_cc);
            }
            my $returned_changes = $sub_comp->update();
            $counts{components}++ if ($returned_changes);
          }
        }
      }

      # Obfuscate some data
      if ($gdpr) {
        my $is_partner = $other_user->is_partner_user();
        my $mail       = $other_user->login;
        my $newmail    = 'nobody+' . $ticket;
        if ($mail =~ /\@redhat.com$/) {
          $newmail .= '@redhat.com';
        }
        elsif ($is_partner) {

          # Partner mail addresses must not be move out of the partner's domain.
          $mail =~ m/(\@.*)$/;
          $newmail .= $1;
        } ## end elsif($is_partner)
        else {
          $newmail .= '@r.c.invalid';
        }

        $other_user->set_login($newmail);
        $other_user->set_name('Persona non grata');
      }
    }

    my $ret_changes;

    eval { $ret_changes = $other_user->update(); };

    if ($@) {
      $dbh->bz_rollback_transaction();
      push(@messages, "User was not updated due to: $@");
      next;
    }
    else {
      $dbh->bz_commit_transaction();
      if ($ret_changes) {
        $changes{$other_user->login}->{account_changes} = $ret_changes;
        $changed = 1;
        my $msg = "User has been " . $action;
        if ($action =~ /e$/) {
          $msg .= q{d};
        }
        else {
          $msg .= q{ed};
        }
        push(@messages, $msg);
      }
      else {
        push(@messages, "No direct changes to user.");
      }
    }

    my @bug_ids;

    # Get open bugs the user has a role on
    if ($close) {
      my $bug_ids = $dbh->selectall_arrayref(
        q{SELECT DISTINCT bug_id
          FROM bugs b1
            LEFT JOIN products p1 ON b1.product_id = p1.id
          WHERE bug_status <> 'CLOSED'
          AND (
               assigned_to  = ?
            OR qa_contact   = ?
            OR docs_contact = ?
          )
          AND p1.classification_id NOT IN (2)
        UNION
          SELECT DISTINCT cc.bug_id
          FROM cc
            LEFT JOIN bugs b2 ON cc.bug_id = b2.bug_id
            LEFT JOIN products p2 ON b2.product_id = p2.id
          WHERE who = ?
          AND b2.bug_status <> 'CLOSED'
          AND p2.classification_id NOT IN (2)
        }, undef, $other_user->id, $other_user->id, $other_user->id, $other_user->id
      );

      @bug_ids = map { $_->[0] } @{$bug_ids};
    }

    # Add any bugs a spammer opened
    if ($spam) {
      my $bug_ids
        = $dbh->selectall_arrayref(
        'SELECT DISTINCT bug_id FROM bugs WHERE reporter = ?',
        undef, $other_user->id);
      my @spam_bug_ids = map { $_->[0] } @{$bug_ids};
      @bug_ids = uniq @bug_ids, @spam_bug_ids;
    }

    $counts{bugs} = scalar @bug_ids;

    foreach my $bug_id (@bug_ids) {
      Bugzilla->job_queue->insert(
        'user_term',
        {
          bug_id  => $bug_id,
          action  => $action,
          login   => $other_user->login,
          actor   => Bugzilla->user->login,
          message => $message,
        }
      );
    }

    $changes{$other_user->id}->{counts}   = \%counts;
    $changes{$other_user->id}->{messages} = \@messages;

    # mail user about change.
    if ($changed && !$close) {
      my $timestamp = $dbh->selectrow_array("SELECT NOW()");

      my %args = (
        to_user => $other_user,
        from    => Bugzilla->params->{'mailfrom'},
        date    => format_time($timestamp, '%a, %d %b %Y %T %z', 'UTC'),
        message => $message,
        action  => $action,
      );

      my $template = Bugzilla->template_inner('en');

      my $msg;
      $template->process('email/account-notify.txt.tmpl', \%args, \$msg);

      MessageToMTA($msg);
    }
  }

  return {users => \%changes};
}
## REDHAT EXTENSION END 1821023

__PACKAGE__->NAME;

__END__

=head1 NAME

Red Hat Extension

=head1 DESCRIPTION

This module contains many customisation for https://bugzilla.redhat.com

=head1 FEATURES

This is a list of some of the features provided by this extension.

Bugzilla::Product::allows_on_dev:
  This field allows a product to use the ON_DEV status.

Bugzilla::Product::allows_no_clear_acks:
  This field allows a product to prevent the pm_ack and dev_ack flags being reset
  when a bug's component changes.

Bugzilla::Component::always_cc:
  A list of user accounts that will always be CCd on mails for bugs, without appearing
  on the bugs CC list.

Doc Types:
  This feature allows the Docs team to label and contribute documentation for bugs.

Group Graphs:
  This feature allows redhat users to render SVG graphs of group relationships.

Red Hat Theme:
  This extension supplies the bulk of the Red Hat Theme.

Admin User Sessions:
  This Admin UI allows site admins to remove a user's login sessions and ban or
  revoke a user's APIKeys.
  See L<Bugzilla::Extension::RedHat::WebService::Bugzilla/terminate-sessions>

Ban leaked APIKeys:
  This Admin UI allows site admins to ban leaked APIKeys.
  See L<Bugzilla::Extension::RedHat::WebService::Bugzilla/admin-apikeys>

Edit comments API:
  This API allows editing the comments on a bug.
  See L<Bugzilla::Extension::RedHat::WebService::Bugzilla/edit-comment>

Manage User Account:
  This API allows various levels of disabling accounts.
  See L<Bugzilla::Extension::RedHat::WebService::Bugzilla/manage_user_access>

=cut

