#!/usr/bin/env perl
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

use strict;
use warnings;

use lib qw(/var/www/html/bugzilla/ .);

use DBI();
use Carp;
use Data::Dumper;
use File::Slurp qw(read_file);
use YAML qw(Load);
use Getopt::Long qw(:config no_auto_abbrev);
use Bugzilla;
use Pod::Usage;
use List::MoreUtils qw(any);

my $man     = 0;       # display man page
my $help    = 0;       # display help
my $compare = 0;
my $types   = undef;
my $verbose = undef;
my $nulls   = undef;
my $conf    = undef;
my $keys    = undef;
my $all     = undef;

GetOptions(
  'help|?'    => \$help,
  'man'       => \$man,
  'compare|c' => \$compare,
  'types|t'   => \$types,
  'nulls|n'   => \$nulls,
  'verbose|v' => \$verbose,
  'conf=s'    => \$conf,
  'keys|k'    => \$keys,
  'all|a'     => \$all,
) || pod2usage(-verbose => 0);

pod2usage(-verbose => 1) if ($help);
pod2usage(-verbose => 2) if ($man);

$Data::Dumper::Sortkeys = 1;

=pod

=head1 NAME

bz-schema-audit.pl - Audit the bz_schema table against the current Bugzilla DB or compare the schemas between two Bugzilla Databases, or.

=head1 Description

This script can compare two Bugzilla databases or audit a single Bugzilla DB to ensure the bz_schema tables matches the actual DB.

To compare two databases you need to supply the path to a YAML file containing the Db details for two Bugzilla databases.

e.g.

=over 4
source:
  host: "foo.example.com"
  name: "bugs"
  user: "bugs"
  password: "secretz"
  type: 'Pg'
  port: 5432
target:
  host: "bar.example.com"
  name: "bugs_test"
  user: "bugs"
  password: "secretz"
  type: 'Pg'
  port: 5432

=back

=head1 SYNOPSIS

bz-schema-audit.pl  [options]

=head1 OPTIONS

=over 4

=item B<--help>

Print a brief help message and exits.

=item B<--man>

Prints the manual page and exits.

=item B<--conf=<path to YAML file>>

The YAML configuration file.

=item B<--compare>

Compare two Bugzilla DBs.

=item B<--types>

Include type checking when auditing the schema.


=item B<--nulls>

Include is null-able checking when auditing the schema.

=item B<--keys>

Include primary & foreign keys checks when auditing the schema.

=item B<--all>

Perfomr alll the audit functions

=item B<--verbose>

Print out some data structures when reporting.

=back

=cut

if ($all) {
  $types = $nulls = $keys = 1;
}

if ($compare) {
  compare();
}
else {
  audit();
}

exit;

##### subs #####
sub compare {

  die("Missing conf file, `bz-schema-audit.pl --conf myconfig.yaml`\n")
    unless ($conf);

  my $yaml = read_file($conf) || die "Could not read file '$conf'";
  my $config = Load($yaml) || die "Could not parse YAML from file '$ARGV[0]'";

  my %con1 = (
    type => $config->{source}{type},
    host => $config->{source}{host},
    name => $config->{source}{name},
    port => $config->{source}{port},
    user => $config->{source}{user},
    pass => $config->{source}{password},
  );
  my %con2 = (
    type => $config->{target}{type},
    host => $config->{target}{host},
    name => $config->{target}{name},
    port => $config->{target}{port},
    user => $config->{target}{user},
    pass => $config->{target}{password},
  );

  my $dbh1 = get_dbh(\%con1);
  my $dbh2 = get_dbh(\%con2);

  my $tables1 = db_list_tables($dbh1, \%con1);
  my $tables2 = db_list_tables($dbh2, \%con2);

  foreach my $table (sort keys %{$tables1}) {
    if (defined $tables2->{$table}) {
      compare_tables($dbh1, \%con1, $dbh2, \%con2, $table);
    }
    else {
      print "table $table in db1 but not in db2\n";
      print(Dumper($tables1->{$table})) if ($verbose);
    }
  }
  foreach my $table (sort keys %$tables2) {
    if (!defined $tables1->{$table}) {
      print "table $table in db2 but not in db1\n";
      print(Dumper($tables2->{$table})) if ($verbose);
    }
  }

  $dbh1->disconnect();
  $dbh2->disconnect();
  return;
}

sub get_dbh {
  my $conn = shift;

  my $attr = {AutoCommit => 0, RaiseError => 1};

  my $dbh;

  if ($conn->{type} eq 'SQLite') {
    $dbh = DBI->connect("dbi:SQLite:dbname=" . $conn->{file}, "", "", $attr);

    if ($conn->{utf8}) {
      $dbh->{sqlite_unicode} = 1;
    }
  }
  else {
    my $dsn = sprintf("dbi:%s:dbname=%s;host=%s;port=%s",
      $conn->{type}, $conn->{name}, $conn->{host}, $conn->{port},);

    if ($conn->{type} eq 'mysql' and $conn->{utf8}) {
      $attr->{mysql_enable_utf8} = 1;
    }
    $dbh = DBI->connect($dsn, $conn->{user}, $conn->{pass}, $attr);
  }

  return ($dbh);
}

sub db_list_tables {
  my ($dbh, $conn) = @_;
  my $schema  = undef;
  my $catalog = undef;
  my $table   = '%';

  # get a list of table names from the database
  if ($conn->{type} =~ m/pg/i) {
    $catalog = 'pg_catalog';
    $schema  = 'public';
  }
  elsif ($conn->{type} =~ m/mysql/i) {
    ## NOP
  }
  else {
    croak "Unknown type '$conn->{type}' on database $conn->{name}";
  }

  my $sth         = $dbh->table_info($catalog, $schema, $table);
  my $tables_info = $sth->fetchall_hashref('TABLE_NAME');

  return ($tables_info);
}

sub compare_tables {
  my ($dbh1, $con1, $dbh2, $con2, $table) = @_;

  my $fields1 = db_list_table_fields($dbh1, $con1, $table);
  my $fields2 = db_list_table_fields($dbh2, $con2, $table);

  foreach my $col (sort keys %$fields1) {
    if (!defined $fields2->{$col}) {
      print "column $table.$col in DB1 but not in DB2\n";
      print(Dumper($fields1->{$col})) if ($verbose);
    }
    elsif (0
      && (
        $fields1->{$col}->{ORDINAL_POSITION} != $fields2->{$col}->{ORDINAL_POSITION}))
    {
      print "column $table.$col is in position "
        . $fields1->{$col}->{ORDINAL_POSITION}
        . " in DB1 but in position "
        . $fields2->{$col}->{ORDINAL_POSITION}
        . " in DB2\n";
    }
  }
  foreach my $col (sort keys %$fields2) {
    print "column $table.$col in DB2 but not in DB1\n"
      if (!defined $fields1->{$col});
  }
  return;
}

sub db_list_table_fields {
  my $dbh    = shift;
  my $conn   = shift;
  my $table  = shift;
  my $column = shift || '%';
  unless ($conn->{type} =~ /^(?:pg|mysql)$/i) {
    croak "Unknown type '$conn->{type}' on database $conn->{name}";
  }

  # get the column information for each table
  my $catalog = undef;
  my $schema  = undef;
  my $sth     = $dbh->column_info($catalog, $schema, $table, $column);
  my $ref     = $dbh->selectall_hashref($sth, 'COLUMN_NAME');

  # initialise COLUMN_KEYS to an array
  for my $field (keys %{$ref}) {
    $ref->{$field}{COLUMN_KEYS} = [];
  }

  # get all the keys of a table in mysql using the sql query
  # describe table_name;
  if ($conn->{type} =~ /^mysql$/i) {

    my $sql = "DESCRIBE $table";

    my $keys_ref = $dbh->selectall_hashref($sql, "Field");
    for my $field (keys %{$keys_ref}) {

      my $key_type = $keys_ref->{$field}{Key};
      if ($ref->{$field}) {
        push @{$ref->{$field}{COLUMN_KEYS}}, $key_type;
      }

    }

  }

  # if the driver is postgres then use sql query
  # to get all the key information from database
  if ($conn->{type} =~ /^Pg$/i) {

    my $sql = "SELECT r.relname AS Table,
                   c.conname AS constraint_name,
                   contype AS constraint_type,
                   conkey AS key_columns
                   FROM pg_class r, pg_constraint c
                   WHERE r.oid=c.conrelid
                   AND relname='$table'";

    my $constraints = $dbh->selectall_hashref($sql, "constraint_name");

    # initialize the KEY_CONSTRAINT to an array
    # to store the ordinal positions of the key columns
    for my $const_name (keys %{$constraints}) {
      $constraints->{$const_name}{KEY_CONSTRAINT} = [];
    }

    # extract the oridinal positions numbers from the
    # string key_columns and store them in the array
    # KEY_CONSTRAINT
    for my $const_name (keys %{$constraints}) {
      my $keys_position = $constraints->{$const_name}{"key_columns"};

      push @{$constraints->{$const_name}{KEY_CONSTRAINT}},
        ($keys_position =~ m/(\d+)/g);

    }

    # store all the key types except for the foreign keys
    # in the array COLUMN_KEYS
    for my $field (keys %{$ref}) {
      for my $const_name (keys %{$constraints}) {
        unless ($constraints->{$const_name}{constraint_type} eq 'f') {
          for my $element (@{$constraints->{$const_name}{KEY_CONSTRAINT}}) {

            if ($ref->{$field}{ORDINAL_POSITION} == $element) {
              push @{$ref->{$field}{COLUMN_KEYS}},
                $constraints->{$const_name}{constraint_type};
            }
          }
        }
      }

    }

    # get the foreign key information using the function foreign_key_info
    # and not the above sql query and that is to be able to get the info
    # about the referenced key and the referenced column
    my $fk_sth
      = $dbh->foreign_key_info(undef, undef, undef, $catalog, $schema, $table);

    if ($fk_sth) {
      my $fk_ref = $fk_sth->fetchall_hashref('FK_COLUMN_NAME');

      for my $field (keys %{$fk_ref}) {
        if ($ref->{$field}) {
          my $pk_table = $fk_ref->{$field}{UK_TABLE_NAME};
          my $pk_field = $fk_ref->{$field}{UK_COLUMN_NAME};
          push @{$ref->{$field}{COLUMN_KEYS}}, "f $pk_table.$pk_field";
        }
      }

    }

  }

  return ($ref);
}

sub audit {
  my $dbh     = Bugzilla->dbh;
  my $lc_hash = Bugzilla->localconfig;

  my %con = (
    host     => $lc_hash->{db_host},
    name     => $lc_hash->{db_name},
    user     => $lc_hash->{db_user},
    password => $lc_hash->{db_pass},
    type     => $lc_hash->{db_driver},
    port     => $lc_hash->{db_port},
  );

  my @sc_tables = $dbh->bz_table_list();
  my $sc_tables;
  map($sc_tables->{$_} = 1, @sc_tables);

  my @db_tables = $dbh->bz_table_list_real();
  my $db_tables;
  map($db_tables->{$_} = 1, @db_tables);

  foreach my $table (sort keys %{$sc_tables}) {
    if (defined $db_tables->{$table}) {
      compare_fields($dbh, $table, \%con);
      compare_indexes($dbh, $table, \%con);
    }
    else {
      print "table $table expected but not in DB\n";
    }
  }

  foreach my $table (sort keys %$db_tables) {
    if (!defined $sc_tables->{$table}) {
      print "table $table in DB but not expected\n";
    }
  }

  $dbh->disconnect();

  return;
}

sub compare_fields {
  my $dbh   = shift;
  my $table = shift;
  my $con   = shift;

  my @sc_fields = $dbh->bz_table_columns($table);
  my $sc_fields;
  map($sc_fields->{$_} = 1, @sc_fields);

  my @db_fields = $dbh->bz_table_columns_real($table);
  my $db_fields;
  map($db_fields->{$_} = 1, @db_fields);

  foreach my $field (sort keys %{$sc_fields}) {
    if (defined $db_fields->{$field}) {
      audit_col($dbh, $table, $field, $con);
    }
    else {
      print "field $table.$field is in bz_schema but not in the DB\n";
    }
  }

  foreach my $field (sort keys %$db_fields) {
    if (!defined $sc_fields->{$field}) {
      print "field $table.$field in the DB but not in bz_schema\n";
    }
  }

  return;
}

sub audit_col {
  my $dbh   = shift;
  my $table = shift;
  my $col   = shift;
  my $con   = shift;

  my $sc_col_info     = $dbh->bz_column_info($table, $col);
  my $sc_col_abs_info = Bugzilla::DB::Schema::ABSTRACT_SCHEMA->{$table}{FIELDS};

  my %expected_cols = ();
  %expected_cols = (@{$sc_col_abs_info}) if ($sc_col_abs_info);

  my $fields1 = db_list_table_fields($dbh, $con, $table, $col);

  my $sc_type = "" . $dbh->_bz_real_schema->convert_type($sc_col_info->{TYPE});
  my $db_type
    = "" . ($fields1->{$col}{pg_type} || $fields1->{$col}{mysql_type_name});
  my $new_type = 'unknown';
  $new_type = $dbh->_bz_real_schema->convert_type($expected_cols{$col}->{TYPE})
    if ($expected_cols{$col}->{TYPE});

  $sc_type =~ s/\s.*$//;
  my $errors = 0;

## BUGBUG not sure this can be made accurate
  if ($types) {
    if ( ($sc_type !~ /^$db_type/i)
      && ($db_type !~ /^$sc_type/i)
      && (lc($db_type) ne lc($sc_type)))
    {
      $errors++;
      print(
        "$table.$col is $sc_type in schema and $db_type in database, and $new_type in a new DB. Raw type $sc_col_info->{TYPE}\n"
      );

      print(Dumper($sc_col_info, $fields1, $sc_col_abs_info)) if ($verbose);
    }
  }

## BUGBUG not sure this can be made accurate
  if ($nulls) {
    my $exp_null = ($expected_cols{$col}->{NOTNULL} ? 'NOT NULLABLE' : 'NULLABLE');
    if ($sc_col_info->{NOTNULL} && $fields1->{$col}{NULLABLE}) {
      $errors++;
      print(
        "$table.$col is NOT NOTNULL in schema and NULLABLE in database and $exp_null in a new DB\n"
      );

    }
    elsif (!defined($sc_col_info->{NOTNULL}) && $fields1->{$col}{NULLABLE}) {
      $errors++;
      print(
        "$table.$col is NULLABLE in schema and NOT NULLABLE in database and is $exp_null in a new DB\n"
      );

      if (defined($sc_col_abs_info)) {

##            print( Dumper( $sc_col_info, $expected_cols{$col} ) );
      }
      else {
        print(
          "$table.$col is in the schema and the database, but Bugzilla doesn't believe it should exist!\n"
        );
      }

    }
  }

  if ($keys) {

    my $sc_fk = $sc_col_info->{REFERENCES};
    my $sc_pk = $sc_col_info->{PRIMARYKEY};
##$dbh->trace('1|pgstart');
    my $sth
      = $dbh->foreign_key_info(undef, undef, undef, undef, undef, $table) || return;

    my $table_fks = $sth->fetchall_hashref("FK_COLUMN_NAME");
##print("table_fks: ", Dumper($table_fks)) if($verbose);
    my $db_fk = $table_fks->{$col};
##print("sc_col_info: ", Dumper($sc_col_info)) if($verbose);
##print("db_fk: ", Dumper($db_fk)) if($verbose);
##print("sc_fk: ", Dumper($sc_fk)) if($verbose);

## BUGBUG getting the key type is complicated because the key type can vary between dbs and key types
    my $col_name = '';
    my $tab_name = '';
    if ($db_fk) {
      $col_name = ($db_fk->{PKCOLUMN_NAME} || 'NULL');
      $tab_name = ($db_fk->{PKTABLE_NAME}  || 'NULL');
    }

    if ((
      $sc_fk
      && (!$db_fk
        || (($sc_fk->{COLUMN} ne $col_name) || ($sc_fk->{TABLE} ne $tab_name)))
      && (!defined($sc_fk->{DELETE})) ## BUGBUG MySQL seems to misreport delete type constraints ...
    ))
    {
      print(
        "FK mismatch: $table.$col is an FK pointing to $sc_fk->{TABLE}.$sc_fk->{COLUMN} in the schema and $tab_name.$col_name in the DB\n"
      );
      print(Dumper($sc_fk, $db_fk)) if ($verbose);
    }

    if (
         $sc_fk
      && $db_fk
      && $sc_fk->{COLUMN} eq $col_name
      && $sc_fk->{TABLE} eq $tab_name

      )
    {
      print("FK's match\n") if ($verbose);
    }

    if (!$sc_fk
      && $db_fk
      && $db_fk->{FK_NAME} ne 'PRIMARY'
      && $db_fk->{FK_NAME} !~ /_idx$/)
    {
      print(
        "$table.$col is not a key the schema and is a key to $tab_name.$col_name the DB\n"
      );
      print(Dumper($sc_fk, $db_fk)) if ($verbose);
    }
  }

  return;
}

sub compare_indexes {
  my $dbh   = shift;
  my $table = shift;
  my $con   = shift;

  my $schema  = undef;
  my $catalog = undef;

  # get a list of table names from the database
  if ($con->{type} =~ m/pg/i) {
    $catalog = 'pg_catalog';
    $schema  = 'public';
  }

  my $sc_indexes = $dbh->bz_table_indexes($table);

  my $sth  = $dbh->statistics_info($catalog, $schema, $table, 0, 1);
  my $hkey = 'INDEX_NAME';
  my $tkey = 'TYPE';
  my $db_indexes;

  if ($sth) {
    $db_indexes = $sth->fetchall_hashref($hkey);
  }
  else {
    $sth = $dbh->prepare("SHOW INDEXES from $table");
    $sth->execute();
    while (my $ref = $sth->fetchrow_hashref()) {
      $db_indexes->{$ref->{Key_name}} = $ref;
    }
    $tkey = 'Index_type';
  }

  foreach my $index (sort keys(%{$sc_indexes})) {
    if (
      !defined $db_indexes->{$index} && none {/^idx_\d+_$index$/}
      keys(%$db_indexes)
      )
    {
      print "index $table.$index is in bz_schema but not in the DB\n";
      print(Dumper($sc_indexes->{$index})) if ($verbose);
    }
  }

  foreach my $index (sort keys %$db_indexes) {
    next if ($db_indexes->{$index}{$tkey} eq 'table');
    next if ($index =~ /_primary$/);
    next if ($index =~ /^PRIMARY$/);
    next if ($index =~ /^idx_\d+_fk_/);
    next if ($index =~ /^fk_/);
    next if ($index =~ /_pkey$/);
    if (!defined $sc_indexes->{$index}) {
      $index =~ /^idx_\d+_(.*)$/;
      if (!$1 || !defined $sc_indexes->{$1}) {
        print "index $table.$index in the DB but not in bz_schema\n";
        print(Dumper($db_indexes->{$index})) if ($verbose);
      }
    }
  }

  return;
}

