# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat::Job::Termination;
use strict;
use warnings;
use 5.10.1;

BEGIN {
  ## no critic (ProhibitStringyEval)
# This is safe and avoids having to conditionally use this module if TheSchwartz is not installed
  eval "use parent qw(TheSchwartz::Worker Bugzilla::Extension)";
  ## use critic
}

# The longest we expect a job to possibly take, in seconds.
use constant grab_for => 600;

# We don't want to fail permanently very easily. Retry for 48 hours.
use constant max_retries => 53;

# The first few retries happen quickly, but after that we wait an hour for
# each retry.
sub retry_delay {
  my ($class, $num_retries) = @_;
  if ($num_retries < 5) {
    return (10, 30, 60, 300, 600)[$num_retries];
  }

  # One hour
  return 60 * 60;
}

sub _handle_bug {
  my $job = shift;

  my $bug_id     = $job->arg->{bug_id};
  my $bug = Bugzilla::Bug->new($bug_id);
  Bugzilla->logger->debug("terminating user on bug " . $bug->bug_id );
  
  my $action     = $job->arg->{action};
  my $login      = $job->arg->{login};
  my $actor      = Bugzilla::User->check($job->arg->{actor});
  my $other_user = Bugzilla::User->check($login);
  my $message    = $job->arg->{message};

  my $spam_product;
  my $spam_prod_name    = Bugzilla->params->{'spam_product'};
  my $spam_component    = Bugzilla->params->{'spam_component'};
  my $nobody_user_name  = Bugzilla->params->{'nobody_user'};
  my $nobody_group_name = Bugzilla->params->{'nobody_group'};

  my $dbh = Bugzilla->dbh;

  # if we are terminating a spammer and they opened a bug, nuke it from orbit.
  if (($action eq 'repudiate') && ($bug->reporter->id == $other_user->id)) {
    $spam_product = Bugzilla::Product->new({name => $spam_prod_name});
    $spam_product->set_is_active(1);
    $spam_product->update();

    my %args = (
      product             => $spam_prod_name,
      component           => $spam_component,
      version             => '1.0',
      status              => 'CLOSED',
      resolution          => 'NOTABUG',
      assigned_to         => $nobody_user_name,
      qa_contact          => '',
      docs_contact        => '',
      target_milestone    => '---',
      target_release      => '---',
      reporter_accessible => 0,
      cclist_accessible   => 0,
    );

    # Now update them all
    $args{groups} = {
      add    => [$nobody_group_name],
      remove => [map { $_->name } @{$bug->groups_in}]
    };
    $args{cc} = {remove => [map { $_->login } @{$bug->cc_users}]};
    $bug->set_all(\%args);
    my ($spam_changes, $delta_ts, $old_bug) = $bug->update();
    if ($spam_changes) {

      # This stops bug mail being sent
      $dbh->do('UPDATE bugs SET lastdiffed = delta_ts WHERE bug_id = ?',
        undef, $bug->id);
    }
    $spam_product->set_is_active(0);
    $spam_product->update();
  }
  else {
    # Replace the user with the default contact for assignee/qa/docs on all
    # open bugs and remove user from CC
    if ($bug->assigned_to->id == $other_user->id) {
      $bug->reset_assigned_to();
    }

    $bug->reset_qa_contact()
      if ($bug->qa_contact && $bug->qa_contact->id == $other_user->id);

    $bug->reset_docs_contact()
      if ($bug->docs_contact && $bug->docs_contact->id == $other_user->id);

    $bug->remove_cc($other_user);

    $bug->add_comment($message, {isprivate => 1});

    my ($changes, $delta_ts, $old_bug) = $bug->update();
    $bug->send_changes($changes, undef, 1);

    # Fire rules engine after bug update email has been sent.
    Bugzilla::Hook::process(
      'bug_end_of_process',
      {
        bug       => $bug,
        timestamp => $delta_ts,
        changes   => $changes,
        old_bug   => $old_bug,
      }
    );
  }

  return;
}

sub work {
  my ($class, $job) = @_;

  Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));

  eval { _handle_bug($job); };

  if ($@) {
    $job->failed($@);
    undef $@;
  }
  else {
    $job->completed;
  }

  return;
}

1;
