# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat::WebService::Flag;

use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::WebService);
use Bugzilla;
use Bugzilla::User;
use Bugzilla::Bug;
use Bugzilla::BugMail;
use Bugzilla::Flag;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(diff_arrays);

use constant PUBLIC_METHODS => qw(
  update
);

# function to update flag
# $call = $rpc->call( 'Flag.update',
#                          {ids     => [1,2]
#                           updates => [
#                               name      => '<flag_name'>,
#                               status    => '<?|+|-|X>',
#                               requestee => '<flag_requestee>', # optional
#                               new       => '<1|0>', # optional
#                               attach_id => '<attach_id>', #optional
#                          }, ....
#                   );
sub update {
  my ($self, $params) = @_;
  my $dbh  = Bugzilla->dbh;
  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $dbh->bz_start_transaction();
  $params->{ids}     || ThrowCodeError('param_required', {param => 'ids'});
  $params->{updates} || ThrowCodeError('param_required', {param => 'updates'});

  my @ids = ref($params->{ids}) ? @{$params->{ids}} : ($params->{ids});
  my @updates
    = ref($params->{updates}) ? @{$params->{updates}} : ($params->{updates});

  my %changes = ();
  foreach my $bug_id (@ids) {
    my $bug = Bugzilla::Bug->check($bug_id);

    # we have to check if the user can edit the bug the product
    # is currently in, before we allow them to change anything.
    Bugzilla->user->can_edit_product($bug->product_id)
      || ThrowUserError("product_edit_denied", {product => $bug->product});

# generate the flag update lookup tables, which include following information
# %type_id_map: type->id map of all possible flagtypes for a specific bug(attachment)
# %requesteeable: indicates if a flagtype is a requesteeable flag
# %multiplicable: indicates if a flagtype is a mulitiplicable flag
# %flag_id_map: flag->flag_id of the flags which have been set for a specific bug(attachment)
    my $bug_flags    = _create_bug_flag_lookup_tables($bug);
    my $attach_flags = _create_attach_flag_lookup_tables($bug);

    my (@flags, @new_flags, %attachment_flags);
    my (%flags_id_map, $flagtypes_info);

    foreach my $data (@updates) {

      # get attachment if 'attach_id' option has been set
      my $attachment;
      if (my $attach_id = $data->{attach_id}) {

        # check if the passed attach_id is a valid attach_id
        foreach my $attach (@{$bug->attachments}) {
          next if $attach->id != $attach_id;
          $attachment = $attach;
        }
        ThrowUserError('invalid_attach_id', {attach_id => $attach_id})
          unless $attachment;
        %flags_id_map   = %{$attach_flags->{flag_id_map}{$attach_id}};
        $flagtypes_info = $attach_flags;
      }
      else {
        # no 'attach_id' option set, get bug flags information
        %flags_id_map   = %{$bug_flags->{flag_id_map}};
        $flagtypes_info = $bug_flags;
      }

      # validate passed arguments
      my ($flag, $flag_id, $requestee)
        = _validate_update_data($data, \%flags_id_map, $flagtypes_info);

      if ($flag_id && (not $data->{'new'})) {

        # flag has been set for this flagtype, update flag
        my %flag_data = (id => $flag_id, status => $data->{'status'});
        if ($requestee) {
          $flag_data{requestee} = $requestee;
        }
        if ($attachment) {
          if (not defined $attachment_flags{$attachment->id}) {
            $attachment_flags{$attachment->id}
              = {obj => $attachment, flags => [], new_flags => [],};
          }
          push(@{$attachment_flags{$attachment->id}{flags}}, \%flag_data);
        }
        else {
          push(@flags, \%flag_data);
        }
      }
      elsif ((not $flag_id) || $data->{'new'}) {

        # flag hasn't been set or need a new flag, update flagtype
        my %flag_data = (type_id => $flagtypes_info->{type_id_map}{$flag},
          status => $data->{status});
        if ($requestee) {
          $flag_data{requestee} = $requestee;
        }
        if ($attachment) {
          if (not defined $attachment_flags{$attachment->id}) {
            $attachment_flags{$attachment->id}
              = {obj => $attachment, flags => [], new_flags => []};
          }
          push(@{$attachment_flags{$attachment->id}{new_flags}}, \%flag_data);
        }
        else {
          push(@new_flags, \%flag_data);
        }
      }
      else {
        ThrowUserError('update_flags_invalid_flag_name');
      }
    }

    $bug->set_flags(\@flags, \@new_flags);
    foreach my $attachment_flag (keys %attachment_flags) {
      $attachment_flags{$attachment_flag}{obj}->set_flags(
        $attachment_flags{$attachment_flag}{flags},
        $attachment_flags{$attachment_flag}{new_flags},
      );
      $attachment_flags{$attachment_flag}{obj}->update();
    }
    $bug->update();

  }

  $dbh->bz_commit_transaction();

  $params->{nomail} ||= 0;
  foreach my $bug_id (@ids) {
    my $bug          = Bugzilla::Bug->check($bug_id);
    my $mail_results = Bugzilla::BugMail::Send($bug->id,
      {changer => Bugzilla->user, nomail => $params->{nomail}});
    foreach my $key (keys %$mail_results) {
      $changes{$bug->id}{$key} = $mail_results->{$key};
    }
  }

  return {flag_updates => \%changes};
}

sub _create_bug_flag_lookup_tables {

  my $bug = shift;

  # The types of flags that can be set on this bug.
  my $bug_flag_types = $bug->flag_types;

  # create the flagname to flag_id map for bug flags
  my %flag_id_map;
  my %flagtype_id_map;
  my %flagtype_requesteeable;
  my %flagtype_multiplicable;
  foreach my $type (@{$bug_flag_types}) {
    foreach my $flag (@{$type->{'flags'}}) {
      push(@{$flag_id_map{$type->name}}, $flag->id);
    }
    $flagtype_id_map{$type->{'name'}} = $type->{'id'};
    if ($type->{'is_requesteeble'}) {
      $flagtype_requesteeable{$type->name} = 1;
    }
    if ($type->{'is_multiplicable'}) {
      $flagtype_multiplicable{$type->name} = 1;
    }
  }

  return {
    type_id_map   => \%flagtype_id_map,
    requesteeable => \%flagtype_requesteeable,
    multiplicable => \%flagtype_multiplicable,
    flag_id_map   => \%flag_id_map
  };

}

sub _create_attach_flag_lookup_tables {

  my $bug = shift;

  # create the flagname to flag_id map for attachment flags
  my %flagtype_id_map;
  my %flagtype_requesteeable;
  my %flagtype_multiplicable;
  my %flag_id_map;

  # The types of flags that can be set on this bug's attachment
  my ($product_id, $component_id) = ($bug->product_id, $bug->component_id);
  my $attach_flag_types = Bugzilla::FlagType::match({
    'target_type'  => 'attachment',
    'product_id'   => $product_id,
    'component_id' => $component_id
  });
  my %attach_id_flagtype_map;
  foreach my $flagtype (@$attach_flag_types) {
    $attach_id_flagtype_map{$flagtype->{id}} = $flagtype->{name};
    if ($flagtype->{'is_requesteeble'}) {
      $flagtype_requesteeable{$flagtype->{name}} = 1;
    }
    if ($flagtype->{'is_multiplicable'}) {
      $flagtype_multiplicable{$flagtype->{name}} = 1;
    }
    $flagtype_id_map{$flagtype->{name}} = $flagtype->{id};
  }

  # create the flagname to flag_id map for attachment flags
  foreach my $attachment (@{$bug->attachments}) {

    # create flags_id_map for each attachment
    my $flags = $attachment->{flags};
    my %flag_ids_map;
    foreach my $flag (@$flags) {
      my $type_id   = $flag->{'type_id'};
      my $type_name = $attach_id_flagtype_map{$type_id};
      push(@{$flag_ids_map{$type_name}}, $flag->{'id'});
    }
    $flag_id_map{$attachment->id} = \%flag_ids_map;
  }

  return {
    type_id_map   => \%flagtype_id_map,
    requesteeable => \%flagtype_requesteeable,
    multiplicable => \%flagtype_multiplicable,
    flag_id_map   => \%flag_id_map
  };

}

sub _validate_update_data {

  my ($data, $flags_id_map, $flagtypes_info) = @_;
  my %flag_id_map = %$flags_id_map;

  # check flag name
  my $flag = $data->{'name'};

  if (!$flag || !$flagtypes_info->{type_id_map}{$flag}) {
    ThrowUserError('update_flags_invalid_flag_name');
  }

  # check passed requestee
  my $requestee = $data->{'requestee'};
  if ($requestee) {

    # check if the new requestee is a valid user
    my $uid = Bugzilla::User::login_to_id($requestee);
    $uid || ThrowUserError('invalid_username', {name => $requestee});
  }

  # check if the flagtype is requesteeable if 'requestee' has been passed
  if ($requestee && !$flagtypes_info->{requesteeable}{$flag}) {
    ThrowUserError('update_flagrequestees_invalid_flag_name');
  }

  my $flag_id;
  if (%flag_id_map && $flag_id_map{$flag}) {

    # if a flagtype has more than one flags, need flag_id for updating flag
    if (@{$flag_id_map{$flag}} > 1 && !$data->{'flag_id'} && !$data->{'new'}) {
      ThrowUserError('update_flags_need_flag_id');
    }

    # check the passed flag_id
    $flag_id = $flag_id_map{$flag}->[0] if $flag_id_map{$flag};
    if ($data->{'flag_id'}) {
      my $valid_flag_id = 0;
      foreach my $id (@{$flag_id_map{$flag}}) {
        next if $id != $data->{'flag_id'};
        $flag_id       = $data->{'flag_id'};
        $valid_flag_id = 1;
      }
      if (!$valid_flag_id) {
        ThrowUserError('update_flags_invalid_flag_id');
      }
    }
  }

  return ($flag, $flag_id, $requestee);

}

1;

__END__

=head1 NAME

Bugzilla::WebService::Flag - The Flag API

=head1 DESCRIPTION

This part of the Bugzilla API allows you to get information on all 
flags (including attachment flags) associated with a specific set of 
bugs.  Also allows updating of flag information for a specific set of 
bugs.

=head1 METHODS

See L<Bugzilla::WebService> for a description of what B<STABLE>, B<UNSTABLE>,
and B<EXPERIMENTAL> mean, and for more information about error codes.

=over

=item C<update> B<EXPERIMENTAL>

=item B<Description>

This allows you to update flag (including attachment's flag) for one or more bugs.

=item B<Params>

=over

=item C<ids> (array), or (int) > - An array of integers, a single integer representing 
one or more bug ids.

=item C<nomail> (boolean) B<Optional> - A boolean value to indicate if mail
with the bug changes should be sent to the relevant users. If set to 1 then
mail will be suppressed, and if ignored or set to 0 then mail will be sent.

=item C<updates> (hash) B<Required> - An array of hashes containing flag fields
and the new values it will be updated to. The hash may contain any of the following:

=over

=item C<name> (string) - The name of the flag that supposes to be updated 

=item C<status> ("?", "+", "-" or "X") - The flag's new status

=item C<requestee> (string) B<Optional> - The new requestee if the flagtype is requesteeable

=item C<flag_id> (id) B<Optional> - Use flag_id to specify the updating flag if there are
more than one flags have been set for a specific flagtype

=item C<new> (boolean) B<Optional> - If this option is set to "1", a new flag will be created
and set new flag values against the newly created flag.

=item C<attach_id> (id) B<Optional> - If this option is set, update attachment's flag
instead of bug flag.

=back

=back

=back

