# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::GroupReview;

use strict;
use warnings;
use 5.10.1;

use parent qw(Exporter);

our @EXPORT = qw(
  get_group_membership
  group_membership_review_bug_update
  page_group_review
  start_group_membership_reviews
);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util;
use Bugzilla::Extension::Workflows::Util;
use Bugzilla::Token;
use Data::Dumper;
use Carp qw/longmess/;

sub page_group_review {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  Bugzilla->login(LOGIN_REQUIRED);
  my $user = Bugzilla->user;
  $vars->{'bless_groups'} = $user->bless_groups();
  $vars->{'group'}        = $input->{'group'};
  $vars->{'min_days'}     = $input->{'min_days'};

  if ($vars->{'group'}) {
    my $args = {group => $vars->{'group'}};

    $args->{min_days} = $vars->{'min_days'} if ($vars->{'min_days'});
    $vars->{'rows'}   = get_group_membership($args);

  }

  return;
}

sub get_group_membership {
  my ($args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $group_name = $args->{group}
    // ThrowCodeError('param_required', {param => 'group'});

  my $min_days = $args->{min_days} || 0;

  my $group = Bugzilla::Group->check({name => $group_name});

  if (!$group || !$user->can_bless($group->id)) {
    ThrowUserError(
      'auth_failure',
      {
        group  => $group_name,
        reason => 'cant_bless',
        action => 'get_group',
        object => 'group'
      }
    );
  }

  my @members;

  foreach my $member (@{$group->members_non_inherited}) {
    next if ($member->login eq Bugzilla->params->{system_user});
    next if ($member->in_group('admin'));
    next if ($member->is_insider && !$user->is_insider);

    if ($member->days_since_login >= $min_days) {
      my $rev_bug  = $member->get_review_bug($group);
      my $rev_id   = '';
      my $rev_days = '';
      if ($rev_bug) {
        $rev_days = $rev_bug->age;
        $rev_id   = $rev_bug->id;
      }
      my %rec = (
        id               => $member->id,
        login            => $member->login,
        days_since_login => $member->days_since_login,
        review_id        => $rev_id,
        review_days      => $rev_days,
      );
      push(@members, \%rec);
    }
  }

  return \@members;
}


sub start_group_membership_reviews {
  my ($args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $group_name = $args->{group}
    // ThrowCodeError('param_required', {param => 'group'});

  my $user_ids = $args->{ids}
    // ThrowCodeError('param_required', {param => 'ids'});

  my $message = <<EOT;
Due to your membership in the group '$group_name', your account requires reconfirmation.

If you do not follow the forgot password process in a timely manner your account may
be terminated.

EOT

  my $group = Bugzilla::Group->check({name => $group_name});

  if (!$group || !$user->can_bless($group->id)) {
    ThrowUserError(
      'auth_failure',
      {
        group  => $group_name,
        reason => 'cant_bless',
        action => 'get_group',
        object => 'group'
      }
    );
  }

  my %ids;
  map { $ids{$_} = 1 } @{$user_ids};

  # Ignore supplied ids that aren't direct members in the group
  my @members = grep { exists $ids{$_->id} } @{$group->members_non_inherited};

  my @results;

  foreach my $member (@members) {
    next if ($member->login eq Bugzilla->params->{system_user});
    next if ($member->in_group('admin'));
    next if ($member->is_insider && !$user->is_insider);

    # Ignore the user if they already have a review bug
    next if ($member->get_review_bug($group));

    my $rev_bug = $member->create_review_bug($user, $group, $message);

    my %rec = (
      id               => $member->id,
      login            => $member->login,
      days_since_login => $member->days_since_login,
      review_id        => $rev_bug->id,
      review_days      => 0,
    );

    push(@results, \%rec);
  }

  return \@results;
}

sub group_membership_review_bug_update {
  my ($args) = @_;
  my $dbh = Bugzilla->dbh;

  my ($bug, $old_bug, $timestamp, $changes)
    = @$args{qw(bug old_bug timestamp changes)};

  return if ($bug->bug_status eq 'CLOSED');

  my $approver = Bugzilla->user;
  my $subject  = $bug->short_desc;
  $subject =~ m/Group membership review for (.*) in (.*)$/;
  my $uname  = $1;
  my $gname  = $2;
  my $member = Bugzilla::User->check($uname);
  my $group  = Bugzilla::Group->check({name => $gname});

  my $terminate = 0;

  ThrowUserError("approvals_subject_change_blocked")
    if (exists $changes->{'short_desc'});

  my $review = $member->get_review_bug($group);

  return unless ($review);

  ThrowUserError("review_bug_mismatch") if ($review->{bug_id} != $bug->id);

  if (exists $changes->{'flagtypes.name'}) {

    my @flags_added = split(/,\s*/, $changes->{'flagtypes.name'}->[1]);

    foreach my $change (@flags_added) {
      my ($group_name, $value);

      if ($change =~ m/^approvals_request_(.+)([\?\+\-])$/) {
        $group_name = $1;
        $value      = $2;
      }

      next unless ($group_name);

      my $flag_name = "approvals_request_$group_name";
      $group_name =~ s/\./ /g;

      my $flagtype   = Bugzilla::FlagType->check({name => $flag_name});
      my $flag_group = Bugzilla::Group->check({name => $group_name});

      ThrowUserError("approvals_flag_mismatch",
        {group => $flag_group, flag_name => $flag_name})
        unless ($flagtype->id == $flag_group->group_workflow->flag_id);

      if ($value eq '-') {
        $terminate = 1;
      }
      else {
        $terminate = 0;
        last;
      }
    }
  }

  my $all_flags_done = 1;
  foreach my $flag (@{$bug->flags}) {
    my $flag_name = $flag->type->name;
    next unless ($flag_name =~ /^approvals_request_(.+)$/);

    if ($flag->status eq '?') {
      $all_flags_done = 0;
      last;
    }
  }

  if ($all_flags_done) {
    Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));
    my $params;
    my $fresh_bug = Bugzilla::Bug->check($bug->id);
    my $msg       = 'All reviews have been completed';

    if ($terminate) {
      $msg .= "\n\nThis account has been terminated.";
    }
    else {
      $msg .= "\n\nThis account is still enabled.";
    }

    $params->{bug_status} = 'CLOSED';
    $params->{resolution} = 'CURRENTRELEASE';
    $params->{comment}    = {body => $msg};
    $fresh_bug->set_all($params);
    $fresh_bug->update();

    if ($terminate) {
      my %args;
      $args{names}   = [$member->login];
      $args{action}  = 'remove';
      $args{message} = 'Account disabled by group membership review process: Bug '
        . $review->{bug_id};
      Bugzilla->user->manage_user_access(\%args);
    }

    Bugzilla->set_user($approver);
  }

  return;

}


1;


__END__

=head1 NAME

Bugzilla::Extension::Workflows::GroupReview

=head1 DESCRIPTION

This workflow allows group managers to force group members to L<reconfirm|Bugzilla::Extension::RedHat::WebService::Bugzilla/reconfirm> their
account.

Managers are defined as users who have Bless on a group.

The tracker bugs created in this process use the flags from the
L<GroupApproval|Bugzilla::Extension::Workflows::GroupApproval/admin-configuration-page>
process, a manager will require access to this flag to complete the termination
path.

The report is accessed via the I<My Links-E<gt>Workflows-E<gt>Review Group Membership> menu entry.

The report shows a list of members in a group, including how many days it has been
since they last logged in.

Managers will be notified of the number of tracking bugs exceeding 30 days of age
in a message banner on the site.

Limitations and constraints are detailed in the L<API|Bugzilla::Extension::Workflows::WebService::GroupReview>.

=head2 The workflow

=over 4

=item A group manager runs the report for a group

=item The manager selects users to review

=item The manager starts the review[s]

=over 4

=item A tracker bug is opened for each review

=item The user account[s] are locked as per L<reconfirm|Bugzilla::Extension::RedHat::WebService::Bugzilla/reconfirm>

=back

=item either:

=over 4

=item The user uses the forgot password process.

=over 4

=item The users account is fully enabled

=item The tracking bug is automatically closed

=back

=item The timer for the user to act expires

=over 4

=item The manager updates the tracking bug by setting the review flag to '-' and saving the change

=item The user account is automatically L<terminated|Bugzilla::Extension::RedHat::WebService::Bugzilla/remove>

=item The tracking bug is automatically closed

=back

=back

=back

=cut
