
use 5.10.1;
use strict;
use warnings;

use lib qw(. lib t);
use Test::More tests => 9;
use Capture::Tiny ':all';
use Bugzilla;
use Bugzilla::User;
use MIME::Entity;
use Mail::GPG;
use File::Temp qw(tempdir);
use File::Path qw(rmtree);

BEGIN {
  $ENV{PROJECT} = 'tests';
}
my $tmpdir = tempdir(CLEANUP => 1);
$ENV{GNUPGHOME} = $tmpdir;

my $user;
my $gpg = Bugzilla->params->{gpg_cmd};

delete @::ENV{'PATH', 'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};

# Some modules throw undefined errors (notably File::Spec::Win32) if
# PATH is undefined.
$ENV{'PATH'} = '';

my ($stdout, $stderr, $exit);
my $dbh      = Bugzilla->dbh();
my $mail     = 'foo@bugzilla.example.com';
my $password = q{This should be secure enough!};
my $body     = <<'EOM';
@product Bugzilla
@component Security
@version 5.0

Get to work!
EOM

my $entity = MIME::Entity->build(
  To       => 'bugzilla@bugzilla.example.com',
  From     => $mail,
  Subject  => "Please do some work!",
  Data     => [$body],
  Encoding => "quoted-printable",
  Charset  => "iso-8859-1",
);

my %init = (always_trust => 1, homedir => $ENV{GNUPGHOME},);

# BAH! if the agent isn't running then supplying pinenetry-mode causes gpg2 to die :-/
($stdout, $stderr, $exit)
  = capture { system("/usr/bin/gpg-agent --daemon -v") };
$stderr =~ /\d+\.(\d+)\.\d+\sstarted/;
my $gpg_ver = $1 // 0;
if ($gpg_ver == 0) {
  diag explain($stdout);
  diag explain($stderr);
  diag explain($exit);
}

my $agent = $stderr !~ /no gpg-agent/;

if ($agent && $gpg_ver >= 2) {
  $init{extra_args} = ['--pinentry-mode=loopback'];
}

my $mg = Mail::GPG->new(
  gpg_call           => $gpg,
  gnupg_hash_init    => \%init,
  default_passphrase => $password,
);

my $OUT;
my $file_name = $ENV{GNUPGHOME} . '/unsigned.eml';
open($OUT, ">", $file_name) || die("can't open email file");
$entity->print($OUT);
close($OUT);

($stdout, $stderr, $exit)
  = capture { system("/usr/bin/perl -T email_in.pl < $file_name") };
like($stderr, qr{Entity is not OpenPGP signed}, "Unsigned mail");

$file_name = $ENV{GNUPGHOME} . '/test_key';
open($OUT, ">", $file_name) || die("can't open email file");
print($OUT <<EOM);
     %echo Generating a basic OpenPGP key
     Key-Type: DSA
     Key-Length: 1024
     Subkey-Type: ELG-E
     Subkey-Length: 1024
     Name-Real: Joe Tester
     Name-Email: $mail
     Expire-Date: 0
     Passphrase: $password 
     # Do a commit here, so that we can later print "done" :-)
     %commit
     %echo done
EOM

close $OUT;

($stdout, $stderr, $exit) = capture {
  system("GNUPGHOME=$ENV{GNUPGHOME} $gpg --batch --gen-key $file_name")
};

like($stderr, qr{marked as ultimately trusted}, "Create key");
$stderr =~ m/key (.*) marked as ultimately trusted/;
my $key_id = $1;

if (-e '/usr/libexec/gpg-preset-passphrase') {
  ($stdout, $stderr, $exit) = capture {
    system("echo $password | /usr/libexec/gpg-preset-passphrase --preset $key_id");
  };
}

my $signed_mail;
eval { $signed_mail = $mg->mime_sign(entity => $entity, key_id => $key_id,); };
diag($@) if ($@);

my $str = $signed_mail ? $signed_mail->as_string() : 'CAMEL';

like($str, qr{Content-Type: multipart/signed}, "encrypted email");

$file_name = $ENV{GNUPGHOME} . '/signed.eml';
open($OUT, ">", $file_name) || die("can't open email file");
$signed_mail->print($OUT);
close($OUT);

($stdout, $stderr, $exit)
  = capture { system("/usr/bin/perl -T email_in.pl < $file_name") };
like($stderr, qr{Mail verification failed}, "no such user");

$user = Bugzilla::User->new({name => $mail});

unless ($user) {
  $user = Bugzilla::User->create({login_name => $mail, cryptpassword => '*'});
  $user->update();
}

$dbh->do("UPDATE profiles SET public_key = NULL WHERE userid = ?",
  undef, $user->id);

($stdout, $stderr, $exit)
  = capture { system("/usr/bin/perl -T email_in.pl < $file_name") };
like($stderr, qr{Mail verification failed}, "no user key");

# set user PGP key.
my $pk = qx{GNUPGHOME=$ENV{GNUPGHOME} $gpg --export -a '$mail'};
$dbh->do("UPDATE profiles SET public_key = ? WHERE userid = ?",
  undef, $pk, $user->id);
$user->add_key();

($stdout, $stderr, $exit)
  = capture { system("/usr/bin/perl -T email_in.pl < $file_name") };
is($stderr, "", "valid signed email and valid user");

## BUGBUG
# Check bug is in correct status!

SKIP: {
  skip "Not yet implimented", 3;

  ($stdout, $stderr, $exit)
    = capture { system("/usr/bin/perl -T email_in.pl < t/mails/encrypted.eml") };
  like($stderr, qr{}, "valid user signed & encrypted email");

  ($stdout, $stderr, $exit)
    = capture { system("/usr/bin/perl -T email_in.pl < t/mails/valid.eml") };
  like($stderr, qr{}, "valid user signed nonmatching key");

  ($stdout, $stderr, $exit)
    = capture { system("/usr/bin/perl -T email_in.pl < t/mails/valid.eml") };
  like($stderr, qr{}, "valid user signed breech ACL");

}

